/**
   @author Shin'ichiro Nakaoka
*/

#include "App.h"
#include "AppUtil.h"
#include "AppConfig.h"
#include "PluginManager.h"
#include "ExtensionManager.h"
#include "ItemManager.h"
#include "ProjectManager.h"
#include "MenuManager.h"
#include "OptionManager.h"
#include "TimeSyncItemEngine.h"
#include "MainWindow.h"
#include "RootItem.h"
#include "FolderItem.h"
#include "SceneItem.h"
#include "View.h"
#include "MessageView.h"
#include "ItemTreeView.h"
#include "ItemPropertyView.h"
#include "SceneView.h"
#include "SceneProviderItemManager.h"
#include "TimeBar.h"
#include "FileBar.h"
#include "CaptureBar.h"
#include "GraphBar.h"
#include "MultiValueSeqGraphView.h"
#include "MultiSE3SeqGraphView.h"
#include "MultiValueSeqItem.h"
#include "MultiSE3SeqItem.h"
#include "MultiAffine3SeqItem.h"
#include "Vector3SeqItem.h"
#include "PathVariableEditor.h"
#include "DescriptionDialog.h"
#include "Licenses.h"
#include "MovieGenerator.h"
#include "LazyCaller.h"

#ifdef CNOID_ENABLE_OSG
#include "OSGSceneView.h"
#include "OSGSceneItem.h"
#endif

#include <cnoid/Config>
#include <cnoid/ValueTree>
#include <QApplication>
#include <QTextCodec>
#include <QIcon>
#include <QGLFormat>
#include <QTextStream>
#include <QFile>
#include <boost/bind.hpp>

#include <csignal>

#ifdef Q_OS_WIN32
#include <windows.h>
#endif

#include <iostream>

#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace cnoid {

    class AppImpl : public QObject
    {
    public:
        AppImpl(App* self, int& argc, char**& argv);
        ~AppImpl();

        App* self;
        QApplication* qapplication;
        int& argc;
        char**& argv;
        ExtensionManager* ext;
        string appName;
        string vendorName;
        DescriptionDialog* descriptionDialog;
        bool doQuit;

        void initialize(const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList);
        int exec();
        void onSigOptionsParsed(boost::program_options::variables_map& v);
        bool processCommandLineOptions();
        virtual bool eventFilter(QObject* watched, QEvent* event);

        void showInformationDialog();
        void onOpenGLVSyncToggled(bool on);
    };
}


namespace {

    boost::signal<void()> sigAboutToQuit_;

    void onCtrl_C_Input(int p)
    {
        callLater(bind(&MainWindow::close, MainWindow::instance()));
    }

#ifdef Q_OS_WIN32
    BOOL WINAPI consoleCtrlHandler(DWORD ctrlChar)
    {
        callLater(bind(&MainWindow::close, MainWindow::instance()));
        return FALSE;
    }
#endif

}


App::App(int& argc, char**& argv)
{
    impl = new AppImpl(this, argc, argv);
}

#include <iostream>
AppImpl::AppImpl(App* self, int& argc, char**& argv)
    : self(self),
      argc(argc),
      argv(argv)
{
    descriptionDialog = 0;

#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
#if defined(__APPLE__) && defined(__MACH__)
    if(!getenv("QT_GRAPHICSSYSTEM")){
        QApplication::setGraphicsSystem("raster"); // to obtain better rendering performance
    }
#endif
#endif

    doQuit = false;

    qapplication = new QApplication(argc, argv);
}


void App::initialize(const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList)
{
    impl->initialize(appName, vendorName, icon, pluginPathList);
}


void AppImpl::initialize( const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList)
{
    this->appName = appName;
    this->vendorName = vendorName;

    setlocale(LC_ALL, ""); // for gettext

#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
    QTextCodec::setCodecForCStrings(QTextCodec::codecForLocale());
#else

#endif
    qapplication->setApplicationName(appName);
    qapplication->setOrganizationName(vendorName);
    qapplication->setWindowIcon(icon);

    ext = new ExtensionManager("Base", false);

    AppConfig::initialize(appName, vendorName);

    // OpenGL settings
    Mapping* glConfig = AppConfig::archive()->openMapping("OpenGL");
    QGLFormat glfmt;
    glfmt.setSwapInterval(glConfig->get("vsync", glfmt.swapInterval()));
    QGLFormat::setDefaultFormat(glfmt);
    
    MainWindow::initialize(appName, ext);
    MessageView::initialize(ext);
    RootItem::initialize(ext);
    ProjectManager::initialize(ext);

    FileBar::initialize(ext);
    TimeBar::initialize(ext);
    ItemTreeView::initialize(ext);
    ItemPropertyView::initialize(ext);

#ifdef CNOID_ENABLE_OSG
    OSGSceneView::initialize(ext);
#endif

    SceneView::initialize(ext);
    SceneProviderItemManager::initialize(ext);
    GraphBar::initialize(ext);
    ext->addView(new MultiValueSeqGraphView());
    ext->addView(new MultiSE3SeqGraphView());
    TimeSyncItemEngineManager::initialize();
    FolderItem::initialize(ext);
    MultiValueSeqItem::initialize(ext);
    MultiSE3SeqItem::initialize(ext);
    MultiAffine3SeqItem::initialize(ext);
    Vector3SeqItem::initialize(ext);
    SceneItem::initialize(ext);

    initializeMovieGenerator(ext);

#ifdef CNOID_ENABLE_OSG
    OSGSceneItem::initialize(ext);
#endif

    CaptureBar::initialize(ext);
    
    PathVariableEditor::initialize(ext);
    
    ext->menuManager().setPath("/Help").addItem(_("About Choreonoid"))
        ->sigTriggered().connect(bind(&AppImpl::showInformationDialog, this));

    // OpenGL settings
    Action* vsyncItem = ext->menuManager().setPath("/Options").addCheckItem(_("Vertical Sync in OpenGL Rendering"));
    vsyncItem->setChecked(glfmt.swapInterval() > 0);
    vsyncItem->sigToggled().connect(bind(&AppImpl::onOpenGLVSyncToggled, this, _1));

    PluginManager::initialize(ext);
    PluginManager::instance()->doStartupLoading(pluginPathList);

    MainWindow::instance()->installEventFilter(this);

    OptionManager& om = ext->optionManager();
    om.addOption("quit", "quit the application just after it is invoked");
    om.sigOptionsParsed().connect(bind(&AppImpl::onSigOptionsParsed, this, _1));

    // Some plugins such as OpenRTM plugin are driven by a library which tries to catch SIGINT.
    // This may block the normal termination by inputting Ctrl+C.
    // To avoid it, the following signal handliers are set.
    std::signal(SIGINT, onCtrl_C_Input);
    std::signal(SIGTERM, onCtrl_C_Input);

#ifdef Q_OS_WIN32
    // The above SIGINT handler seems to work even on Windows
    // when Choreonoid is compiled as a console-program,
    // and the following handler only works for a console-program, too.
    // Hence the folloing handler for Windows is currently disabled.
    // SetConsoleCtrlHandler(consoleCtrlHandler, TRUE);
#endif

}


App::~App()
{
    if(impl){
        delete impl;
    }
}


AppImpl::~AppImpl()
{
    AppConfig::flush();
    delete qapplication;
}


int App::exec()
{
    return impl->exec();
}


int AppImpl::exec()
{
    MainWindow* mainWindow = MainWindow::instance();

    processCommandLineOptions();

    if(!mainWindow->isVisible()){
        mainWindow->show();
    }

    int result = 0;
    
    if(doQuit){
        MessageView::instance()->flush();
    } else {
        result = qapplication->exec();
    }

    PluginManager::finalize();
    delete ext;
    delete mainWindow;
    
    return result;
}


bool AppImpl::eventFilter(QObject* watched, QEvent* event)
{
    MainWindow* mainWindow = MainWindow::instance();
    if(watched == mainWindow){
        if(event->type() == QEvent::Close){
            sigAboutToQuit_();
            mainWindow->storeWindowStateConfig();
            event->accept();
            return true;
        } 
    }
                
    return false;
}


SignalProxy< boost::signal<void()> > cnoid::sigAboutToQuit()
{
    return sigAboutToQuit_;
}


void AppImpl::onSigOptionsParsed(boost::program_options::variables_map& v)
{
    if(v.count("quit")){
        doQuit = true;
    }
}
    

bool AppImpl::processCommandLineOptions()
{
    if(!ext->optionManager().parseCommandLine(argc, argv)){
        //put error messages
    }

    return false;
}


void AppImpl::showInformationDialog()
{
    if(!descriptionDialog){

        descriptionDialog = new DescriptionDialog();

        descriptionDialog->setWindowTitle(_("About Choreonoid"));

        QFile resource(":/Base/LICENSE");
        if(resource.open(QIODevice::ReadOnly | QIODevice::Text)){
            QTextStream license(&resource);
            descriptionDialog->setDescription(
                QString("Choreonoid Version %1\n\n").arg(CNOID_FULL_VERSION_STRING) +
                license.readAll());
        }
    }

    descriptionDialog->show();
}


void AppImpl::onOpenGLVSyncToggled(bool on)
{
    Mapping* glConfig = AppConfig::archive()->openMapping("OpenGL");
    glConfig->write("vsync", (on ? 1 : 0));

    // show messag dialog to prompt restart here
}
