/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "GLSceneRenderer.h"
#include <cnoid/ValueTree>
#include <cnoid/EigenUtil>
#include <Eigen/StdVector>
#ifdef _WIN32
#include <Windows.h>
#endif
#include <GL/glew.h>
#include <boost/scoped_ptr.hpp>
#include <boost/unordered_map.hpp>
#include <boost/bind.hpp>
#include <boost/variant.hpp>
#include <boost/dynamic_bitset.hpp>
#include <iostream>

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {

#ifdef CNOID_SCENE_GRAPH_USE_FLOAT
#define CNOID_GL_FLOAT GL_FLOAT
    inline void glLoadMatrix(const GLfloat* m) { glLoadMatrixf(m); }
    inline void glMultMatrix(const GLfloat* m) { glMultMatrixf(m); }
    inline void glVertex3(GLfloat x, GLfloat y, GLfloat z) { glVertex3f(x, y, z); }
    inline void glNormal3(GLfloat nx, GLfloat ny, GLfloat nz) { glNormal3f(nx, ny, nz); }
#else
#define CNOID_GL_FLOAT GL_DOUBLE
    inline void glLoadMatrix(const GLdouble* m) { glLoadMatrixd(m); }
    inline void glMultMatrix(const GLdouble* m) { glMultMatrixd(m); }
    inline void glVertex3(GLdouble x, GLdouble y, GLdouble z) { glVertex3d(x, y, z); }
    inline void glNormal3(GLdouble nx, GLdouble ny, GLdouble nz) { glNormal3d(nx, ny, nz); }
#endif

    const bool USE_DISPLAY_LISTS = true;
    const bool USE_VBO = false;
    const bool USE_INDEXING = false;
    const bool SHOW_IMAGE_FOR_PICKING = false;

    const float MinLineWidthForPicking = 5.0f;

    typedef vector<SgAffine3, Eigen::aligned_allocator<SgAffine3> > SgAffine3Array;

    struct TransparentShapeInfo
    {
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
        SgShape* shape;
        unsigned int pickId;
        SgAffine3 V; // view matrix
    };
    typedef shared_ptr<TransparentShapeInfo> TransparentShapeInfoPtr;

    
    /*
      A set of variables associated with a scene node
    */
    class Cache : public Referenced
    {
    public:
        GLuint listID;
        GLuint listIDforPicking;
        bool useIDforPicking;
        GLuint bufferNames[3];
        GLuint size;
        vector<TransparentShapeInfoPtr> transparentShapes;
        
        Cache() {
            listID = 0;
            useIDforPicking = false;
            listIDforPicking = 0;
            for(int i=0; i < 3; ++i){
                bufferNames[i] = GL_INVALID_VALUE;
            }
        }
        ~Cache() {
            if(listID){
                glDeleteLists(listID, 1);
            }
            if(listIDforPicking){
                glDeleteLists(listIDforPicking, 1);
            }
            for(int i=0; i < 3; ++i){
                if(bufferNames[i] != GL_INVALID_VALUE){
                    glDeleteBuffers(1, &bufferNames[i]);
                }
            }
        }
        GLuint& vertexBufferName() { return bufferNames[0]; }
        GLuint& normalBufferName() { return bufferNames[1]; }
        GLuint& indexBufferName() { return bufferNames[2]; }
    };
    typedef intrusive_ptr<Cache> CachePtr;


    struct PreproNode
    {
        PreproNode() : parent(0), child(0), next(0) { }
        ~PreproNode() {
            if(child) delete child;
            if(next) delete next;
        }
        enum { GROUP, TRANSFORM, PREPROCESSED, LIGHT, CAMERA };
        variant<SgGroup*, SgTransform*, SgPreprocessed*, SgLight*, SgCamera*> node;
        SgNode* base;
        PreproNode* parent;
        PreproNode* child;
        PreproNode* next;
        template<class T> void setNode(T* n){
            node = n;
            base = n;
        }
    };
    
    
    class PreproTreeExtractor : public SgVisitor
    {
        PreproNode* node;
        bool found;
    public:
        PreproNode* apply(SgNode* node);
        virtual void visitGroup(SgGroup* group);
        virtual void visitTransform(SgTransform* transform);
        virtual void visitShape(SgShape* shape);
        virtual void visitPointSet(SgPointSet* pointSet);        
        virtual void visitLineSet(SgLineSet* lineSet);        
        virtual void visitPreprocessed(SgPreprocessed* preprocessed);
        virtual void visitLight(SgLight* light);
        virtual void visitCamera(SgCamera* camera);
    };
}


PreproNode* PreproTreeExtractor::apply(SgNode* snode)
{
    node = 0;
    found = false;
    snode->accept(*this);
    return node;
}


void PreproTreeExtractor::visitGroup(SgGroup* group)
{
    bool foundInSubTree = false;

    PreproNode* self = new PreproNode();
    self->setNode(group);

    for(SgGroup::reverse_iterator p = group->rbegin(); p != group->rend(); ++p){

        node = 0;
        found = false;
        
        (*p)->accept(*this);
        
        if(node){
            if(found){
                node->parent = self;
                node->next = self->child;
                self->child = node;
                foundInSubTree = true;
            } else {
                delete node;
            }
        }
    }
            
    found = foundInSubTree;

    if(found){
        node = self;
    } else {
        delete self;
        node = 0;
    }
}


void PreproTreeExtractor::visitTransform(SgTransform* transform)
{
    visitGroup(transform);
    if(node){
        node->setNode(transform);
    }
}


void PreproTreeExtractor::visitShape(SgShape* shape)
{

}


void PreproTreeExtractor::visitPointSet(SgPointSet* shape)
{

}


void PreproTreeExtractor::visitLineSet(SgLineSet* shape)
{

}


void PreproTreeExtractor::visitPreprocessed(SgPreprocessed* preprocessed)
{
    node = new PreproNode();
    node->setNode(preprocessed);
    found = true;
}


void PreproTreeExtractor::visitLight(SgLight* light)
{
    node = new PreproNode();
    node->setNode(light);
    found = true;
}


void PreproTreeExtractor::visitCamera(SgCamera* camera)
{
    node = new PreproNode();
    node->setNode(camera);
    found = true;
}


namespace cnoid {

    class GLSceneRendererImpl
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
        
        GLSceneRendererImpl(GLSceneRenderer* self);
        ~GLSceneRendererImpl();

        GLSceneRenderer* self;

        boost::signal<void()> sigRenderingRequest;

        SgGroupPtr root;

        SgNodePath indexedEntities;

        SgAffine3Array Vstack; // stack of the mode/view matrices

        struct Buf {
            SgVertexArray vertices;
            SgIndexArray triangles;
            SgNormalArray normals;
        };
        scoped_ptr<Buf> buf;
        
        struct SgObjectPtrHash {
            std::size_t operator()(const SgObjectPtr& p) const {
                return boost::hash_value<SgObject*>(p.get());
            }
        };
        typedef unordered_map<SgObjectPtr, CachePtr, SgObjectPtrHash> CacheMap;
        CacheMap cacheMaps[2];
        bool doUnusedCacheCheck;
        bool isCheckingUnusedCaches;
        bool hasValidNextCacheMap;
        bool isCacheClearRequested;
        int currentCacheMapIndex;
        CacheMap* currentCacheMap;
        CacheMap* nextCacheMap;
        Cache* currentCache;

        int currentCacheTopViewMatrixIndex;

        bool doPreprocessedNodeTreeExtraction;
        scoped_ptr<PreproNode> preproTree;

        Array4i viewport;

        struct CameraInfo
        {
            EIGEN_MAKE_ALIGNED_OPERATOR_NEW

            CameraInfo() : camera(0), M(SgAffine3::Identity()), node(0) { }
            CameraInfo(SgCamera* camera, const SgAffine3& M, PreproNode* node)
                : camera(camera), M(M), node(node) { }
            SgCamera* camera;

            // I want to use 'T' here, but that causes a compile error (c2327) for VC++2010.
            // 'T' is also used as a template parameter in a template code of Eigen,
            // and it seems that the name of a template parameter and this member conflicts.
            // So I changed the name to 'M'.
            // This behavior of VC++ seems stupid!!!
            SgAffine3 M;
            PreproNode* node;
        };

        typedef vector<CameraInfo, Eigen::aligned_allocator<CameraInfo> > CameraInfoArray;
        CameraInfoArray cameras1;
        CameraInfoArray cameras2;
        CameraInfoArray* cameras;
        CameraInfoArray* prevCameras;

        bool camerasChanged;
        bool currentCameraRemoved;
        int currentCameraIndex;
        SgCamera* currentCamera;
        signal<void()> sigCamerasChanged;
        vector<SgNodePath> cameraPaths;
        
        GLfloat aspectRatio; // width / height;
        GLfloat zNear;
        GLfloat zFar;

        SgTransformPtr defaultCameraTransform;
        SgPerspectiveCameraPtr defaultPersCamera;
        SgOrthographicCameraPtr defaultOrthoCamera;
        CameraInfo defaultCameraInfo;
        int numDefaultCameras;
        bool isDefaultCameraCurrent;

        Affine3 lastViewMatrix;
        Matrix4 lastProjectionMatrix;

        struct LightInfo
        {
            EIGEN_MAKE_ALIGNED_OPERATOR_NEW

            LightInfo() : light(0) { }
            LightInfo(SgLight* light, const SgAffine3& M) : light(light), M(M) { } 
            SgLight* light;
            SgAffine3 M;
        };
        vector<LightInfo, Eigen::aligned_allocator<LightInfo> > lights;

        int numSystemLights;
        int prevNumLights;

        Vector3f bgColor;

        SgLightPtr headLight;
        std::set<SgLightPtr> defaultLights;
        bool additionalLightsEnabled;

        GLSceneRenderer::PolygonMode polygonMode;
        bool defaultLighting;
        SgMaterialPtr defaultMaterial;
        Vector4f defaultColor;
        GLfloat defaultPointSize;
        GLfloat defaultLineWidth;
        
        bool doNormalVisualization;
        SgFloat normalLength;

        bool isCompiling;
        bool isPicking;

        GLdouble pickX;
        GLdouble pickY;
        typedef shared_ptr<SgNodePath> SgNodePathPtr;
        SgNodePath currentNodePath;
        vector<SgNodePathPtr> pickingNodePathList;
        SgNodePath pickedNodePath;
        Vector3 pickedPoint;

        vector<TransparentShapeInfoPtr> transparentShapeInfos;

        MappingPtr property;
        boost::signal<void()> sigPropertyUpdated;

        // OpenGL states
        enum StateFlag {
            CURRENT_COLOR,
            DIFFUSE_COLOR,
            AMBIENT_COLOR,
            EMISSION_COLOR,
            SPECULAR_COLOR,
            SHININESS,
            CULL_FACE,
            CCW,
            LIGHTING,
            LIGHT_MODEL_TWO_SIDE,
            BLEND,
            DEPTH_MASK,
            POINT_SIZE,
            LINE_WIDTH,
            NUM_STATE_FLAGS
        };

        dynamic_bitset<> stateFlag;
        
        Vector4f currentColor;
        Vector4f diffuseColor;
        Vector4f ambientColor;
        Vector4f emissionColor;
        Vector4f specularColor;
        float shininess;
        bool isCullFaceEnabled;
        bool isCCW;
        bool isLightingEnabled;
        bool isLightModelTwoSide;
        bool isBlendEnabled;
        bool isDepthMaskEnabled;
        float pointSize;
        float lineWidth;

        void addEntity(SgNode* node);
        void removeEntity(SgNode* node);
        void onSceneGraphUpdated(const SgUpdate& update);
        void updateCameraPaths();
        bool getSimplifiedCameraPathStrings(int index, vector<string>& out_pathStrings);
        void setCurrentCamera(int index, bool doRenderingRequest);
        bool setCurrentCamera(std::vector<std::string>& strings, bool doRenderingRequest);
        bool initializeGL();
        void setViewport(int x, int y, int width, int height);
        void beginRendering(bool doRenderingCommands);
        void extractPreproNodes(PreproNode* node, const SgAffine3& T);
        void renderCamera();
        void getViewVolume(
            const SgOrthographicCamera& camera,
            GLfloat& out_left, GLfloat& out_right, GLfloat& out_bottom, GLfloat& out_top, GLfloat& out_zNear, GLfloat& out_zFar) const;
        void renderLights();
        void renderLight(const SgLight* light, GLint id, const SgAffine3& T);
        void endRendering();
        void render();
        bool pick(int x, int y);
        inline void setPickColor(unsigned int id);
        inline unsigned int pushPickName(SgNode* node, bool doSetColor = true);
        void popPickName();
        void visitInvariantGroup(SgInvariantGroup* group);
        void visitShape(SgShape* shape);
        void visitPointSet(SgPointSet* pointSet);
        void visitLineSet(SgLineSet* lineSet);
        void renderMaterial(SgMaterial* material);
        void renderMesh(SgMesh* mesh);
        void renderTransparentShapes();
        void writeVertexBuffers(SgMesh* mesh, Cache* cache);

        void clearGLState();
        void setColor(const Vector4f& color);
        void setDiffuseColor(const Vector4f& color);
        void setAmbientColor(const Vector4f& color);
        void setEmissionColor(const Vector4f& color);
        void setSpecularColor(const Vector4f& color);
        void setShininess(float shininess);
        void enableCullFace(bool on);
        void setFrontCCW(bool on);
        void enableLighting(bool on);
        void setLightModelTwoSide(bool on);
        void enableBlend(bool on);
        void enableDepthMask(bool on);
        void setPointSize(float size);
        void setLineWidth(float width);
        void getCurrentCameraTransform(SgAffine3& T);
    };
}


GLSceneRenderer::GLSceneRenderer()
{
    impl = new GLSceneRendererImpl(this);
}


GLSceneRendererImpl::GLSceneRendererImpl(GLSceneRenderer* self)
    : self(self)
{
    root = new SgGroup();
    root->setName("Root");
    root->sigUpdated().connect(bind(&GLSceneRendererImpl::onSceneGraphUpdated, this, _1));

    Vstack.reserve(16);
    
    buf.reset(new Buf);

    doUnusedCacheCheck = true;
    currentCacheMapIndex = 0;
    hasValidNextCacheMap = false;
    isCacheClearRequested = false;
    currentCacheMap = &cacheMaps[0];
    nextCacheMap = &cacheMaps[1];

    doPreprocessedNodeTreeExtraction = true;

    bgColor << 0.1f, 0.1f, 0.3f; // dark blue

    aspectRatio = 1.0f;
    zNear = 0.01f;
    zFar = 10000.0f;

    cameras = &cameras1;
    prevCameras = &cameras2;
    currentCameraIndex = 0;
    currentCamera = 0;

    defaultCameraTransform = new SgTransform();
    defaultCameraTransform->setTransform(
        SgCamera::positionLookingAt(
            SgVector3(4.0, 2.0, 1.5), SgVector3(0.0, 0.0, 1.0), SgVector3::UnitZ()));

    defaultPersCamera = new SgPerspectiveCamera();
    defaultPersCamera->setName("Perspective");
    defaultPersCamera->setFieldOfView(0.6978);
    defaultCameraTransform->addChild(defaultPersCamera);
    defaultCameraInfo.camera = defaultPersCamera.get();

    defaultOrthoCamera = new SgOrthographicCamera();
    defaultOrthoCamera->setName("Orthographic");
    defaultOrthoCamera->setHeight(20.0f);
    defaultCameraTransform->addChild(defaultOrthoCamera);

    isDefaultCameraCurrent = true;
    numDefaultCameras = 2;
    root->addChild(defaultCameraTransform);

    lastViewMatrix.setIdentity();
    lastProjectionMatrix.setIdentity();

    numSystemLights = 1;
    prevNumLights = 0;

    headLight = new SgDirectionalLight();
    headLight->setAmbientIntensity(0.0);
    additionalLightsEnabled = true;

    polygonMode = GLSceneRenderer::FILL_MODE;
    defaultLighting = true;
    defaultMaterial = new SgMaterial;
    defaultColor << 1.0f, 1.0f, 1.0f, 1.0f;
    defaultPointSize = 1.0f;
    defaultLineWidth = 1.0f;
    
    doNormalVisualization = false;
    normalLength = 0.0;

    isCompiling = false;
    isPicking = false;
    pickedPoint.setZero();

    stateFlag.resize(NUM_STATE_FLAGS, false);
    clearGLState();

    property = new Mapping();
}


GLSceneRenderer::~GLSceneRenderer()
{
    delete impl;
}


GLSceneRendererImpl::~GLSceneRendererImpl()
{

}


SgGroup* GLSceneRenderer::sceneRoot()
{
    return impl->root.get();
}


void GLSceneRendererImpl::onSceneGraphUpdated(const SgUpdate& update)
{
    if(update.action() & (SgUpdate::ADDED | SgUpdate::REMOVED)){
        doPreprocessedNodeTreeExtraction = true;
    }
    sigRenderingRequest();
}


SignalProxy<boost::signal<void()> > GLSceneRenderer::sigRenderingRequest()
{
    return impl->sigRenderingRequest;
}


SignalProxy<boost::signal<void()> > GLSceneRenderer::sigCamerasChanged() const
{
    return impl->sigCamerasChanged;
}


int GLSceneRenderer::numCameras() const
{
    return impl->cameras->size();
}


const std::vector<SgNode*>& GLSceneRenderer::cameraPath(int index) const
{
    if(impl->cameraPaths.empty()){
        impl->updateCameraPaths();
    }
    return impl->cameraPaths[index];
}


void GLSceneRendererImpl::updateCameraPaths()
{
    vector<SgNode*> tmpPath;
    const int n = cameras->size();
    cameraPaths.resize(n);
    for(int i=0; i < n; ++i){
        CameraInfo& info = (*cameras)[i];
        tmpPath.clear();
        PreproNode* node = info.node;
        while(node){
            tmpPath.push_back(node->base);
            node = node->parent;
        }
        if(!tmpPath.empty()){
            tmpPath.pop_back(); // remove the root node
            vector<SgNode*>& path = cameraPaths[i];
            path.resize(tmpPath.size());
            std::copy(tmpPath.rbegin(), tmpPath.rend(), path.begin());
        }
    }
}


bool GLSceneRenderer::getSimplifiedCameraPathStrings(int index, vector<string>& out_pathStrings) const
{
    return impl->getSimplifiedCameraPathStrings(index, out_pathStrings);
}


bool GLSceneRendererImpl::getSimplifiedCameraPathStrings(int index, vector<string>& out_pathStrings)
{
    if(cameraPaths.empty()){
        updateCameraPaths();
    }
    out_pathStrings.clear();
    if(index < cameraPaths.size()){
        const vector<SgNode*>& cameraPath = cameraPaths[index];
        const string& name = cameraPath.back()->name();
        if(!name.empty()){
            size_t n = cameraPath.size() - 1;
            for(size_t i=0; i < n; ++i){
                const string& element = cameraPath[i]->name();
                if(!element.empty()){
                    out_pathStrings.push_back(element);
                    break;
                }
            }
            out_pathStrings.push_back(name);
        }
    }
    return !out_pathStrings.empty();
}


void GLSceneRenderer::setCurrentCamera(int index)
{
    impl->setCurrentCamera(index, true);
}


void GLSceneRendererImpl::setCurrentCamera(int index, bool doRenderingRequest)
{
    if(index < cameras->size()){
        SgCamera* newCamera = (*cameras)[index].camera;
        if(newCamera != currentCamera){
            currentCameraIndex = index;
            currentCamera = newCamera;
            isDefaultCameraCurrent = (index < numDefaultCameras);
            if(doRenderingRequest){
                sigRenderingRequest();
            }
        }
    }
}


bool GLSceneRenderer::setCurrentCamera(std::vector<std::string>& simplifiedPathStrings)
{
    return impl->setCurrentCamera(simplifiedPathStrings, true);
}


bool GLSceneRendererImpl::setCurrentCamera(std::vector<std::string>& strings, bool doRenderingRequest)
{
    bool found = false;
    
    if(cameraPaths.empty()){
        updateCameraPaths();
    }

    if(!strings.empty()){
        vector<int> candidates;
        const string& name = (strings.size() == 1) ? strings.front() : strings.back();
        for(size_t i=0; i < cameraPaths.size(); ++i){
            const vector<SgNode*> cameraPath = cameraPaths[i];
            if(cameraPath.back()->name() == name){
                candidates.push_back(i);
            }
        }
        if(candidates.size() == 1){
            setCurrentCamera(candidates.front(), doRenderingRequest);
            found = true;
        } else if(candidates.size() >= 2){
            if(strings.size() == 2){
                const string& owner = strings.front();
                for(size_t i=0; i < candidates.size(); ++i){
                    const vector<SgNode*> cameraPath = cameraPaths[i];
                    if(cameraPath.front()->name() == owner){
                        setCurrentCamera(i, doRenderingRequest);
                        found = true;
                    }
                }
            }
        }
    }
    return found;
}


SgCamera* GLSceneRenderer::currentCamera() const
{
    return impl->currentCamera;
}


int GLSceneRenderer::currentCameraIndex() const
{
    return impl->currentCameraIndex;
}


SgTransform* GLSceneRenderer::defaultCameraTransform() const
{
    return impl->defaultCameraTransform.get();
}


SgPerspectiveCamera* GLSceneRenderer::defaultPerspectiveCamera() const
{
    return impl->defaultPersCamera.get();
}


SgOrthographicCamera* GLSceneRenderer::defaultOrthographicCamera() const
{
    return impl->defaultOrthoCamera.get();
}


bool GLSceneRenderer::isDefaultCameraCurrent() const
{
    return impl->isDefaultCameraCurrent;
}


bool GLSceneRenderer::initializeGL()
{
    return impl->initializeGL();
}


bool GLSceneRendererImpl::initializeGL()
{
    GLenum err = glewInit();
    if(err != GLEW_OK){
        return false;
    }

    isCullFaceEnabled = false;
    if(isCullFaceEnabled){
        glEnable(GL_CULL_FACE);
        GLint twoSide = 0;
        isLightModelTwoSide = false;
        glLightModeliv(GL_LIGHT_MODEL_TWO_SIDE, &twoSide);
    } else {
        glDisable(GL_CULL_FACE);
        GLint twoSide = 1;
        isLightModelTwoSide = true;
        glLightModeliv(GL_LIGHT_MODEL_TWO_SIDE, &twoSide);
    }
    
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_NORMALIZE);
    setFrontCCW(true);

    return true;
}


void GLSceneRenderer::setDefaultClippingDepth(double zNear, double zFar)
{
    impl->zNear = zNear;
    impl->zFar = zFar;
}


void GLSceneRenderer::getDefaultClippingDepth(double& zNear, double& zFar)
{
    zNear = impl->zNear;
    zFar = impl->zFar;
}


void GLSceneRenderer::setViewport(int x, int y, int width, int height)
{
    impl->setViewport(x, y, width, height);
}


void GLSceneRendererImpl::setViewport(int x, int y, int width, int height)
{
    aspectRatio = (double)width / height;
    viewport << x, y, width, height;
    glViewport(x, y, width, height);
}


const Array4i& GLSceneRenderer::viewport() const
{
    return impl->viewport;
}


void GLSceneRenderer::getViewport(int& out_x, int& out_y, int& out_width, int& out_height) const
{
    out_x = impl->viewport[0];
    out_y = impl->viewport[1];
    out_width = impl->viewport[2];
    out_height = impl->viewport[3];
}    


double GLSceneRenderer::aspectRatio() const
{
    return impl->aspectRatio;
}


void GLSceneRenderer::requestToClearCache()
{
    impl->isCacheClearRequested = true;
}


void GLSceneRenderer::initializeRendering()
{
    impl->beginRendering(false);
}


void GLSceneRenderer::beginRendering()
{
    impl->beginRendering(true);
}


void GLSceneRendererImpl::beginRendering(bool doRenderingCommands)
{
    isCheckingUnusedCaches = isPicking ? false : doUnusedCacheCheck;

    if(isCacheClearRequested){
        cacheMaps[0].clear();
        cacheMaps[1].clear();
        hasValidNextCacheMap = false;
        isCheckingUnusedCaches = false;
        isCacheClearRequested = false;
    }
    if(hasValidNextCacheMap){
        currentCacheMapIndex = 1 - currentCacheMapIndex;
        currentCacheMap = &cacheMaps[currentCacheMapIndex];
        nextCacheMap = &cacheMaps[1 - currentCacheMapIndex];
        hasValidNextCacheMap = false;
    }
    currentCache = 0;

    if(doRenderingCommands){
        if(isPicking){
            currentNodePath.clear();
            pickingNodePathList.clear();
            glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        } else {
            glClearColor(bgColor[0], bgColor[1], bgColor[2], 1.0f);
        }
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    if(doPreprocessedNodeTreeExtraction){
        PreproTreeExtractor extractor;
        preproTree.reset(extractor.apply(root.get()));
        doPreprocessedNodeTreeExtraction = false;
    }

    std::swap(cameras, prevCameras);
    cameras->clear();
    camerasChanged = false;
    currentCameraRemoved = true;
    
    lights.clear();

    if(preproTree){
        extractPreproNodes(preproTree.get(), SgAffine3::Identity());
    }

    if(!camerasChanged){
        if(cameras->size() != prevCameras->size()){
            camerasChanged = true;
        }
    }
    if(camerasChanged){
        if(currentCameraRemoved){
            currentCameraIndex = 0;
        }
        cameraPaths.clear();
        sigCamerasChanged();
    }

    setCurrentCamera(currentCameraIndex, false);

    if(doRenderingCommands){
        renderCamera();
        if(!isPicking){
            renderLights();
        }
        if(isPicking){
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        } else {
            switch(polygonMode){
            case GLSceneRenderer::FILL_MODE:
                glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                break;
            case GLSceneRenderer::LINE_MODE:
                glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
                break;
            case GLSceneRenderer::POINT_MODE:
                glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
                break;
            }
        }
        clearGLState();
        
        setColor(defaultColor);
        setPointSize(defaultPointSize);
        //glEnable(GL_POINT_SMOOTH);
        setLineWidth(defaultLineWidth);
        //glEnable(GL_LINE_SMOOTH);

        transparentShapeInfos.clear();
    }
}


void GLSceneRendererImpl::extractPreproNodes(PreproNode* node, const SgAffine3& T)
{
    switch(node->node.which()){

    case PreproNode::GROUP:
        for(PreproNode* childNode = node->child; childNode; childNode = childNode->next){
            extractPreproNodes(childNode, T);
        }
        break;
        
    case PreproNode::TRANSFORM:
        {
            SgTransform* transform = get<SgTransform*>(node->node);
            const SgAffine3 T2 = T * transform->T();
            for(PreproNode* childNode = node->child; childNode; childNode = childNode->next){
                extractPreproNodes(childNode, T2);
            }
        }
        break;
        
    case PreproNode::PREPROCESSED:
        // call additional functions
        break;

    case PreproNode::LIGHT:
    {
        SgLight* light = get<SgLight*>(node->node);
        if(additionalLightsEnabled || defaultLights.find(light) != defaultLights.end()){
            lights.push_back(LightInfo(light, T));
        }
        break;
    }

    case PreproNode::CAMERA:
        {
            SgCamera* camera = get<SgCamera*>(node->node);
            size_t index = cameras->size();
            if(!camerasChanged){
                if(index >= prevCameras->size() || camera != (*prevCameras)[index].camera){
                    camerasChanged = true;
                }
            }
            if(camera == currentCamera){
                currentCameraRemoved = false;
                currentCameraIndex = cameras->size();
            }
            cameras->push_back(CameraInfo(camera, T, node));
        }
        break;

    default:
        break;
    }
}


void GLSceneRendererImpl::renderCamera()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // set projection
    if(SgPerspectiveCamera* camera = dynamic_cast<SgPerspectiveCamera*>(currentCamera)){
        gluPerspective(degree(camera->fovy(aspectRatio)), aspectRatio, zNear, zFar);
        
    } else if(SgOrthographicCamera* camera = dynamic_cast<SgOrthographicCamera*>(currentCamera)){
        GLfloat left, right, bottom, top, near_, far_;
        getViewVolume(*camera, left, right, bottom, top, near_, far_);
        glOrtho(left, right, bottom, top, near_, far_);
        
    } else {
        gluPerspective(40.0f, aspectRatio, zNear, zFar);
    }

    GLdouble P[16];
    glGetDoublev(GL_PROJECTION_MATRIX, P);
    lastProjectionMatrix = Eigen::Map<Eigen::Matrix<GLdouble, 4, 4> >(P).cast<Affine3::Scalar>();
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    // Render headlight
    if(!isPicking && defaultLighting){
        glEnable(GL_LIGHTING);

        if(!headLight->on()){
            glDisable(GL_LIGHT0);
        } else {
            static const SgAffine3 I = SgAffine3::Identity();
            renderLight(headLight.get(), GL_LIGHT0, I);
        }
    } else {
        glDisable(GL_LIGHTING);
    }

    Vstack.clear();
    Vstack.push_back((*cameras)[currentCameraIndex].M.inverse(Eigen::Isometry));
    const SgAffine3& V = Vstack.back();
    lastViewMatrix = V.cast<Affine3::Scalar>();
    glLoadMatrix(V.data());
}


void GLSceneRendererImpl::getViewVolume
(const SgOrthographicCamera& camera,
 GLfloat& out_left, GLfloat& out_right, GLfloat& out_bottom, GLfloat& out_top, GLfloat& out_zNear, GLfloat& out_zFar) const
{
    GLfloat h = camera.height();
    out_top = h / 2.0f;
    out_bottom = -h / 2.0f;
    GLfloat w = h * aspectRatio;
    out_left = -w / 2.0f;
    out_right = w / 2.0f;
    out_zNear = -zFar;
    out_zFar = zFar;
}


void GLSceneRendererImpl::renderLights()
{
    GLint maxLights;
    glGetIntegerv(GL_MAX_LIGHTS, &maxLights);
    maxLights -= numSystemLights;
    
    if(lights.size() > maxLights){
        lights.resize(maxLights);
    }
        
    for(size_t i=0; i < lights.size(); ++i){
        const LightInfo& info = lights[i];
        const GLint id = GL_LIGHT0 + numSystemLights + i;
        renderLight(info.light, id, info.M);
    }

    for(size_t i = lights.size(); i < prevNumLights; ++i){
        const GLint lightID = GL_LIGHT0 + numSystemLights + i;
        glDisable(lightID);
    }

    prevNumLights = lights.size();
}


void GLSceneRendererImpl::renderLight(const SgLight* light, GLint id, const SgAffine3& T)
{
    bool isValid = false;

    if(light->on()){
        
        if(const SgDirectionalLight* dirLight = dynamic_cast<const SgDirectionalLight*>(light)){

            isValid = true;
            
            Vector4f pos;
            pos << (T.linear() * -dirLight->direction()).cast<GLfloat>(), 0.0f;
            glLightfv(id, GL_POSITION, pos.data());
            
            /*
              glLightf(id, GL_CONSTANT_ATTENUATION, 0.0f);
              glLightf(id, GL_LINEAR_ATTENUATION, 0.0f);
              glLightf(id, GL_QUADRATIC_ATTENUATION, 0.0f);
            */
            
        } else if(const SgPointLight* pointLight = dynamic_cast<const SgPointLight*>(light)){
            
            isValid = true;
            
            Vector4f pos;
            pos << (T * pointLight->location()).cast<GLfloat>(), 1.0f;
            glLightfv(id, GL_POSITION, pos.data());
            
            glLightf(id, GL_CONSTANT_ATTENUATION, pointLight->attenuation()[0]); 
            glLightf(id, GL_LINEAR_ATTENUATION, pointLight->attenuation()[1]); 
            glLightf(id, GL_QUADRATIC_ATTENUATION, pointLight->attenuation()[2]);
            
            if(const SgSpotLight* spotLight = dynamic_cast<const SgSpotLight*>(pointLight)){
                Vector3f direction = (T.linear() * spotLight->direction()).cast<GLfloat>();
                glLightfv(id, GL_SPOT_DIRECTION, direction.data());
                glLightf(id, GL_SPOT_CUTOFF, degree(spotLight->cutOffAngle()));
                glLightf(id, GL_SPOT_EXPONENT, 0.5f);
                
            } else {
                glLightf(id, GL_SPOT_CUTOFF, 180.0f);
            }
        }
    }
        
    if(!isValid){
        glDisable(id);
        
    } else {
        Vector4f diffuse;
        diffuse << (light->intensity() * light->color()).cast<GLfloat>(), 1.0f;
        glLightfv(id, GL_DIFFUSE, diffuse.data());
        glLightfv(id, GL_SPECULAR, diffuse.data());
        
        Vector4f ambient;
        ambient << (light->ambientIntensity() * light->color()).cast<GLfloat>(), 1.0f;
        glLightfv(id, GL_AMBIENT, ambient.data());
        
        glEnable(id);
    }
}


void GLSceneRenderer::endRendering()
{
    impl->endRendering();
}


void GLSceneRendererImpl::endRendering()
{
    if(isCheckingUnusedCaches){
        currentCacheMap->clear();
        hasValidNextCacheMap = true;
    }
}


void GLSceneRenderer::render()
{
    impl->render();
}


void GLSceneRendererImpl::render()
{
    beginRendering(true);

    root->accept(*self);

    if(!transparentShapeInfos.empty()){
        renderTransparentShapes();
    }

    endRendering();
}


bool GLSceneRenderer::pick(int x, int y)
{
    return impl->pick(x, y);
}


/*
  http://stackoverflow.com/questions/4040616/opengl-gl-select-or-manual-collision-detection
  http://content.gpwiki.org/index.php/OpenGL_Selection_Using_Unique_Color_IDs
  http://www.opengl-tutorial.org/miscellaneous/clicking-on-objects/picking-with-an-opengl-hack/
  http://www.codeproject.com/Articles/35139/Interactive-Techniques-in-Three-dimensional-Scenes#_OpenGL_Picking_by
  http://en.wikibooks.org/wiki/OpenGL_Programming/Object_selection

   Use a Framebuffer object?
*/
bool GLSceneRendererImpl::pick(int x, int y)
{
    glPushAttrib(GL_ENABLE_BIT);

    //glDisable(GL_LIGHTING); // disable this later in 'renderCamera()'
    glDisable(GL_BLEND);
    glDisable(GL_MULTISAMPLE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);

    if(!SHOW_IMAGE_FOR_PICKING){
        glScissor(x, y, 1, 1);
        glEnable(GL_SCISSOR_TEST);
    }
    
    isPicking = true;
    render();
    isPicking = false;

    glPopAttrib();

    GLfloat color[4];
    glReadPixels(x, y, 1, 1, GL_RGBA, GL_FLOAT, color);
    if(SHOW_IMAGE_FOR_PICKING){
        color[2] = 0.0f;
    }
    unsigned int id = (color[0] * 255) + ((int)(color[1] * 255) << 8) + ((int)(color[2] * 255) << 16) - 1;

    pickedNodePath.clear();

    if(0 < id && id < pickingNodePathList.size()){
        pickedNodePath = *pickingNodePathList[id];
        GLfloat depth;
        glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
        GLdouble ox, oy, oz;
        gluUnProject(x, y, depth,
                     lastViewMatrix.matrix().cast<GLdouble>().data(),
                     lastProjectionMatrix.cast<GLdouble>().data(),
                     viewport.cast<GLint>().data(),
                     &ox, &oy, &oz);
        pickedPoint << ox, oy, oz;
    }

    return !pickedNodePath.empty();
}


const std::vector<SgNode*>& GLSceneRenderer::pickedNodePath() const
{
    return impl->pickedNodePath;
}


const Vector3& GLSceneRenderer::pickedPoint() const
{
    return impl->pickedPoint;
}



inline void GLSceneRendererImpl::setPickColor(unsigned int id)
{
    float r = (id & 0xff) / 255.0;
    float g = ((id >> 8) & 0xff) / 255.0;
    float b = ((id >> 16) & 0xff) / 255.0;
    if(SHOW_IMAGE_FOR_PICKING){
        b = 1.0f;
    }
    glColor4f(r, g, b, 1.0f);
    currentColor << r, g, b, 1.0f;
}
        

/**
   @return id of the current object
*/
inline unsigned int GLSceneRendererImpl::pushPickName(SgNode* node, bool doSetColor)
{
    unsigned int id = 0;
    
    if(isPicking && !isCompiling){
        id = pickingNodePathList.size() + 1;
        currentNodePath.push_back(node);
        pickingNodePathList.push_back(make_shared<SgNodePath>(currentNodePath));
        if(doSetColor){
            setPickColor(id);
        }
    }

    return id;
}


inline void GLSceneRendererImpl::popPickName()
{
    if(isPicking && !isCompiling){
        currentNodePath.pop_back();
    }
}


void GLSceneRenderer::visitGroup(SgGroup* group)
{
    impl->pushPickName(group);
    SgVisitor::visitGroup(group);
    impl->popPickName();
}


void GLSceneRenderer::visitInvariantGroup(SgInvariantGroup* group)
{
    impl->visitInvariantGroup(group);
}


void GLSceneRendererImpl::visitInvariantGroup(SgInvariantGroup* group)
{
    if(!USE_DISPLAY_LISTS || isCompiling){
        self->visitGroup(group);

    } else {
        Cache* cache;
        CacheMap::iterator p = currentCacheMap->find(group);
        if(p == currentCacheMap->end()){
            cache = new Cache();
            currentCacheMap->insert(CacheMap::value_type(group, cache));
        } else {
            cache = p->second.get();
        }

        if(!cache->listID && !isPicking){
            currentCache = cache;
            currentCacheTopViewMatrixIndex = Vstack.size() - 1;

            cache->listID = glGenLists(1);
            if(cache->listID){
                glNewList(cache->listID, GL_COMPILE);

                isCompiling = true;
                clearGLState();
                self->visitGroup(group);
                isCompiling = false;

                if(stateFlag[LIGHTING] || stateFlag[CURRENT_COLOR]){
                    cache->useIDforPicking = true;
                }
                glEndList();
            }
        }

        GLuint listID = cache->listID;

        if(listID){
            if(isPicking && cache->useIDforPicking){
                if(!cache->listIDforPicking){
                    currentCache = cache;
                    currentCacheTopViewMatrixIndex = Vstack.size() - 1;
                    cache->listIDforPicking = glGenLists(1);
                    if(cache->listIDforPicking){
                        glNewList(cache->listIDforPicking, GL_COMPILE);
                        isCompiling = true;
                        clearGLState();
                        self->visitGroup(group);
                        isCompiling = false;
                        glEndList();
                    }
                }
                listID = cache->listIDforPicking;
            }
            if(listID){
                const unsigned int pickId = pushPickName(group);
                glPushAttrib(GL_ENABLE_BIT);
                glCallList(listID);
                glPopAttrib();
                clearGLState();
                popPickName();

                const vector<TransparentShapeInfoPtr>& transparentShapes = cache->transparentShapes;
                if(!transparentShapes.empty()){
                    const SgAffine3& V = Vstack.back();
                    for(size_t i=0; i < transparentShapes.size(); ++i){
                        const TransparentShapeInfo& src = *transparentShapes[i];
                        TransparentShapeInfoPtr info = make_shared_aligned<TransparentShapeInfo>();
                        info->shape = src.shape;
                        info->pickId = pickId;
                        info->V = V * src.V;
                        transparentShapeInfos.push_back(info);
                    }
                }
            }

            if(isCheckingUnusedCaches){
                nextCacheMap->insert(CacheMap::value_type(group, cache));
            }
        }
    }
    currentCache = 0;
}


void GLSceneRenderer::visitTransform(SgTransform* transform)
{
    SgAffine3Array& Vstack = impl->Vstack;
    Vstack.push_back(Vstack.back() * transform->T());

    glPushMatrix();

    glMultMatrix(transform->T().data());

    /*
      if(isNotRotationMatrix){
      glPushAttrib(GL_ENABLE_BIT);
      glEnable(GL_NORMALIZE);
      }
    */
    
    visitGroup(transform);
    
    /*
      if(isNotRotationMatrix){
      glPopAttrib();
      }
    */
    
    glPopMatrix();
    Vstack.pop_back();
}


void GLSceneRendererImpl::visitShape(SgShape* shape)
{
    SgMesh* mesh = shape->mesh();
    if(mesh){
        if(mesh->vertices() && !mesh->vertices()->empty()){

            SgMaterial* material = shape->material();
            
            if(material && material->transparency() > 0.0){
                // A transparent shape is rendered later
                TransparentShapeInfoPtr info = make_shared_aligned<TransparentShapeInfo>();
                info->shape = shape;
                if(isCompiling){
                    info->V = Vstack[currentCacheTopViewMatrixIndex].inverse() * Vstack.back();
                    currentCache->transparentShapes.push_back(info);
                } else {
                    info->V = Vstack.back();
                    info->pickId = pushPickName(shape, false);
                    popPickName();
                    transparentShapeInfos.push_back(info);
                }
            } else {
                pushPickName(shape);
                if(!isPicking){
                    if(material){
                        renderMaterial(shape->material());
                    }
                }
                renderMesh(mesh);
                popPickName();
            }
        }
    }
}


void GLSceneRenderer::visitShape(SgShape* shape)
{
    impl->visitShape(shape);
}


void GLSceneRendererImpl::renderMaterial(SgMaterial* material)
{
    float alpha = 1.0 - material->transparency();

    Vector4f color;
    color << material->diffuseColor().cast<GLfloat>(), alpha;
    setDiffuseColor(color);
        
    color.head<3>() *= material->ambientIntensity();
    setAmbientColor(color);

    color << material->emissiveColor().cast<GLfloat>(), alpha;
    setEmissionColor(color);
    
    color << material->specularColor().cast<GLfloat>(), alpha;
    setSpecularColor(color);
    
    float shininess = (127.0 * material->shininess()) + 1.0;
    setShininess(shininess);
}


/**
   \todo sort the shape nodes by the distance from the viewpoint
*/
void GLSceneRendererImpl::renderTransparentShapes()
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();;
    
    if(!isPicking){
        enableBlend(true);
    }
    
    const int n = transparentShapeInfos.size();
    for(int i=0; i < n; ++i){
        TransparentShapeInfo& info = *transparentShapeInfos[i];
        glMatrixMode(GL_MODELVIEW);
        glLoadMatrix(info.V.data());
        SgShape* shape = info.shape;
        if(isPicking){
            setPickColor(info.pickId);
        } else {
            renderMaterial(shape->material());
        }
        renderMesh(shape->mesh());
    }

    if(!isPicking){
        enableBlend(false);
    }

    transparentShapeInfos.clear();

    glPopMatrix();
}


void GLSceneRendererImpl::renderMesh(SgMesh* mesh)
{
    /*
    if(mesh->vertices){
        cout << "vertices: ";
        for(size_t i=0; i < mesh->vertices->size(); ++i){
            const SgVector3& v = mesh->vertices->at(i);
            cout << "(" << v.x() << ", " << v.y() << ", " << v.z() << "), ";
        }
        cout << endl;
    }
    if(!mesh->triangles.empty()){
        cout << "triangles: ";
        for(size_t i=0; i < mesh->triangles.size(); ++i){
            cout << mesh->triangles[i] << ", ";
        }
        cout << endl;
    }
    if(mesh->normals){
        cout << "normals: ";
        for(size_t i=0; i < mesh->normals->size(); ++i){
            const SgVector3& v = mesh->normals->at(i);
            cout << "(" << v.x() << ", " << v.y() << ", " << v.z() << "), ";
        }
        cout << endl;
    }
    if(!mesh->normalIndices.empty()){
        cout << "normalIndices: ";
        for(size_t i=0; i < mesh->normalIndices.size(); ++i){
            cout << mesh->normalIndices[i] << ", ";
        }
        cout << endl;
    }
    */
    
    const bool ENABLE_CULLING = true;
    if(ENABLE_CULLING){
        enableCullFace(mesh->isSolid());
        setLightModelTwoSide(!mesh->isSolid());
    } else {
        enableCullFace(false);
        setLightModelTwoSide(true);
    }

    //glShadeModel(GL_FLAT);
    //glShadeModel(GL_SMOOTH);

    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    if(!USE_VBO){
        writeVertexBuffers(mesh, 0);

    } else {
        CacheMap::iterator it = currentCacheMap->find(mesh);
        if(it == currentCacheMap->end()){
            it = currentCacheMap->insert(CacheMap::value_type(mesh, new Cache)).first;
            writeVertexBuffers(mesh, it->second.get());
        }
        if(isCheckingUnusedCaches){
            nextCacheMap->insert(*it);
        }
        Cache& cache = *it->second;
        if(cache.vertexBufferName() != GL_INVALID_VALUE){
            glEnableClientState(GL_VERTEX_ARRAY);
            glBindBuffer(GL_ARRAY_BUFFER, cache.vertexBufferName());
            glVertexPointer(3, CNOID_GL_FLOAT, 0, 0);
                        
            if(cache.normalBufferName() != GL_INVALID_VALUE){
                glEnableClientState(GL_NORMAL_ARRAY);
                glBindBuffer(GL_ARRAY_BUFFER, cache.normalBufferName());
                glNormalPointer(CNOID_GL_FLOAT, 0, 0);
            }
            if(USE_INDEXING){
                if(cache.indexBufferName() != GL_INVALID_VALUE){
                    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cache.indexBufferName());
                    glDrawElements(GL_TRIANGLES, cache.size, GL_UNSIGNED_INT, 0);
                }
            } else {
                glDrawArrays(GL_TRIANGLES, 0, cache.size);
            }
        }
    }

    glPopClientAttrib();
}


void GLSceneRendererImpl::writeVertexBuffers(SgMesh* mesh, Cache* cache)
{
    SgVertexArray& orgVertices = *mesh->vertices();
    SgIndexArray& orgTriangles = mesh->triangles();
    const size_t numTriangles = orgTriangles.size() / 3;
    const size_t totalNumVertices = orgTriangles.size();
    const bool hasNormals = !mesh->normals()->empty();
    SgVertexArray* vertices = 0;
    SgNormalArray* normals = 0;
    SgIndexArray* triangles = 0;

    if(USE_INDEXING){
        vertices = &orgVertices;
        triangles = &orgTriangles;
        if(hasNormals){
            if(mesh->normalIndices().empty() && mesh->normals()->size() == orgVertices.size()){
                normals = mesh->normals();
            } else {
                normals = &buf->normals;
                normals->resize(orgVertices.size());
            }
        }
    } else {
        vertices = &buf->vertices;
        vertices->reserve(totalNumVertices);
        vertices->clear();
        if(hasNormals){
            normals = &buf->normals;
            normals->reserve(totalNumVertices);
            normals->clear();
        }
    }
    
    int faceVertexIndex = 0;
    int numFaceVertices = 0;
    
    for(size_t i=0; i < numTriangles; ++i){
        for(size_t j=0; j < 3; ++j){
            const int orgVertexIndex = orgTriangles[faceVertexIndex];
            if(!USE_INDEXING){
                vertices->push_back(orgVertices[orgVertexIndex]);
            }
            if(hasNormals){
                if(mesh->normalIndices().empty()){
                    if(!USE_INDEXING){
                        normals->push_back(mesh->normals()->at(orgVertexIndex));
                    }
                } else {
                    const int normalIndex = mesh->normalIndices()[faceVertexIndex];
                    if(USE_INDEXING){
                        normals->at(orgVertexIndex) = mesh->normals()->at(normalIndex);
                    } else {
                        normals->push_back(mesh->normals()->at(normalIndex));
                    }
                }
            }
            ++faceVertexIndex;
        }
    }

    if(USE_VBO){
        if(cache->vertexBufferName() == GL_INVALID_VALUE){
            glGenBuffers(1, &cache->vertexBufferName());
            glBindBuffer(GL_ARRAY_BUFFER, cache->vertexBufferName());
            glBufferData(GL_ARRAY_BUFFER, vertices->size() * sizeof(SgVector3), vertices->data(), GL_STATIC_DRAW);
        }
    } else {
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, CNOID_GL_FLOAT, 0, vertices->data());
    }

    if(normals){
        if(USE_VBO){
            if(cache->normalBufferName() == GL_INVALID_VALUE){
                glGenBuffers(1, &cache->normalBufferName());
                glBindBuffer(GL_ARRAY_BUFFER, cache->normalBufferName());
                glBufferData(GL_ARRAY_BUFFER, normals->size() * sizeof(SgVector3), normals->data(), GL_STATIC_DRAW);
            }
        } else {
            glEnableClientState(GL_NORMAL_ARRAY);
            glNormalPointer(CNOID_GL_FLOAT, 0, normals->data());
        }
    }
    if(USE_VBO){
        if(!triangles){
            cache->size = vertices->size();

        } else if(cache->indexBufferName() == GL_INVALID_VALUE){
            glGenBuffers(1, &cache->indexBufferName());
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cache->indexBufferName());
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, triangles->size(), &triangles->front(), GL_STATIC_DRAW);
            cache->size = triangles->size();
        }
            
    } else {
        if(USE_INDEXING){
            glDrawElements(GL_TRIANGLES, triangles->size(), GL_UNSIGNED_INT, &triangles->front());
        } else {
            glDrawArrays(GL_TRIANGLES, 0, vertices->size());
        }
    }

    if(doNormalVisualization && !isPicking){
        enableLighting(false);
        if(!USE_INDEXING){
            vector<SgVector3> lines;
            for(size_t i=0; i < vertices->size(); ++i){
                const SgVector3& v = (*vertices)[i];
                lines.push_back(v);
                lines.push_back(v + (*normals)[i] * normalLength);
            }
            glDisableClientState(GL_NORMAL_ARRAY);
            glVertexPointer(3, CNOID_GL_FLOAT, 0, lines.front().data());
            setColor(Vector4f(0.0f, 1.0f, 0.0f, 1.0f));
            glDrawArrays(GL_LINES, 0, lines.size());
        }
        enableLighting(true);
    }
}


void GLSceneRenderer::visitPointSet(SgPointSet* pointSet)
{
    impl->visitPointSet(pointSet);
}


void GLSceneRendererImpl::visitPointSet(SgPointSet* pointSet)
{
    const SgVertexArray* points = pointSet->vertices();
    const SgNormalArray* normals = pointSet->normals();
    
    if(points && !points->empty()){
        setColor(defaultColor);
        if(!normals){
            enableLighting(false);
        } else {
            enableCullFace(false);
            setLightModelTwoSide(true);
            renderMaterial(defaultMaterial.get());
        }
        pushPickName(pointSet);
        glBegin(GL_POINTS);
        const int numPoints = points->size();
        for(int i=0; i < numPoints; ++i){
            if(normals){
                const SgVector3& n = normals->at(i);
                glNormal3(n.x(), n.y(), n.z());
            }
            const SgVector3& v = points->at(i);
            glVertex3(v.x(), v.y(), v.z());
        }
        glEnd();
        popPickName();
        if(!normals){
            enableLighting(true);
        }
    }
}


void GLSceneRenderer::visitLineSet(SgLineSet* lineSet)
{

}


void GLSceneRenderer::visitPreprocessed(SgPreprocessed* preprocessed)
{

}


void GLSceneRenderer::visitLight(SgLight* light)
{

}


Mapping* GLSceneRenderer::property()
{
    return impl->property.get();
}


SignalProxy<boost::signal<void()> > GLSceneRenderer::sigPropertyUpdated()
{
    return impl->sigPropertyUpdated;
}


void GLSceneRenderer::notifyPropertyUpdate()
{
    impl->sigPropertyUpdated();
}


void GLSceneRendererImpl::clearGLState()
{
    stateFlag.reset();
    
    //! \todo get the current state from a GL context
	diffuseColor << 0.0f, 0.0f, 0.0f, 0.0f;
	ambientColor << 0.0f, 0.0f, 0.0f, 0.0f;
	emissionColor << 0.0f, 0.0f, 0.0f, 0.0f;
	specularColor << 0.0f, 0.0f, 0.0f, 0.0f;
    shininess = 0.0f;
    isCullFaceEnabled = true;
    isCCW = false;
    isLightingEnabled = true;
    isLightModelTwoSide = false;
    isBlendEnabled = true;
    isDepthMaskEnabled = true;

    pointSize = defaultPointSize;
    lineWidth = defaultLineWidth;
}


void GLSceneRendererImpl::setColor(const Vector4f& color)
{
    if(!isPicking){
        if(!stateFlag[CURRENT_COLOR] || color != currentColor){
            glColor4f(color[0], color[1], color[2], color[3]);
            currentColor = color;
            stateFlag.set(CURRENT_COLOR);
        }
    }
}


void GLSceneRenderer::setColor(const Vector4f& color)
{
    impl->setColor(color);
}


void GLSceneRendererImpl::setDiffuseColor(const Vector4f& color)
{
    if(!stateFlag[DIFFUSE_COLOR] || diffuseColor != color){
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color.data());
        diffuseColor = color;
        stateFlag.set(DIFFUSE_COLOR);
    }
}


void GLSceneRenderer::setDiffuseColor(const Vector4f& color)
{
    impl->setDiffuseColor(color);
}


void GLSceneRendererImpl::setAmbientColor(const Vector4f& color)
{
    if(!stateFlag[AMBIENT_COLOR] || ambientColor != color){
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color.data());
        ambientColor = color;
        stateFlag.set(AMBIENT_COLOR);
    }
}


void GLSceneRenderer::setAmbientColor(const Vector4f& color)
{
    impl->setAmbientColor(color);
}


void GLSceneRendererImpl::setEmissionColor(const Vector4f& color)
{
    if(!stateFlag[EMISSION_COLOR] || emissionColor != color){
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color.data());
        emissionColor = color;
        stateFlag.set(EMISSION_COLOR);
    }
}


void GLSceneRenderer::setEmissionColor(const Vector4f& color)
{
    impl->setEmissionColor(color);
}


void GLSceneRendererImpl::setSpecularColor(const Vector4f& color)
{
    if(!stateFlag[SPECULAR_COLOR] || specularColor != color){
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color.data());
        specularColor = color;
        stateFlag.set(SPECULAR_COLOR);
    }
}


void GLSceneRenderer::setSpecularColor(const Vector4f& color)
{
    impl->setSpecularColor(color);
}


void GLSceneRendererImpl::setShininess(float s)
{
    if(!stateFlag[SHININESS] || shininess != s){
        glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, s);
        shininess = s;
        stateFlag.set(SHININESS);
    }
}


void GLSceneRenderer::setShininess(float s)
{
    impl->setShininess(s);
}


void GLSceneRendererImpl::enableCullFace(bool on)
{
    if(!stateFlag[CULL_FACE] || isCullFaceEnabled != on){
        if(on){
            glEnable(GL_CULL_FACE);
        } else {
            glDisable(GL_CULL_FACE);
        }
        isCullFaceEnabled = on;
        stateFlag.set(CULL_FACE);
    }
}


void GLSceneRenderer::enableCullFace(bool on)
{
    impl->enableCullFace(on);
}


void GLSceneRendererImpl::setFrontCCW(bool on)
{
    if(!stateFlag[CCW] || isCCW != on){
        if(on){
            glFrontFace(GL_CCW);
        } else {
            glFrontFace(GL_CW);
        }
        isCCW = on;
        stateFlag.set(CCW);
    }
}


void GLSceneRenderer::setFrontCCW(bool on)
{
    impl->setFrontCCW(on);
}


/**
   Lighting should not be enabled in rendering code
   which may be rendered with displaylists.
*/
void GLSceneRendererImpl::enableLighting(bool on)
{
    if(isPicking || !defaultLighting){
        return;
    }
    if(!stateFlag[LIGHTING] || isLightingEnabled != on){
        if(on){
            glEnable(GL_LIGHTING);
        } else {
            glDisable(GL_LIGHTING);
        }
        isLightingEnabled = on;
        stateFlag.set(LIGHTING);
    }
}


void GLSceneRenderer::enableLighting(bool on)
{
    impl->enableLighting(on);
}


void GLSceneRendererImpl::setLightModelTwoSide(bool on)
{
    if(!stateFlag[LIGHT_MODEL_TWO_SIDE] || isLightModelTwoSide != on){
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, on ? GL_TRUE : GL_FALSE);
        isLightModelTwoSide = on;
        stateFlag.set(LIGHT_MODEL_TWO_SIDE);
    }
}


void GLSceneRenderer::setLightModelTwoSide(bool on)
{
    impl->setLightModelTwoSide(on);
}


void GLSceneRendererImpl::enableBlend(bool on)
{
    if(isPicking){
        return;
    }
    if(!stateFlag[BLEND] || isBlendEnabled != on){
        if(on){
            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            enableDepthMask(false);
        } else {
            glDisable(GL_BLEND);
            enableDepthMask(true);
        }
        isBlendEnabled = on;
        stateFlag.set(BLEND);
    }
}


void GLSceneRenderer::enableBlend(bool on)
{
    impl->enableBlend(on);
}


void GLSceneRendererImpl::enableDepthMask(bool on)
{
    if(!stateFlag[DEPTH_MASK] || isDepthMaskEnabled != on){
        glDepthMask(on);
        isDepthMaskEnabled = on;
        stateFlag.set(DEPTH_MASK);
    }
}


void GLSceneRenderer::enableDepthMask(bool on)
{
    impl->enableDepthMask(on);
}


void GLSceneRendererImpl::setPointSize(float size)
{
    if(!stateFlag[POINT_SIZE] || pointSize != size){
        if(isPicking){
            glPointSize(std::max(size, MinLineWidthForPicking));
        } else {
            glPointSize(size);
        }
        pointSize = size;
        stateFlag.set(POINT_SIZE);
    }
}


void GLSceneRenderer::setPointSize(float size)
{
    impl->setPointSize(size);
}


void GLSceneRendererImpl::setLineWidth(float width)
{
    if(!stateFlag[LINE_WIDTH] || lineWidth != width){
        if(isPicking){
            glLineWidth(std::max(width, MinLineWidthForPicking));
        } else {
            glLineWidth(width);
        }
        lineWidth = width;
        stateFlag.set(LINE_WIDTH);
    }
}


void GLSceneRenderer::setLineWidth(float width)
{
    impl->setLineWidth(width);
}



SgObject* SgCustomGLNode::duplicate(const CopyFlags& copyFlags)
{
    return new SgCustomGLNode(*this);
}


void SgCustomGLNode::accept(SgVisitor& visitor)
{
    GLSceneRenderer* renderer = dynamic_cast<GLSceneRenderer*>(&visitor);
    if(renderer){
        renderer->impl->pushPickName(this);
        render(*renderer);
        renderer->impl->popPickName();
    }
}
    

void SgCustomGLNode::render(GLSceneRenderer& renderer)
{
    if(renderingFunction){
        renderingFunction(renderer);
    }
}


void SgCustomGLNode::setRenderingFunction(RenderingFunction f)
{
    renderingFunction = f;
}



const Vector3f& GLSceneRenderer::backgroundColor() const
{
    return impl->bgColor;
}


void GLSceneRenderer::setBackgroundColor(const Vector3f& color)
{
    impl->bgColor = color;
}


SgLight* GLSceneRenderer::headLight()
{
    return impl->headLight.get();
}


void GLSceneRenderer::setHeadLight(SgLightPtr light)
{
    impl->headLight = light;
}


void GLSceneRenderer::setAsDefaultLight(SgLightPtr light)
{
    impl->defaultLights.insert(light);
}


void GLSceneRenderer::unsetDefaultLight(SgLightPtr light)
{
    impl->defaultLights.erase(light);
}


void GLSceneRenderer::enableAdditionalLights(bool on)
{
    impl->additionalLightsEnabled = on;
}


void GLSceneRenderer::setPolygonMode(PolygonMode mode)
{
    impl->polygonMode = mode;
}


void GLSceneRenderer::setDefaultLighting(bool on)
{
    impl->defaultLighting = on;
    requestToClearCache();
}


SgMaterial* GLSceneRenderer::defaultMaterial()
{
    return impl->defaultMaterial.get();
}


void GLSceneRenderer::setDefaultColor(const Vector4f& color)
{
    impl->defaultColor = color;
}


void GLSceneRenderer::setDefaultPointSize(double size)
{
    impl->defaultPointSize = size;
}


void GLSceneRenderer::setDefaultLineWidth(double width)
{
    impl->defaultLineWidth = width;
}


void GLSceneRenderer::showNormalVectors(double length)
{
    impl->doNormalVisualization = (length > 0.0);
    impl->normalLength = length;
}


void GLSceneRenderer::enableUnusedCacheCheck(bool on)
{
    if(!on){
        impl->nextCacheMap->clear();
    }
    impl->doUnusedCacheCheck = on;
}


const Affine3& GLSceneRenderer::lastViewMatrix() const
{
    return impl->lastViewMatrix;
}


const Matrix4& GLSceneRenderer::lastProjectionMatrix() const
{
    return impl->lastProjectionMatrix;
}


void GLSceneRenderer::getViewFrustum
(const SgPerspectiveCamera& camera,
 double& left, double& right, double& bottom, double& top, double& zNear, double& zFar) const
{
    top = impl->zNear * tan(camera.fovy(impl->aspectRatio) / 2.0);
    bottom = -top;
    right = top * impl->aspectRatio;
    left = -right;
    zNear = impl->zNear;
    zFar = impl->zFar;
}


void GLSceneRenderer::getViewVolume
(const SgOrthographicCamera& camera,
 double& out_left, double& out_right, double& out_bottom, double& out_top, double& out_near, double& out_far) const
{
    GLfloat left, right, bottom, top, near_, far_;
    impl->getViewVolume(camera, left, right, bottom, top, near_, far_);
    out_left = left;
    out_right = right;
    out_bottom = bottom;
    out_top = top;
    out_near = near_;
    out_far = far_;
}
