/**
    @file
   @author Shin'ichiro Nakaoka
*/

#include "SceneItem.h"
#include "ItemManager.h"
#include "Archive.h"
#include "PutPropertyFunction.h"
#include <cnoid/FileUtil>
#include <cnoid/PointCloudUtil>
#include <cnoid/Exception>
#include <boost/bind.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;


void SceneItem::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    if(!initialized){
        ext->itemManager().registerClass<SceneItem>(N_("SceneItem"));
        ext->itemManager().addLoader<SceneItem>(
            _("Point Cloud (PCD)"), "PCD-FILE", "pcd",
            bind(&SceneItem::loadPCD, _1, _2, _3), ItemManager::PRIORITY_CONVERSION);
        initialized = true;
    }
}


SceneItem::SceneItem()
    : topNode_(new SgTransform())
{

}


SceneItem::SceneItem(const SceneItem& org)
    : Item(org),
      topNode_(new SgTransform(*org.topNode(), SgObject::SHALLOW_COPY))
{
    
}


SceneItem::~SceneItem()
{
    
}


void SceneItem::setName(const std::string& name)
{
    topNode_->setName(name);
    Item::setName(name);
}


SgNode* SceneItem::scene()
{
    return topNode_.get();
}


ItemPtr SceneItem::doDuplicate() const
{
    return new SceneItem(*this);
}


void SceneItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("file"), getFilename(filesystem::path(lastAccessedFilePath())));
}


bool SceneItem::store(Archive& archive)
{
    if(!lastAccessedFilePath().empty()){
        archive.writeRelocatablePath("file", lastAccessedFilePath());
        archive.write("format", lastAccessedFileFormatId());
    }
    return true;
}


bool SceneItem::restore(const Archive& archive)
{
    std::string filename, formatId;
    if(archive.readRelocatablePath("file", filename) && archive.read("format", formatId)){
        if(load(filename, archive.currentParentItem(), formatId)){
            return true;
        }
    }
    return false;
}


bool SceneItem::loadPCD(const std::string& filename, std::ostream& os)
{
    try {
        SgPointSetPtr pointSet = cnoid::loadPCD(filename);
        if(pointSet){
            SgInvariantGroupPtr group = new SgInvariantGroup();
            group->addChild(pointSet);
            topNode_->addChild(group);
            os << pointSet->vertices()->size() << " points have been loaded.";
            return true;
        }
    } catch (boost::exception& ex) {
        if(std::string const * message = get_error_info<error_info_message>(ex)){
            os << *message;
        }
    }
    return false;
}

     
        
