/**
   @author Shin'ichiro Nakaoka
*/

#include "SceneView.h"
#include "GLSceneRenderer.h"
#include "SceneViewEditable.h"
#include "ExtensionManager.h"
#include "MainWindow.h"
#include "ToolBar.h"
#include "Dialog.h"
#include "SpinBox.h"
#include "ComboBox.h"
#include "Separator.h"
#include "Archive.h"
#include "MessageView.h"
#include "MenuManager.h"
#include "Timer.h"
#include "LazyCaller.h"
#include <cnoid/EigenArchive>
#include <QGLWidget>
#include <QGLPixelBuffer>
#include <QLabel>
#include <QBoxLayout>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QFileDialog>
#include <QDialogButtonBox>
#include <QColorDialog>
#include <QElapsedTimer>
#include <QMessageBox>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <set>
#include <iostream>
#include "gettext.h"

#ifdef _WIN32
#undef near
#undef far
#endif

using namespace std;
using namespace boost;
using namespace Eigen;
using namespace cnoid;

namespace {

#ifdef CNOID_SCENE_GRAPH_USE_FLOAT
    inline void glMultMatrix(const GLfloat* m) { glMultMatrixf(m); }
#else
    inline void glMultMatrix(const GLfloat* m) { glMultMatrixd(m); }
#endif

    const bool TRACE_FUNCTIONS = false;
    const bool SHOW_IMAGE_FOR_PICKING = false;
    
    SceneView* mainSceneView = 0;

    class SetupDialog : public Dialog
    {
    public:
        SpinBox fieldOfViewSpin;
        DoubleSpinBox zNearSpin;
        DoubleSpinBox zFarSpin;
        CheckBox lightingCheck;
        CheckBox headLightCheck;
        DoubleSpinBox headLightIntensitySpin;
        CheckBox worldLightCheck;
        DoubleSpinBox worldLightIntensitySpin;
        DoubleSpinBox worldLightAmbientSpin;
        CheckBox additionalLightsCheck;
        CheckBox floorGridCheck;
        DoubleSpinBox floorGridSpanSpin;
        DoubleSpinBox floorGridIntervalSpin;
        PushButton backgroundColorButton;
        PushButton gridColorButton;
        PushButton defaultColorButton;
        DoubleSpinBox pointSizeSpin;
        DoubleSpinBox lineWidthSpin;
        CheckBox pointModeCheck;
        CheckBox normalVisualizationCheck;
        DoubleSpinBox normalLengthSpin;
        CheckBox coordinateAxesCheck;
        CheckBox fpsCheck;
        PushButton fpsTestButton;
        CheckBox bufferForPickingCheck;

        LazyCaller updateDefaultLightsLater;

        SetupDialog(SceneViewImpl* impl);
        void storeState(Archive& archive);
        void restoreState(const Archive& archive);
    };


    /**
       \note Z axis should always be the upper vertical direciton.
    */
    SgAffine3 normalizedCameraTransform(const SgAffine3& T)
    {
        SgVector3 x, y;
        SgVector3 z = T.linear().col(2).normalized();
        
        if(fabsf(z.dot(SgVector3::UnitZ())) > 0.9f){
            x = T.linear().col(0).normalized();
            y = z.cross(x);
        } else {
            y = T.linear().col(1);
            if(y.dot(SgVector3::UnitZ()) >= 0.0f){
                x = SgVector3::UnitZ().cross(z).normalized();
                y = z.cross(x);
            } else {
                x = z.cross(SgVector3::UnitZ()).normalized();
                y = z.cross(x);
            }
        }
        SgAffine3 N;
        N.linear() << x, y, z;
        N.translation() = T.translation();
        return N;
    }
}


namespace cnoid {

    class SceneViewImpl : public QGLWidget
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SceneViewImpl(SceneView* self);
        ~SceneViewImpl();

        SceneView* self;
        ostream& os;

        GLSceneRenderer renderer;
        QGLPixelBuffer* buffer;
        LazyCaller initializeRenderingCaller;
        Vector4f gridColor;

        SgDirectionalLightPtr worldLight;

        enum OperationMode { VIEW_MODE, EDIT_MODE } operationMode;

        static SceneViewEditable::SceneMode sceneMode(OperationMode mode) {
            return (mode == VIEW_MODE) ? SceneViewEditable::VIEW_MODE : SceneViewEditable::EDIT_MODE;
        }
        
        enum DragMode { NO_DRAGGING, EDITING, VIEW_ROTATION, VIEW_TRANSLATION, VIEW_ZOOM } dragMode;

        set<SceneViewEditable*> editableEntities;

        SceneViewEditable* currentEntity;

        SgFloat orgMouseX;
        SgFloat orgMouseY;
        SgVector3 orgPointedPos;
        SgTransform* cameraTransform;
        SgPerspectiveCamera* persCamera;
        SgOrthographicCamera* orthoCamera;
        SgAffine3 orgCameraPosition;
        float orgOrthoCameraHeight;
        
        SgFloat dragAngleRatio;
        SgFloat viewTranslationRatioX;
        SgFloat viewTranslationRatioY;

        SceneViewEvent latestEvent;

        Vector3 lastClickedPoint;

        SgGroupPtr systemNodeGroup;

        SgCustomGLNodePtr coordinateAxes;
        SgTransform* xAxis;
        SgTransform* yAxis;
        SgTransform* zAxis;

        SgCustomGLNodePtr floorGrid;

        double fps;
        Timer fpsTimer;
        int fpsCounter;
        Timer fpsRenderingTimer;
        bool fpsRendered;
        
        SetupDialog setup;
        QLabel* indicatorLabel;
        ToolButton* editModeToggle;
        ToolButton* walkThroughModeToggle;
        ToolButton* collisionToggle;
        ToolButton* wireframeToggle;
        ComboBox* cameraCombo;
        signals::connection cameraComboConnection;

        Menu* popupMenu;
        MenuManager menuManager;

        ToolBar* createToolBar();

        virtual void initializeGL();
        virtual void resizeGL(int width, int height);
        virtual void paintGL();

        void renderFloorGrid(GLSceneRenderer& renderer);
        void setupCoordinateAxes();
        void renderCoordinateAxes(GLSceneRenderer& renderer);

        void addEntity(SgNode* entity);
        void removeEntity(SgNode* entity);

        void showFPS(bool on);
        void doFPSTest();
        void onFPSUpdateRequest();
        void onFPSRenderingRequest();
        void renderFPS();

        void showBackgroundColorDialog();
        void showGridColorDialog();
        void showDefaultColorDialog();

        void updateCurrentCamera(bool doUpdateCameraCombo);
        void setCurrentCamera(int index, bool doUpdateCameraCombo);
        void setCurrentCameraPath(std::vector<std::string>& simplifiedPathStrings, bool doUpdateCameraCombo);
        void onCamerasChanged();
        
        void onLineWidthChanged(double width);
        void onPointSizeChanged(double width);
        void onPolygonModeChanged();
        void toggleCollisions(bool on);
        void onFieldOfViewChanged();
        void onClippingDepthChanged();
        void onLightingToggled(bool on);
        void updateDefaultLights();
        void onNormalVisualizationChanged();

        void onEditModeToggled(bool on);
        void setOperationMode(OperationMode mode);
        void flipOperationMode();
        void viewAll();

        void onEntityAdded(SgNode* node);
        void onEntityRemoved(SgNode* node);

        void onBufferForPickingToggled(bool on);
        SceneViewEditable* findPointedEntity();
        void updateLastClickedPoint();
        void updateLatestEvent(QKeyEvent* event);
        void updateLatestEvent(int x, int y, int modifiers);
        void updateLatestEvent(QMouseEvent* event);

        virtual void keyPressEvent(QKeyEvent* event);
        virtual void keyReleaseEvent(QKeyEvent* event);
        virtual void mousePressEvent(QMouseEvent* event);
        virtual void mouseDoubleClickEvent(QMouseEvent* event);
        virtual void mouseMoveEvent(QMouseEvent* event);
        void updatePointerPosition();
        virtual void mouseReleaseEvent(QMouseEvent* event);
        virtual void wheelEvent(QWheelEvent* event);
        virtual void focusOutEvent(QFocusEvent* event);

        void startViewChange();
        void startViewRotation();
        void dragViewRotation();
        void startViewTranslation();
        void dragViewTranslation();
        void startViewZoom();
        void dragViewZoom();
        void zoomView(SgFloat ratio);

        void showViewModePopupMenu(const QPoint& globalPos);
        void showEditModePopupMenu(SceneViewEditable* entity, const QPoint& globalPos);
        void onPopupMenuSelectionDone();

        void activateSystemNode(SgNodePtr node, bool on);

        bool saveImage(const std::string& filename);
        void setScreenSize(int width, int height);
        bool storeState(Archive& archive);
        bool restoreState(const Archive& archive);
        void restoreCurrentCamera(const Mapping& cameraData);
    };
    
}


SceneView* SceneView::instance()
{
    assert(mainSceneView);
    return mainSceneView;
}


void SceneView::initialize(ExtensionManager* ext)
{
    if(!mainSceneView){
        mainSceneView = new SceneView();
        ext->addView(mainSceneView);
        ext->addToolBar(mainSceneView->impl->createToolBar());
    }
}


SceneView::SceneView()
{
    setName(N_("Scene"));
    setDefaultLayoutArea(View::RIGHT);
    
    impl = new SceneViewImpl(this);

    QVBoxLayout* vbox = new QVBoxLayout();
    vbox->addWidget(impl);
    setLayout(vbox);
}


SceneViewImpl::SceneViewImpl(SceneView* self)
    : QGLWidget(self),
      self(self),
      setup(this),
      os(MessageView::mainInstance()->cout())
{
    setFocusPolicy(Qt::WheelFocus);

    setAutoBufferSwap(true);
    
    renderer.enableUnusedCacheCheck(true);
    renderer.sigRenderingRequest().connect(bind(&SceneViewImpl::update, this));
    renderer.sigCamerasChanged().connect(bind(&SceneViewImpl::onCamerasChanged, this));

    buffer = 0;

    initializeRenderingCaller.set(bind(&GLSceneRenderer::initializeRendering, &renderer));

    cameraTransform = renderer.defaultCameraTransform();
    persCamera = renderer.defaultPerspectiveCamera();
    orthoCamera = renderer.defaultOrthographicCamera();
    
    gridColor << 0.9f, 0.9f, 0.9f, 1.0f;

    setAutoFillBackground(false);
    setMouseTracking(true);

    operationMode = VIEW_MODE;
    dragMode = NO_DRAGGING;
    currentEntity = 0;

    latestEvent.sceneView_ = self;

    lastClickedPoint.setZero();
    
    indicatorLabel = new QLabel();
    indicatorLabel->setAlignment(Qt::AlignLeft);
    QFont font = indicatorLabel->font();
    font.setFixedPitch(true);
    indicatorLabel->setFont(font);

    SgGroup* root = renderer.sceneRoot();

    worldLight = new SgDirectionalLight();
    worldLight->setDirection(Vector3(0.0, 0.0, -1.0));
    root->addChild(worldLight);
    renderer.setAsDefaultLight(worldLight);

    updateDefaultLights();

    systemNodeGroup = new SgGroup();
    root->addChild(systemNodeGroup);

    setupCoordinateAxes();

    floorGrid = new SgCustomGLNode(bind(&SceneViewImpl::renderFloorGrid, this, _1));
    activateSystemNode(floorGrid, setup.floorGridCheck.isChecked());

    fpsTimer.sigTimeout().connect(bind(&SceneViewImpl::onFPSUpdateRequest, this));
    fpsRenderingTimer.setSingleShot(true);
    fpsRenderingTimer.sigTimeout().connect(bind(&SceneViewImpl::onFPSRenderingRequest, this));

    popupMenu = new Menu(self);
    menuManager.setTopMenu(popupMenu);
    popupMenu->sigTriggered().connect(bind(&SceneViewImpl::onPopupMenuSelectionDone, this));

}


SceneView::~SceneView()
{
    delete impl;
}


SceneViewImpl::~SceneViewImpl()
{
    if(buffer){
        delete buffer;
    }
    delete indicatorLabel;
}


void SceneView::onActivated()
{

}


void SceneView::onDeactivated()
{

}


QWidget* SceneView::indicatorOnInfoBar()
{
    return impl->indicatorLabel;
}


ToolBar* SceneViewImpl::createToolBar()
{
    ToolBar* toolBar = new ToolBar(N_("SceneBar"));

    editModeToggle = toolBar->addToggleButton(QIcon(":/Base/icons/sceneedit.png"), _("Switch to the edit mode"));
    editModeToggle->setChecked(false);
    editModeToggle->sigToggled().connect(bind(&SceneViewImpl::onEditModeToggled, this, _1));

    walkThroughModeToggle = toolBar->addToggleButton(QIcon(":/Base/icons/walkthrough.png"), _("Walk-through mode"));
    walkThroughModeToggle->setChecked(false);

    cameraCombo = new ComboBox();
    cameraCombo->setToolTip(_("Select a camera"));
    cameraCombo->setMinimumContentsLength(1);
    cameraCombo->setSizeAdjustPolicy(QComboBox::AdjustToMinimumContentsLengthWithIcon);
    cameraComboConnection =
        cameraCombo->sigCurrentIndexChanged().connect(
            bind(&SceneViewImpl::setCurrentCamera, this, _1, false));
    toolBar->addWidget(cameraCombo);

    toolBar->addButton(QIcon(":/Base/icons/viewfitting.png"), _("Move the camera to look at the objects"))
        ->sigClicked().connect(bind(&SceneViewImpl::viewAll, this));

    collisionToggle = toolBar->addToggleButton(QIcon(":/Base/icons/collisionlines.png"), _("Toggle the collision visualization"));
    collisionToggle->setChecked(true);
    collisionToggle->sigToggled().connect(bind(&SceneViewImpl::toggleCollisions, this, _1));
    toggleCollisions(true);
    
    wireframeToggle = toolBar->addToggleButton(QIcon(":/Base/icons/wireframe.png"), _("Toggle the wireframe mode"));
    wireframeToggle->sigToggled().connect(bind(&SceneViewImpl::onPolygonModeChanged, this));

    toolBar->addButton(QIcon(":/Base/icons/setup.png"), _("Open the dialog to setup SceneView"))
        ->sigClicked().connect(bind(&SetupDialog::show, &setup));

    return toolBar;
}


void SceneViewImpl::initializeGL()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::initializeGL()" << endl;
    }

    if(!renderer.initializeGL()){
        os << "OpenGL initialization failed." << endl;
        // This view shoulbe be disabled when the glew initialization is failed.
    }
}


void SceneViewImpl::resizeGL(int width, int height)
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::resizeGL()" << endl;
    }

    renderer.setViewport(0, 0, width, height);
}


void SceneView::addEntity(SgNodePtr entity)
{
    impl->addEntity(entity.get());
}


void SceneViewImpl::addEntity(SgNode* entity)
{
    renderer.sceneRoot()->addChild(entity);

    if(SceneViewEditable* editable = dynamic_cast<SceneViewEditable*>(entity)){
        editableEntities.insert(editable);
        editable->onSceneModeChanged(sceneMode(operationMode));
    }
    
    initializeRenderingCaller.request();

    renderer.sceneRoot()->notifyUpdate(SgUpdate::ADDED);
}


void SceneView::removeEntity(SgNodePtr entity)
{
    impl->removeEntity(entity.get());
}


void SceneViewImpl::removeEntity(SgNode* entity)
{
    renderer.sceneRoot()->removeChild(entity);

    if(SceneViewEditable* editable = dynamic_cast<SceneViewEditable*>(entity)){
        if(editable == currentEntity){
            currentEntity = 0;
        }
        editableEntities.erase(editable);
    }
    
    initializeRenderingCaller.request();

    renderer.sceneRoot()->notifyUpdate(SgUpdate::REMOVED);
}


void SceneViewImpl::paintGL()
{
    if(TRACE_FUNCTIONS){
        static int counter = 0;
        os << "SceneViewImpl::paintGL() " << counter++ << endl;
    }

    renderer.render();

    if(fpsTimer.isActive()){
        renderFPS();
    }
}


void SceneViewImpl::renderFloorGrid(GLSceneRenderer& renderer)
{
    if(!setup.floorGridCheck.isChecked()){
        return;
    }

    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);

    renderer.setColor(gridColor);

    float half = setup.floorGridSpanSpin.value() / 2.0f;
    float interval = setup.floorGridIntervalSpin.value();
    float i = 0.0f;
    float x = 0.0f;

    glBegin(GL_LINES);

    do {
        x = i * interval;
        // y-line
        glVertex3f( x, -half, 0.0f);
        glVertex3f( x,  half, 0.0f);
        glVertex3f(-x, -half, 0.0f);
        glVertex3f(-x,  half, 0.0f);
        // x-line
        glVertex3f(-half,  x, 0.0f);
        glVertex3f( half,  x, 0.0f);
        glVertex3f(-half, -x, 0.0f);
        glVertex3f( half, -x, 0.0f);
        ++i;
    } while(x < half);

    glEnd();

    glPopAttrib();
}


void SceneViewImpl::renderCoordinateAxes(GLSceneRenderer& renderer)
{
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    int x, y, w, h;
    renderer.getViewport(x, y, w, h);
    glOrtho((double)(x - 26), (double)(w - 26), (double)(y - 24), (double)(h - 24), -100.0, 100.0);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    SgAffine3 transform(cameraTransform->rotation());
    SgAffine3 inv = transform.inverse();
    glMultMatrix(inv.data());

    renderer.setColor(Vector4f(1.0,0.0,0.0,0.0));
    renderer.visitTransform(xAxis);
    renderer.setColor(Vector4f(0.0,1.0,0.0,0.0));
    renderer.visitTransform(yAxis);
    renderer.setColor(Vector4f(0.4,0.6,1.0,0.0));
    renderer.visitTransform(zAxis);
    
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPopAttrib();
}


void SceneViewImpl::renderFPS()
{
    renderer.setColor(Vector4f(1.0f, 1.0f, 1.0f, 1.0f));
    renderText(20, 20, QString("FPS: %1").arg(fps));
    fpsRendered = true;
    ++fpsCounter;
}


void SceneViewImpl::onFPSUpdateRequest()
{
    double oldFps = fps;
    fps = fpsCounter / 0.5;
    fpsCounter = 0;
    fpsRendered = false;
    if(oldFps > 0.0 || fps > 0.0){
        fpsRenderingTimer.start(100);
    }
}
    

void SceneViewImpl::onFPSRenderingRequest()
{
    if(!fpsRendered){
        update();
        --fpsCounter;
    }
}


void SceneViewImpl::showFPS(bool on)
{
    if(on){
        fpsCounter = 0;
        fpsTimer.start(500);
    } else {
        fpsTimer.stop();
    }
    onFPSUpdateRequest();
}


void SceneViewImpl::doFPSTest()
{
    const SgVector3 p = lastClickedPoint.cast<SgFloat>();
    const SgAffine3 C = cameraTransform->T();

    QElapsedTimer timer;
    timer.start();
    
    for(double i=1.0; i <= 360.0; i+=1.0){
        double a = 3.14159265 * i / 180.0;
        cameraTransform->setTransform(
            normalizedCameraTransform(
                SgTranslation3(p) *
                SgAngleAxis(a, SgVector3::UnitZ()) *
                SgTranslation3(-p) *
                C));
        glDraw();
    }

    double time = timer.elapsed() / 1000.0;
    fps = 360.0 / time;
    fpsCounter = 0;

    QMessageBox::information(&setup, _("FPS Test Result"),
                             QString(_("FPS: %1 frames / %2 [s] = %3")).arg(360).arg(time).arg(fps));

    update();
}


void SceneViewImpl::onEditModeToggled(bool on)
{
    setOperationMode(on ? EDIT_MODE : VIEW_MODE);
}


void SceneViewImpl::setOperationMode(OperationMode mode)
{
    operationMode = mode;

    set<SceneViewEditable*>::iterator p;
    for(p = editableEntities.begin(); p != editableEntities.end(); ++p){
        SceneViewEditable* editable = *p;
        editable->onSceneModeChanged(sceneMode(operationMode));
    }
}


void SceneViewImpl::flipOperationMode()
{
    editModeToggle->setChecked(operationMode == VIEW_MODE);
}


void SceneViewImpl::viewAll()
{
    const SgBoundingBox& bbox = renderer.sceneRoot()->boundingBox();
    if(bbox.empty()){
        return;
    }
    const SgFloat radius = bbox.boundingSphereRadius();

    SgPerspectiveCamera* persCamera = renderer.defaultPerspectiveCamera();
    double left, right, bottom, top, zNear, zFar;
    renderer.getViewFrustum(*persCamera, left, right, bottom, top, zNear, zFar);
    
    SgAffine3& T = cameraTransform->T();
    T.translation() +=
        (bbox.center() - T.translation())
        + T.rotation() * SgVector3(0, 0, 2.0 * radius * zNear / (right - left));

    SgOrthographicCamera* orthoCamera = renderer.defaultOrthographicCamera();
    const double a = renderer.aspectRatio();
    if(a >= 1.0){
        orthoCamera->setHeight(radius * 2.0);
    } else {
        orthoCamera->setHeight(radius * 2.0 / a);
    }

    update();
}


void SceneViewImpl::onBufferForPickingToggled(bool on)
{
    if(!on){
        if(buffer){
            delete buffer;
            buffer = 0;
        }
    }
}


SceneViewEditable* SceneViewImpl::findPointedEntity()
{
    SceneViewEditable* editable = 0;

    if(setup.bufferForPickingCheck.isChecked()){
        const QSize s = size();
        if(buffer && (buffer->size() != s)){
            delete buffer;
            buffer = 0;
        }
        if(!buffer){
            if(QGLPixelBuffer::hasOpenGLPbuffers()){
                QGLFormat f = format();
                f.setDoubleBuffer(false);
                buffer = new QGLPixelBuffer(s, f, this);
                buffer->makeCurrent();
                glEnable(GL_DEPTH_TEST);
            }
        }
    }

    if(buffer){
        buffer->makeCurrent();
    } else {
        QGLWidget::makeCurrent();
    }

    bool picked = renderer.pick(latestEvent.x(), latestEvent.y());

    if(SHOW_IMAGE_FOR_PICKING && !buffer){
        swapBuffers();
    }

    if(!picked){
        latestEvent.nodePath_.clear();
    } else {
        latestEvent.point_ = renderer.pickedPoint();
        latestEvent.nodePath_ = renderer.pickedNodePath();

        vector<SgNode*>& path = latestEvent.nodePath_;
        for(vector<SgNode*>::reverse_iterator p = path.rbegin(); p != path.rend(); ++p){
            SgNode* node = *p;
            editable = dynamic_cast<SceneViewEditable*>(node);
            if(editable){
                if(!editable->isEditable())
                    editable = 0;
                break;
            }
        }
    }

    latestEvent.viewMatrix_ = renderer.lastViewMatrix();
    latestEvent.projectionMatrix_ = renderer.lastProjectionMatrix();
    latestEvent.viewport_ = renderer.viewport();
    
    return editable;
}


void SceneViewImpl::updateLastClickedPoint()
{
    const SgNodePath& path = latestEvent.nodePath();
    if(!path.empty()){
        if(path.back() != floorGrid.get()){
            lastClickedPoint = latestEvent.point();
        }
    }
}

    
void SceneViewImpl::updateLatestEvent(QKeyEvent* event)
{
    latestEvent.modifiers_ = event->modifiers();
    latestEvent.key_ = event->key();
}


void SceneViewImpl::updateLatestEvent(int x, int y, int modifiers)
{
    latestEvent.x_ = x;
    latestEvent.y_ = height() - y - 1;
    latestEvent.modifiers_ = modifiers;
}


void SceneViewImpl::updateLatestEvent(QMouseEvent* event)
{
    updateLatestEvent(event->x(), event->y(), event->modifiers());
    latestEvent.button_ = event->button();
}


void SceneViewImpl::keyPressEvent(QKeyEvent* event)
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::keyPressEvent()" << endl;
    }

    bool handled = false;
    updateLatestEvent(event);
    SceneViewEditable* entity = findPointedEntity();

    switch(event->key()){

    case Qt::Key_Escape:
        flipOperationMode();
        handled = true;
        break;
        
    case Qt::Key_Z:
        if(event->modifiers() & Qt::ControlModifier){
            if(entity){
                if(event->modifiers() & Qt::ShiftModifier){
                    entity->onRedoRequest();
                } else {
                    entity->onUndoRequest();
                }
                handled = true;
            }
        }
        break;

    case Qt::Key_W:
        walkThroughModeToggle->toggle();
        handled = true;
        break;

    case Qt::Key_Space:
    {
        updateLastClickedPoint();
        latestEvent.button_ = Qt::MidButton;
        startViewChange();
        handled = true;
        break;
    }

    default:
        break;
    }

    if(!handled && operationMode == EDIT_MODE){
        if(entity){
            handled = entity->onKeyPressEvent(latestEvent);
        }
    }

    if(!handled){
        event->setAccepted(false);
    }
}


void SceneViewImpl::keyReleaseEvent(QKeyEvent* event)
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::keyReleaseEvent()" << endl;
    }

    bool handled = false;
    
    updateLatestEvent(event);
    
    if(event->key() == Qt::Key_Space){
        dragMode = NO_DRAGGING;
        handled = true;

    } else if(dragMode == VIEW_ZOOM && (event->key() == Qt::Key_Control)){
        dragMode = NO_DRAGGING;
        handled = true;
        
    } else if(operationMode == EDIT_MODE){
        if(currentEntity){
            handled = currentEntity->onKeyReleaseEvent(latestEvent);
        }
    }

    if(!handled){
        event->setAccepted(false);
    }
}


void SceneViewImpl::mousePressEvent(QMouseEvent* event)
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::mousePressEvent()" << endl;
    }
    
    bool handled = false;
    updateLatestEvent(event);

    SceneViewEditable* entity = findPointedEntity();
    updateLastClickedPoint();
    
    OperationMode tmpOperationMode = operationMode;
    if(event->modifiers() & Qt::AltModifier){
        tmpOperationMode = VIEW_MODE;
    }
    
    if(tmpOperationMode == EDIT_MODE){
        if(event->button() == Qt::RightButton){
            showEditModePopupMenu(entity, event->globalPos());
            handled = true;
        } else if(entity){
            if(entity->onButtonPressEvent(latestEvent)){
                handled = true;
                currentEntity = entity;
                dragMode = EDITING;
            }
        }
    }

    if(!handled){
        if(operationMode == VIEW_MODE && event->button() == Qt::RightButton){
            showViewModePopupMenu(event->globalPos());
            handled = true;
        }
    }

    if(!handled){
        startViewChange();
    }
}


void SceneViewImpl::mouseDoubleClickEvent(QMouseEvent* event)
{
    updateLatestEvent(event);

    SceneViewEditable* entity = findPointedEntity();
    bool handled = false;
    
    if(operationMode == EDIT_MODE && entity){
        handled = entity->onDoubleClickEvent(latestEvent);
    }
    if(!handled){
        flipOperationMode();
    }
}


void SceneViewImpl::mouseReleaseEvent(QMouseEvent* event)
{
    updateLatestEvent(event);

    if(operationMode == EDIT_MODE){
        if(currentEntity){
            findPointedEntity(); // update pointed scene path of latestEvent
            currentEntity->onButtonReleaseEvent(latestEvent);
        }
    }
    
    dragMode = NO_DRAGGING;
}


void SceneViewImpl::mouseMoveEvent(QMouseEvent* event)
{
    updateLatestEvent(event);

    switch(dragMode){

    case EDITING:
        currentEntity->onPointerMoveEvent(latestEvent);
        break;

    case VIEW_ROTATION:
        dragViewRotation();
        break;
        
    case VIEW_TRANSLATION:
        dragViewTranslation();
        break;
        
    case VIEW_ZOOM:
        dragViewZoom();
        break;
        
    default:
        updatePointerPosition();
        break;
    }
}


void SceneViewImpl::updatePointerPosition()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::updatePointerPosition()" << endl;
    }
    
    SceneViewEditable* entity = findPointedEntity();
    
    if(operationMode == EDIT_MODE){
        if(currentEntity && (currentEntity != entity)){
            currentEntity->onPointerLeaveEvent(latestEvent);
        }
        if(entity){
            if(entity->onPointerMoveEvent(latestEvent)){
                currentEntity = entity;
                if(!hasFocus()){
                    setFocus(Qt::MouseFocusReason);
                }
            }
        }
    } else {
        static boost::format f(_("Position = (%.3f %.3f %.3f)"));
        const Vector3& p = latestEvent.point();
        self->updateIndicator(str(f % p.x() % p.y() % p.z()));
    }
}


void SceneViewImpl::wheelEvent(QWheelEvent* event)
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::wheelEvent()" << endl;
        os << "delta: " << event->delta() << endl;
    }

    updateLatestEvent(event->x(), event->y(), event->modifiers());
    SceneViewEditable* entity = findPointedEntity();
    updateLastClickedPoint();

    const double s = event->delta() / 8.0 / 15.0;
    latestEvent.wheelSteps_ = s;

    bool handled = false;
    if(operationMode == EDIT_MODE){
        SceneViewEditable* entity = findPointedEntity();
        if(entity && entity->onScrollEvent(latestEvent)){
            currentEntity = entity;
            handled = true;
        }
    }    

    if(renderer.isDefaultCameraCurrent()){
        if(!handled || operationMode == VIEW_MODE){
            if(event->orientation() == Qt::Vertical){
                zoomView(0.25 * s);
            }
        }
    }
}


void SceneViewImpl::focusOutEvent(QFocusEvent* event)
{


}


void SceneViewImpl::startViewChange()
{
    if(renderer.isDefaultCameraCurrent()){

        switch(latestEvent.button()){
                
        case Qt::LeftButton:
            startViewRotation();
            break;
            
        case Qt::MidButton:
            if(latestEvent.modifiers() & Qt::ControlModifier){
                startViewZoom();
            } else {
                startViewTranslation();
            }
            break;
            
        default:
            break;
        }
    }
}
       

void SceneViewImpl::startViewRotation()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::startViewRotation()" << endl;
    }
    
    orgMouseX = latestEvent.x();
    orgMouseY = latestEvent.y();
    orgCameraPosition = cameraTransform->T();

    if(walkThroughModeToggle->isChecked()){
        orgPointedPos = orgCameraPosition.translation();
        dragAngleRatio = 0.01f;
    } else {
        orgPointedPos = lastClickedPoint.cast<SgFloat>();
        dragAngleRatio = 0.01f;
    }
    
    dragMode = VIEW_ROTATION;
}


void SceneViewImpl::dragViewRotation()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::dragViewRotation()" << endl;
    }
    
    const double dx = latestEvent.x() - orgMouseX;
    const double dy = latestEvent.y() - orgMouseY;
    const SgVector3 right = SgCamera::right(orgCameraPosition);

    cameraTransform->setTransform(
        normalizedCameraTransform(
            SgTranslation3(orgPointedPos) *
            SgAngleAxis(-dx * dragAngleRatio, SgVector3::UnitZ()) *
            SgAngleAxis(dy * dragAngleRatio, right) *
            SgTranslation3(-orgPointedPos) *
            orgCameraPosition));
    update();
}


void SceneViewImpl::startViewTranslation()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::startViewTranslation()" << endl;
    }

    const SgAffine3& C = cameraTransform->T();

    if(walkThroughModeToggle->isChecked()){
        viewTranslationRatioX = -0.005;
        viewTranslationRatioY = -0.005;

    } else {
        int x, y, width, height;
        renderer.getViewport(x, y, width, height);
        const double aspect = (double)width / height;
        double r, cw, ch;
        SgCamera* camera = renderer.currentCamera();
        if(camera == persCamera){
            const double fovy = persCamera->fovy(aspect);
            r = (lastClickedPoint.cast<SgFloat>() - C.translation()).dot(SgCamera::direction(C));
            ch = tanf(fovy / 2.0) * 2.0;
            cw = aspect * ch;
        } else if(camera == orthoCamera) {
            r = 1.0;
            ch = orthoCamera->height();
            cw = aspect * ch;
        }
        viewTranslationRatioX = r * cw / width;
        viewTranslationRatioY = r * ch / height;
    }
    
    orgMouseX = latestEvent.x();
    orgMouseY = latestEvent.y();
    orgCameraPosition = C;
    dragMode = VIEW_TRANSLATION;
}


void SceneViewImpl::dragViewTranslation()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::dragViewTranslation()" << endl;
    }
    
    const SgFloat dx = viewTranslationRatioX * (latestEvent.x() - orgMouseX);
    const SgFloat dy = viewTranslationRatioY * (latestEvent.y() - orgMouseY);

    cameraTransform->setTransform(
        normalizedCameraTransform(
            SgTranslation3(-dy * SgCamera::up(orgCameraPosition)) *
            SgTranslation3(-dx * SgCamera::right(orgCameraPosition)) *
            orgCameraPosition));
    update();
}


void SceneViewImpl::startViewZoom()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::startViewZoom()" << endl;
    }

    orgMouseY = latestEvent.y();
    orgCameraPosition = cameraTransform->T();
    orgOrthoCameraHeight = orthoCamera->height();
    dragMode = VIEW_ZOOM;
}


void SceneViewImpl::dragViewZoom()
{
    if(TRACE_FUNCTIONS){
        os << "SceneViewImpl::dragViewZoom()" << endl;
    }

    SgCamera* camera = renderer.currentCamera();

    
    const SgFloat dy = latestEvent.y() - orgMouseY;
    const SgFloat ratio = expf(dy * 0.01);

    if(camera == persCamera){
        const SgAffine3& C = orgCameraPosition;
        const SgVector3 v = SgCamera::direction(C);

        if(walkThroughModeToggle->isChecked()){
            SgFloat speed = 0.02;
            cameraTransform->setTranslation(C.translation() + speed * dy * v);
            
        } else {
            const SgFloat l0 = (lastClickedPoint.cast<SgFloat>() - C.translation()).dot(v);
            cameraTransform->setTranslation(C.translation() + v * (l0 * (-ratio + 1.0)));
        }

    } else if(camera == orthoCamera){
        orthoCamera->setHeight(orgOrthoCameraHeight * ratio);
    }
    update();
}


void SceneViewImpl::zoomView(SgFloat ratio)
{
    SgCamera* camera = renderer.currentCamera();
    if(camera == persCamera){
        const SgAffine3& C = cameraTransform->T();
        const SgVector3 v = SgCamera::direction(C);
        
        if(walkThroughModeToggle->isChecked()){
            if(latestEvent.modifiers() & Qt::ShiftModifier){
                ratio *= 5.0;
            }
            cameraTransform->translation() += ratio * v;

        } else {
            if(latestEvent.modifiers() & Qt::ShiftModifier){
                ratio *= 0.2;
            }
            const SgFloat dz = ratio * (lastClickedPoint.cast<SgFloat>() - C.translation()).dot(v);
            cameraTransform->translation() -= dz * v;
        }

    } else if(camera == orthoCamera){
        orthoCamera->setHeight(orthoCamera->height() * expf(ratio));
    }
    update();
}


void SceneViewImpl::showViewModePopupMenu(const QPoint& globalPos)
{
    menuManager.releaseAllItems();
    
    menuManager.addItem(_("Edit Mode"))
        ->sigTriggered().connect(bind(&SceneViewImpl::flipOperationMode, this));

    popupMenu->popup(globalPos);
}


void SceneViewImpl::showEditModePopupMenu(SceneViewEditable* entity, const QPoint& globalPos)
{
    menuManager.releaseAllItems();
    
    menuManager.setBackwardMode().addItem(_("View Mode"))
        ->sigTriggered().connect(bind(&SceneViewImpl::flipOperationMode, this));
    
    if(entity){
        menuManager.addSeparator();
        menuManager.setPath("/");
        entity->onContextMenuRequest(latestEvent, menuManager);
    }
    
    popupMenu->popup(globalPos);
}


void SceneViewImpl::onPopupMenuSelectionDone()
{
    menuManager.releaseAllItems();
}


void SceneViewImpl::showBackgroundColorDialog()
{
    const Vector3f& c = renderer.backgroundColor();
    QColor newColor =
        QColorDialog::getColor(
            QColor::fromRgbF(c[0], c[1], c[2], 1.0f),
            MainWindow::instance(), _("Background Color"));
    
    if(newColor.isValid()){
        renderer.setBackgroundColor(Vector3f(newColor.redF(), newColor.greenF(), newColor.blueF()));
        update();
    }
}


void SceneViewImpl::showGridColorDialog()
{
    QColor newColor = QColorDialog::getColor(
        QColor::fromRgbF(gridColor[0], gridColor[1], gridColor[2], gridColor[3]),
        MainWindow::instance(), _("Floor Grid Color"));
    
    if(newColor.isValid()){
        gridColor << newColor.redF(), newColor.greenF(), newColor.blueF(), newColor.alphaF();
        update();
    }
}


void SceneViewImpl::showDefaultColorDialog()
{
    QColor c = QColorDialog::getColor(
        QColor::fromRgbF(gridColor[0], gridColor[1], gridColor[2], gridColor[3]),
        MainWindow::instance(), _("Default Color"));
    
    if(c.isValid()){
        Vector4f color(c.redF(), c.greenF(), c.blueF(), c.alphaF());
        renderer.setDefaultColor(color);
        renderer.defaultMaterial()->setDiffuseColor(Vector3f(c.redF(), c.greenF(), c.blueF()));
        renderer.defaultMaterial()->setTransparency(1.0f - c.alphaF());
        renderer.requestToClearCache();
        update();
    }
}


void SceneViewImpl::updateCurrentCamera(bool doUpdateCameraCombo)
{
    if(doUpdateCameraCombo){
        cameraComboConnection.block();
        cameraCombo->setCurrentIndex(renderer.currentCameraIndex());
        cameraComboConnection.unblock();
    }

    latestEvent.cameraPath_ = renderer.cameraPath(renderer.currentCameraIndex());
}
    
        
void SceneViewImpl::setCurrentCamera(int index, bool doUpdateCameraCombo)
{
    renderer.setCurrentCamera(index);
    updateCurrentCamera(doUpdateCameraCombo);
}


void SceneViewImpl::setCurrentCameraPath(std::vector<std::string>& simplifiedPathStrings, bool doUpdateCameraCombo)
{
    renderer.setCurrentCamera(simplifiedPathStrings);
    updateCurrentCamera(doUpdateCameraCombo);
}


void SceneViewImpl::onCamerasChanged()
{
    cameraComboConnection.block();
    vector<string> pathStrings;
    cameraCombo->clear();
    const int n = renderer.numCameras();
    for(int i=0; i < n; ++i){
        renderer.getSimplifiedCameraPathStrings(i, pathStrings);
        string label;
        if(pathStrings.empty()){
            label = str(boost::format(_("Camera %1%")) % i);
        } else if(pathStrings.size() == 1){
            label = pathStrings.front();
        } else {
            label = pathStrings.front() + " - " + pathStrings.back();
        }
        cameraCombo->addItem(label.c_str());
    }
    cameraCombo->setCurrentIndex(renderer.currentCameraIndex());
    cameraComboConnection.unblock();

    updateCurrentCamera(false);
}


void SceneViewImpl::onLineWidthChanged(double width)
{
    renderer.setDefaultLineWidth(width);
    renderer.requestToClearCache();
    update();
}


void SceneViewImpl::onPointSizeChanged(double size)
{
    renderer.setDefaultPointSize(size);
    renderer.requestToClearCache();
    update();
}


void SceneViewImpl::onPolygonModeChanged()
{
    if(wireframeToggle->isChecked()){
        if(setup.pointModeCheck.isChecked()){
            renderer.setPolygonMode(GLSceneRenderer::POINT_MODE);
        } else {
            renderer.setPolygonMode(GLSceneRenderer::LINE_MODE);
        }
    } else {
        renderer.setPolygonMode(GLSceneRenderer::FILL_MODE);
    }
    update();
}


void SceneViewImpl::toggleCollisions(bool on)
{
    renderer.property()->write("collision", on);
    renderer.notifyPropertyUpdate();
    update();
}


void SceneViewImpl::onFieldOfViewChanged()
{
    persCamera->setFieldOfView(PI * setup.fieldOfViewSpin.value() / 180.0);
    update();
}


void SceneViewImpl::onClippingDepthChanged()
{
    renderer.setDefaultClippingDepth(setup.zNearSpin.value(), setup.zFarSpin.value());
    update();
}


void SceneViewImpl::onLightingToggled(bool on)
{
    renderer.setDefaultLighting(on);
    update();
}


void SceneViewImpl::updateDefaultLights()
{
    SgLight* headLight = renderer.headLight();
    headLight->on(setup.headLightCheck.isChecked());
    headLight->setIntensity(setup.headLightIntensitySpin.value());

    worldLight->on(setup.worldLightCheck.isChecked());
    worldLight->setIntensity(setup.worldLightIntensitySpin.value());
    worldLight->setAmbientIntensity(setup.worldLightAmbientSpin.value());

    renderer.enableAdditionalLights(setup.additionalLightsCheck.isChecked());

    renderer.sceneRoot()->notifyUpdate();
}


void SceneViewImpl::onNormalVisualizationChanged()
{
    if(setup.normalVisualizationCheck.isChecked()){
        renderer.showNormalVectors(setup.normalLengthSpin.value());
    } else {
        renderer.showNormalVectors(0.0);
    }
    renderer.requestToClearCache();
    update();
}


bool SceneView::saveImage(const std::string& filename)
{
    return impl->saveImage(filename);
}


bool SceneViewImpl::saveImage(const std::string& filename)
{
    return grabFrameBuffer().save(filename.c_str());
}


QImage SceneView::getImage()
{
    return impl->grabFrameBuffer();
}


void SceneView::setScreenSize(int width, int height)
{
    impl->setScreenSize(width, height);
}


void SceneViewImpl::setScreenSize(int width, int height)
{
    QRect r = self->geometry();
    setGeometry((r.width() - width) / 2, (r.height() - height) / 2, width, height);
}


void SceneView::updateIndicator(const std::string& text)
{
    impl->indicatorLabel->setText(text.c_str());
}


bool SceneView::storeState(Archive& archive)
{
    return impl->storeState(archive);
}


bool SceneViewImpl::storeState(Archive& archive)
{
    archive.write("walkthrough", walkThroughModeToggle->isChecked());
    archive.write("showCollisions", collisionToggle->isChecked());
    archive.write("wireframe", wireframeToggle->isChecked());

    setup.storeState(archive);

    Mapping& cameraData = *archive.createMapping("camera");

    vector<string> cameraStrings;
    if(renderer.getSimplifiedCameraPathStrings(renderer.currentCameraIndex(), cameraStrings)){
        if(cameraStrings.size() == 1){
            cameraData.write("current", cameraStrings.front());
        } else {
            Listing& pathNode = *cameraData.createListing("current");
            pathNode.setFlowStyle(true);
            for(size_t i=0; i < cameraStrings.size(); ++i){
                pathNode.append(cameraStrings[i]);
            }
        }
    }
    const SgAffine3& C = cameraTransform->T();
    write(cameraData, "eye", C.translation());
    write(cameraData, "direction", SgCamera::direction(C));
    write(cameraData, "up", SgCamera::up(C));

    cameraData.write("fieldOfView", persCamera->fieldOfView());
    cameraData.write("near", setup.zNearSpin.value());
    cameraData.write("far", setup.zFarSpin.value());
    cameraData.write("orthoHeight", orthoCamera->height());

    write(archive, "backgroundColor", renderer.backgroundColor());
    write(archive, "gridColor", gridColor);
    
    return true;
}


bool SceneView::restoreState(const Archive& archive)
{
    return impl->restoreState(archive);
}


bool SceneViewImpl::restoreState(const Archive& archive)
{
    bool updated = false;
    
    walkThroughModeToggle->setChecked(archive.get("walkthrough", walkThroughModeToggle->isChecked()));
    collisionToggle->setChecked(archive.get("showCollisions", collisionToggle->isChecked()));
    wireframeToggle->setChecked(archive.get("wireframe", wireframeToggle->isChecked()));
    
    setup.restoreState(archive);
    
    const Mapping& cameraData = *archive.findMapping("camera");
    if(cameraData.isValid()){
        SgVector3 eye, direction, up;
        if(read(cameraData, "eye", eye) &&
           read(cameraData, "direction", direction) &&
           read(cameraData, "up", up)){
            cameraTransform->setPosition(SgCamera::positionLookingFor(eye, direction, up));
            updated = true;
        }
        double fov;
        if(cameraData.read("fieldOfView", fov)){
            persCamera->setFieldOfView(fov);
            updated = true;
        }
        double height;
        if(cameraData.read("orthoHeight", height)){
            orthoCamera->setHeight(height);
            updated = true;
        }
        double zNear, zFar;
        renderer.getDefaultClippingDepth(zNear, zFar);
        setup.zNearSpin.setValue(cameraData.get("near", zNear));
        setup.zFarSpin.setValue(cameraData.get("far", zFar));
        
        archive.addPostProcess(bind(&SceneViewImpl::restoreCurrentCamera, this, ref(cameraData)));
    }

    Vector3f bgColor;
    if(read(archive, "backgroundColor", bgColor)){
        renderer.setBackgroundColor(bgColor);
        updated = true;
    }
    if(read(archive, "gridColor", gridColor)){
        updated = true;
    }
    if(updated){
        update();
    }
    return true;
}


void SceneViewImpl::restoreCurrentCamera(const Mapping& cameraData)
{
    // get a path strings of the current camera
    vector<string> cameraStrings;
    const ValueNode& current = *cameraData.find("current");
    if(current.isString()){
        cameraStrings.push_back(current.toString());
    } else if(current.isListing()){
        const Listing& pathNode = *current.toListing();
        for(int i=0; i < pathNode.size(); ++i){
            cameraStrings.push_back(pathNode[i].toString());
        }
    }
    if(!cameraStrings.empty()){
        setCurrentCameraPath(cameraStrings, true);
    }
}


void SceneViewImpl::setupCoordinateAxes()
{
    coordinateAxes = new SgCustomGLNode(bind(&SceneViewImpl::renderCoordinateAxes, this, _1));
    
    float length = 16;
    float width = 2;

    SgMeshPtr cylinder = new SgMesh;
    cylinder->setCylinder(width, length);
    SgMeshPtr cone = new SgMesh;
    cone->setCone(width*2.0,width*4.0);

    SgShapePtr xCylinderShape = new SgShape;
    xCylinderShape->setMesh(cylinder);
    SgShapePtr xConeShape = new SgShape;
    xConeShape->setMesh(cone);
    SgTransform* xtransform = new SgTransform;
    xtransform->setTranslation(SgVector3(0.0f, length/2.0, 0.0f));
    xtransform->addChild(xConeShape);
    xAxis = new SgTransform;
    xAxis->setTranslation(SgVector3(length/2.0f, 0.0, 0.0));
    AngleAxis xangleA(1.571, Vector3(0.0, 0.0, -1.0));
    xAxis->setRotation(xangleA.toRotationMatrix ());
    xAxis->addChild(xCylinderShape);
    xAxis->addChild(xtransform);

    SgShapePtr yCylinderShape = new SgShape;
    yCylinderShape->setMesh(cylinder);
    SgShapePtr yConeShape = new SgShape;
    yConeShape->setMesh(cone);
    SgTransform* ytransform = new SgTransform;
    ytransform->setTranslation(SgVector3(0.0f, length/2.0, 0.0f));
    ytransform->addChild(yConeShape);
    yAxis = new SgTransform;
    yAxis->setTranslation(SgVector3(0.0f, length/2.0f, 0.0f));
    yAxis->addChild(yCylinderShape);
    yAxis->addChild(ytransform);

    SgShapePtr zCylinderShape = new SgShape;
    zCylinderShape->setMesh(cylinder);
    SgShapePtr zConeShape = new SgShape;
    zConeShape->setMesh(cone);
    SgTransform* ztransform = new SgTransform;
    ztransform->setTranslation(SgVector3(0.0f, length/2.0, 0.0f));
    ztransform->addChild(zConeShape);
    zAxis = new SgTransform;
    AngleAxis zangleA(1.571, Vector3(1.0, 0.0, 0.0));
    zAxis->setRotation(zangleA.toRotationMatrix ());
    zAxis->setTranslation(SgVector3(0.0f, 0.0f, length/2.0f));
    zAxis->addChild(zCylinderShape);
    zAxis->addChild(ztransform);

    activateSystemNode(coordinateAxes, setup.coordinateAxesCheck.isChecked());
}


void SceneViewImpl::activateSystemNode(SgNodePtr node, bool on)
{
    if(on){
        systemNodeGroup->addChild(node, true);
    } else {
        systemNodeGroup->removeChild(node, true);
    }
}


SetupDialog::SetupDialog(SceneViewImpl* impl)
{
    setWindowTitle(_("SceneView Setup"));

    QHBoxLayout* hbox;
    QVBoxLayout* vbox = new QVBoxLayout();

    vbox->addLayout(new HSeparatorBox(new QLabel(_("Default Camera"))));
    hbox = new QHBoxLayout();
    hbox->addWidget(new QLabel(_("Field of view")));
    fieldOfViewSpin.setRange(1, 179);
    fieldOfViewSpin.setValue(45);
    fieldOfViewSpin.sigValueChanged().connect(bind(&SceneViewImpl::onFieldOfViewChanged, impl));
    hbox->addWidget(&fieldOfViewSpin);
    hbox->addWidget(new QLabel(_("[deg]")));
    hbox->addStretch();
    vbox->addLayout(hbox);

    hbox = new QHBoxLayout();
    hbox->addWidget(new QLabel(_("Clipping depth")));
    double zNear, zFar;
    impl->renderer.getDefaultClippingDepth(zNear, zFar);
    hbox->addSpacing(8);
    hbox->addWidget(new QLabel(_("Near")));
    zNearSpin.setDecimals(4);
    zNearSpin.setRange(0.0001, 9.9999);
    zNearSpin.setSingleStep(0.0001);
    zNearSpin.setValue(zNear);
    zNearSpin.sigValueChanged().connect(bind(&SceneViewImpl::onClippingDepthChanged, impl));
    hbox->addWidget(&zNearSpin);
    hbox->addWidget(new QLabel(_("Far")));
    zFarSpin.setDecimals(1);
    zFarSpin.setRange(0.1, 9999999.9);
    zFarSpin.setSingleStep(0.1);
    zFarSpin.setValue(zFar);
    zFarSpin.sigValueChanged().connect(bind(&SceneViewImpl::onClippingDepthChanged, impl));
    hbox->addWidget(&zFarSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);

    updateDefaultLightsLater.set(bind(&SceneViewImpl::updateDefaultLights, impl));
    
    vbox->addLayout(new HSeparatorBox(new QLabel(_("Light"))));
    hbox = new QHBoxLayout();
    lightingCheck.setText(_("Lighiting"));
    lightingCheck.setChecked(true);
    lightingCheck.sigToggled().connect(bind(&SceneViewImpl::onLightingToggled, impl, _1));
    hbox->addWidget(&lightingCheck);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    hbox = new QHBoxLayout();
    headLightCheck.setText(_("Head light"));
    headLightCheck.setChecked(true);
    headLightCheck.sigToggled().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&headLightCheck);

    hbox->addWidget(new QLabel(_("Intensity")));
    headLightIntensitySpin.setDecimals(2);
    headLightIntensitySpin.setSingleStep(0.01);    
    headLightIntensitySpin.setRange(0.0, 1.0);
    headLightIntensitySpin.setValue(0.75);
    headLightIntensitySpin.sigValueChanged().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&headLightIntensitySpin);
    hbox->addStretch();
    vbox->addLayout(hbox);

    hbox = new QHBoxLayout();
    worldLightCheck.setText(_("World light"));
    worldLightCheck.setChecked(true);
    worldLightCheck.sigToggled().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&worldLightCheck);

    hbox->addWidget(new QLabel(_("Intensity")));
    worldLightIntensitySpin.setDecimals(2);
    worldLightIntensitySpin.setSingleStep(0.01);    
    worldLightIntensitySpin.setRange(0.0, 1.0);
    worldLightIntensitySpin.setValue(0.5);
    worldLightIntensitySpin.sigValueChanged().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&worldLightIntensitySpin);

    hbox->addWidget(new QLabel(_("Ambient")));
    worldLightAmbientSpin.setDecimals(2);
    worldLightAmbientSpin.setSingleStep(0.01);    
    worldLightAmbientSpin.setRange(0.0, 1.0);
    worldLightAmbientSpin.setValue(0.3);
    worldLightAmbientSpin.sigValueChanged().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&worldLightAmbientSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    hbox = new QHBoxLayout();
    additionalLightsCheck.setText(_("Additional lights"));
    additionalLightsCheck.setChecked(true);
    additionalLightsCheck.sigToggled().connect(bind(updateDefaultLightsLater));
    hbox->addWidget(&additionalLightsCheck);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    vbox->addLayout(new HSeparatorBox(new QLabel(_("Background"))));
    hbox = new QHBoxLayout();
    backgroundColorButton.setText(_("Background color"));
    backgroundColorButton.sigClicked().connect(bind(&SceneViewImpl::showBackgroundColorDialog, impl));
    hbox->addWidget(&backgroundColorButton);
    hbox->addStretch();
    vbox->addLayout(hbox);

    hbox = new QHBoxLayout();
    floorGridCheck.setText(_("Show the floor grid"));
    floorGridCheck.setChecked(true);
    floorGridCheck.sigToggled().connect(
        bind(&SceneViewImpl::activateSystemNode, impl, boost::ref(impl->floorGrid), _1));
    hbox->addWidget(&floorGridCheck);
    
    hbox->addWidget(new QLabel(_("Span")));
    floorGridSpanSpin.setAlignment(Qt::AlignCenter);
    floorGridSpanSpin.setDecimals(1);
    floorGridSpanSpin.setRange(0.0, 99.9);
    floorGridSpanSpin.setSingleStep(0.1);
    floorGridSpanSpin.setValue(10.0);
    floorGridSpanSpin.sigValueChanged().connect(bind(&SceneViewImpl::update, impl));
    hbox->addWidget(&floorGridSpanSpin);
    hbox->addSpacing(8);
    
    hbox->addWidget(new QLabel(_("Interval")));
    floorGridIntervalSpin.setAlignment(Qt::AlignCenter);
    floorGridIntervalSpin.setDecimals(2);
    floorGridIntervalSpin.setRange(0.01, 9.99);
    floorGridIntervalSpin.setSingleStep(0.01);
    floorGridIntervalSpin.setValue(0.5);
    floorGridIntervalSpin.sigValueChanged().connect(bind(&SceneViewImpl::update, impl));
    hbox->addWidget(&floorGridIntervalSpin);

    gridColorButton.setText(_("Color"));
    gridColorButton.sigClicked().connect(bind(&SceneViewImpl::showGridColorDialog, impl));
    hbox->addWidget(&gridColorButton);
    hbox->addStretch();
    vbox->addLayout(hbox);

    vbox->addWidget(new HSeparator());
    hbox = new QHBoxLayout();
    defaultColorButton.setText(_("Default color"));
    defaultColorButton.sigClicked().connect(bind(&SceneViewImpl::showDefaultColorDialog, impl));
    hbox->addWidget(&defaultColorButton);
    
    hbox->addWidget(new QLabel(_("Default line width")));
    lineWidthSpin.setDecimals(1);
    lineWidthSpin.setRange(0.1, 9.9);
    lineWidthSpin.setSingleStep(0.1);
    lineWidthSpin.setValue(1.0);
    lineWidthSpin.sigValueChanged().connect(bind(&SceneViewImpl::onLineWidthChanged, impl, _1));
    hbox->addWidget(&lineWidthSpin);

    hbox->addWidget(new QLabel(_("Default point size")));
    pointSizeSpin.setDecimals(1);
    pointSizeSpin.setRange(0.1, 9.9);
    pointSizeSpin.setSingleStep(0.1);
    pointSizeSpin.setValue(1.0);
    pointSizeSpin.sigValueChanged().connect(bind(&SceneViewImpl::onPointSizeChanged, impl, _1));
    hbox->addWidget(&pointSizeSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    hbox = new QHBoxLayout();
    pointModeCheck.setText(_("Do point rendering in the wireframe mode"));
    pointModeCheck.sigToggled().connect(
        bind(&SceneViewImpl::onPolygonModeChanged, impl));
    hbox->addWidget(&pointModeCheck);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    hbox = new QHBoxLayout();
    normalVisualizationCheck.setText(_("Normal Visualization"));
    normalVisualizationCheck.sigToggled().connect(
        bind(&SceneViewImpl::onNormalVisualizationChanged, impl));
    hbox->addWidget(&normalVisualizationCheck);
    normalLengthSpin.setDecimals(3);
    normalLengthSpin.setRange(0.0, 1000.0);
    normalLengthSpin.setSingleStep(0.001);
    normalLengthSpin.setValue(0.01);
    normalLengthSpin.sigValueChanged().connect(
        bind(&SceneViewImpl::onNormalVisualizationChanged, impl));
    hbox->addWidget(&normalLengthSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    vbox->addWidget(new HSeparator());

    hbox = new QHBoxLayout();
    coordinateAxesCheck.setText(_("Coordinate axes"));
    coordinateAxesCheck.setChecked(true);
    coordinateAxesCheck.sigToggled().connect(
        bind(&SceneViewImpl::activateSystemNode, impl, boost::ref(impl->coordinateAxes), _1));
    hbox->addWidget(&coordinateAxesCheck);
    
    fpsCheck.setText(_("Show FPS"));
    fpsCheck.setChecked(false);
    fpsCheck.sigToggled().connect(bind(&SceneViewImpl::showFPS, impl, _1));
    hbox->addWidget(&fpsCheck);

    fpsTestButton.setText(_("Test"));
    fpsTestButton.sigClicked().connect(bind(&SceneViewImpl::doFPSTest, impl));
    hbox->addWidget(&fpsTestButton);
    hbox->addStretch();
    vbox->addLayout(hbox);

    hbox = new QHBoxLayout();
    bufferForPickingCheck.setText(_("Use an OpenGL pixel buffer for picking"));
    bufferForPickingCheck.setChecked(true);
    bufferForPickingCheck.sigToggled().connect(bind(&SceneViewImpl::onBufferForPickingToggled, impl, _1));
    hbox->addWidget(&bufferForPickingCheck);
    hbox->addStretch();
    vbox->addLayout(hbox);

    vbox->addWidget(new HSeparator());
    QPushButton* okButton = new QPushButton(_("&Ok"));
    okButton->setDefault(true);
    QDialogButtonBox* buttonBox = new QDialogButtonBox(this);
    buttonBox->addButton(okButton, QDialogButtonBox::AcceptRole);
    connect(buttonBox,SIGNAL(accepted()), this, SLOT(accept()));
    vbox->addWidget(buttonBox);
    
    setLayout(vbox);
}


void SetupDialog::storeState(Archive& archive)
{
    archive.write("defaultHeadLight", headLightCheck.isChecked());
    archive.write("defaultHeadLightIntensity", headLightIntensitySpin.value());
    archive.write("worldLight", worldLightCheck.isChecked());
    archive.write("worldLightIntensity", worldLightIntensitySpin.value());
    archive.write("worldLightAmbient", worldLightAmbientSpin.value());
    archive.write("additionalLights", additionalLightsCheck.isChecked());
    archive.write("floorGrid", floorGridCheck.isChecked());
    archive.write("floorGridSpan", floorGridSpanSpin.value());
    archive.write("floorGridInterval", floorGridIntervalSpin.value());
    archive.write("normalVisualization", normalVisualizationCheck.isChecked());
    archive.write("normalLength", normalLengthSpin.value());
    archive.write("coordinateAxes", coordinateAxesCheck.isChecked());
    archive.write("showFPS", fpsCheck.isChecked());
    archive.write("useBufferForPicking", bufferForPickingCheck.isChecked());
}


void SetupDialog::restoreState(const Archive& archive)
{
    headLightCheck.setChecked(archive.get("defaultHeadLight", headLightCheck.isChecked()));
    headLightIntensitySpin.setValue(archive.get("defaultHeadLightIntensity", headLightIntensitySpin.value()));
    worldLightCheck.setChecked(archive.get("worldLight", worldLightCheck.isChecked()));
    worldLightIntensitySpin.setValue(archive.get("worldLightIntensity", worldLightIntensitySpin.value()));
    worldLightAmbientSpin.setValue(archive.get("worldLightAmbient", worldLightAmbientSpin.value()));
    additionalLightsCheck.setChecked(archive.get("additionalLights", additionalLightsCheck.isChecked()));
    floorGridCheck.setChecked(archive.get("floorGrid", floorGridCheck.isChecked()));
    floorGridSpanSpin.setValue(archive.get("floorGridSpan", floorGridSpanSpin.value()));
    floorGridIntervalSpin.setValue(archive.get("floorGridInterval", floorGridIntervalSpin.value()));
    normalVisualizationCheck.setChecked(archive.get("normalVisualization", normalVisualizationCheck.isChecked()));
    normalLengthSpin.setValue(archive.get("normalLength", normalLengthSpin.value()));
    coordinateAxesCheck.setChecked(archive.get("coordinateAxes", coordinateAxesCheck.isChecked()));
    fpsCheck.setChecked(archive.get("showFPS", fpsCheck.isChecked()));
    bufferForPickingCheck.setChecked(archive.get("useBufferForPicking", bufferForPickingCheck.isChecked()));
}
