/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "Sensor.h"

using namespace cnoid;


Sensor::Sensor()
{

}


Sensor::Sensor(const Sensor& org, bool stateOnly)
    : Device(org, stateOnly)
{

}


void Sensor::forEachActualType(boost::function<bool(const std::type_info& type)> func)
{
    if(!func(typeid(Sensor))){
        Device::forEachActualType(func);
    }
}


ForceSensor::ForceSensor()
    : spec(new Spec)
{
    setStateSize(6);
    F_.setZero();
    spec->F_max.setConstant(std::numeric_limits<double>::max());
}


ForceSensor::ForceSensor(const ForceSensor& org, bool stateOnly)
    : Sensor(org, stateOnly)
{
    copyStateFrom(org);

    if(!stateOnly){
        spec.reset(new Spec);
        if(org.spec){
            spec->F_max = org.spec->F_max;
        } else {
            spec->F_max.setConstant(std::numeric_limits<double>::max());
        }
    }
}


void ForceSensor::copyStateFrom(const ForceSensor& other)
{
    F_ = other.F_;
}


bool ForceSensor::copyStateFrom(const DeviceState& other)
{
    if(const ForceSensor* sensor = dynamic_cast<const ForceSensor*>(&other)){
        copyStateFrom(*sensor);
        return true;
    }
    return false;
}


DeviceState* ForceSensor::cloneState()
{
    return new ForceSensor(*this, true);
}


Device* ForceSensor::clone()
{
    return new ForceSensor(*this);
}


void ForceSensor::forEachActualType(boost::function<bool(const std::type_info& type)> func)
{
    if(!func(typeid(ForceSensor))){
        Sensor::forEachActualType(func);
    }
}


void ForceSensor::initializeState()
{
    F_.setZero();
}


void ForceSensor::readState(const double* buf)
{
    F_ = Eigen::Map<const Vector6>(buf);
}


void ForceSensor::writeState(double* out_buf) const
{
    Eigen::Map<Vector6>(out_buf) << F_;
}


RateGyroSensor::RateGyroSensor()
    : spec(new Spec)
{
    setStateSize(3);
    w_.setZero();
    spec->w_max.setConstant(std::numeric_limits<double>::max());
}


RateGyroSensor::RateGyroSensor(const RateGyroSensor& org, bool stateOnly)
    : Sensor(org, stateOnly)
{
    copyStateFrom(org);

    if(!stateOnly){
        spec.reset(new Spec);
        if(org.spec){
            spec->w_max = org.spec->w_max;
        } else {
            spec->w_max.setConstant(std::numeric_limits<double>::max());
        }
    }
}


void RateGyroSensor::copyStateFrom(const RateGyroSensor& other)
{
    w_ = other.w_;
}


bool RateGyroSensor::copyStateFrom(const DeviceState& other)
{
    if(const RateGyroSensor* sensor = dynamic_cast<const RateGyroSensor*>(&other)){
        copyStateFrom(*sensor);
        return true;
    }
    return false;
}


DeviceState* RateGyroSensor::cloneState()
{
    return new RateGyroSensor(*this, true);
}


Device* RateGyroSensor::clone()
{
    return new RateGyroSensor(*this);
}


void RateGyroSensor::forEachActualType(boost::function<bool(const std::type_info& type)> func)
{
    if(!func(typeid(RateGyroSensor))){
        Sensor::forEachActualType(func);
    }
}


void RateGyroSensor::initializeState()
{
    w_.setZero();
}


void RateGyroSensor::readState(const double* buf)
{
    w_ = Eigen::Map<const Vector3>(buf);
}


void RateGyroSensor::writeState(double* out_buf) const
{
    Eigen::Map<Vector3>(out_buf) << w_;
}


AccelSensor::AccelSensor()
    : spec(new Spec)
{
    setStateSize(3);
    dv_.setZero();
    spec->dv_max.setConstant(std::numeric_limits<double>::max());
}


AccelSensor::AccelSensor(const AccelSensor& org, bool stateOnly)
    : Sensor(org, stateOnly)
{
    copyStateFrom(org);

    if(!stateOnly){
        spec.reset(new Spec);
        if(org.spec){
            spec->dv_max = org.spec->dv_max;
        } else {
            spec->dv_max.setConstant(std::numeric_limits<double>::max());
        }
    }
}


void AccelSensor::copyStateFrom(const AccelSensor& other)
{
    dv_ = other.dv_;
}


bool AccelSensor::copyStateFrom(const DeviceState& other)
{
    if(const AccelSensor* sensor = dynamic_cast<const AccelSensor*>(&other)){
        copyStateFrom(*sensor);
        return true;
    }
    return false;
}


DeviceState* AccelSensor::cloneState()
{
    return new AccelSensor(*this, true);
}


Device* AccelSensor::clone()
{
    return new AccelSensor(*this);
}


void AccelSensor::forEachActualType(boost::function<bool(const std::type_info& type)> func)
{
    if(!func(typeid(AccelSensor))){
        Sensor::forEachActualType(func);
    }
}


void AccelSensor::initializeState()
{
    dv_.setZero();
}


void AccelSensor::readState(const double* buf)
{
    dv_ = Eigen::Map<const Vector3>(buf);
}


void AccelSensor::writeState(double* out_buf) const
{
    Eigen::Map<Vector3>(out_buf) << dv_;
}
