/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "AISTSimulatorItem.h"
#include "BodyMotionItem.h"
#include "ControllerItem.h"
#include <cnoid/ItemManager>
#include <cnoid/Archive>
#include <cnoid/EigenArchive>
#include <cnoid/DyWorld>
#include <cnoid/DyBody>
#include <cnoid/ForwardDynamicsCBM>
#include <cnoid/ConstraintForceSolver>
#include <cnoid/FloatingNumberString>
#include <cnoid/EigenUtil>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <iomanip>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

// for Windows
#undef min
#undef max

namespace {

    const bool TRACE_FUNCTIONS = false;
    const bool ENABLE_DEBUG_OUTPUT = false;
    const double DEFAULT_GRAVITY_ACCELERATION = 9.80665;
    enum DynamicsMode { FORWARD_DYNAMICS = 0, HG_DYNAMICS, KINEMATICS, N_DYNAMICS_MODES };
    enum IntegrationMode { EULER_INTEGRATION = 0, RUNGE_KUTTA_INTEGRATION, N_INTEGRATION_MODES };

    class HighGainControllerItem : public ControllerItem
    {
        BodyPtr body;
        MultiValueSeqPtr qseqRef;
        int currentFrame;
        int lastFrame;
        int numJoints;

    public:
        HighGainControllerItem(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem) {
            qseqRef = bodyMotionItem->jointPosSeq();
            setName(str(fmt(_("HighGain Controller with %1%")) % bodyMotionItem->name()));
        }

        virtual bool start(const Target& target) {
            body = target.body();
            currentFrame = 0;
            lastFrame = std::max(0, qseqRef->numFrames() - 1);
            numJoints = std::min(body->numJoints(), qseqRef->numParts());
            if(qseqRef->numFrames() == 0){
                putMessage(_("Reference motion is empty()."));
                return false;
            }
            if(fabs(qseqRef->frameRate() - (1.0 / target.worldTimeStep())) > 1.0e-6){
                putMessage(_("The frame rate of the reference motion is different from the world frame rate."));
                return false;
            }
            control();
            return true;
        }

        virtual double timeStep() const {
            return qseqRef->getTimeStep();
        }
        
        virtual void input() { }

        virtual bool control() {

            if(++currentFrame > lastFrame){
                currentFrame = lastFrame;
                return false;
            }
            return true;
        }
        
        virtual void output() {

            int prevFrame = std::max(currentFrame - 1, 0);
            int nextFrame = std::min(currentFrame + 1, lastFrame);
            
            MultiValueSeq::Frame q0 = qseqRef->frame(prevFrame);
            MultiValueSeq::Frame q1 = qseqRef->frame(currentFrame);
            MultiValueSeq::Frame q2 = qseqRef->frame(nextFrame);

            double dt = qseqRef->getTimeStep();
            double dt2 = dt * dt;

            for(int i=0; i < numJoints; ++i){
                Link* joint = body->joint(i);
                joint->q() = q1[i];
                joint->dq() = (q2[i] - q1[i]) / dt;
                joint->ddq() = (q2[i] - 2.0 * q1[i] + q0[i]) / dt2;
            }
        }
        
        virtual void stop() { }

    };
}


namespace cnoid {
  
    class AISTSimulatorItemImpl
    {
    public:
        AISTSimulatorItem* self;

        World<ConstraintForceSolver> world;
        
        Selection dynamicsMode;
        Selection integrationMode;
        Vector3 gravity;
        double staticFriction;
        double slipFriction;
        //double cullingThresh;
        FloatingNumberString cullingThresh;
        FloatingNumberString errorCriterion;
        int maxNumIterations;
        FloatingNumberString contactCorrectionDepth;
        FloatingNumberString contactCorrectionVelocityRatio;
        double epsilon;
        bool is2Dmode;

        typedef std::map<Body*, int> BodyIndexMap;
        BodyIndexMap bodyIndexMap;

        AISTSimulatorItemImpl(AISTSimulatorItem* self);
        AISTSimulatorItemImpl(AISTSimulatorItem* self, const AISTSimulatorItemImpl& org);
        bool initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies);
        void addBody(const SimulationBodyPtr& simBody);
        void addColdetLinkPair(const BodyPtr& body1, Link* link1, const BodyPtr& body2, Link* link2);
        bool stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies);
        void doPutProperties(PutPropertyFunction& putProperty);
        bool store(Archive& archive);
        bool restore(const Archive& archive);

        // for debug
        ofstream os;
    };
}


void AISTSimulatorItem::initialize(ExtensionManager* ext)
{
    ext->itemManager().registerClass<AISTSimulatorItem>(N_("AISTSimulatorItem"));
    ext->itemManager().addCreationPanel<AISTSimulatorItem>();
}


AISTSimulatorItem::AISTSimulatorItem()
{
    impl = new AISTSimulatorItemImpl(this);
}


AISTSimulatorItemImpl::AISTSimulatorItemImpl(AISTSimulatorItem* self)
    : self(self),
      dynamicsMode(N_DYNAMICS_MODES, CNOID_GETTEXT_DOMAIN_NAME),
      integrationMode(N_INTEGRATION_MODES, CNOID_GETTEXT_DOMAIN_NAME)
{
    dynamicsMode.setSymbol(FORWARD_DYNAMICS,  N_("Forward dynamics"));
    dynamicsMode.setSymbol(HG_DYNAMICS,       N_("High-gain dynamics"));
    dynamicsMode.setSymbol(KINEMATICS,        N_("Kinematics"));

    integrationMode.setSymbol(EULER_INTEGRATION,  N_("Euler"));
    integrationMode.setSymbol(RUNGE_KUTTA_INTEGRATION,  N_("Runge Kutta"));
    integrationMode.select(RUNGE_KUTTA_INTEGRATION);
    
    gravity << 0.0, 0.0, -DEFAULT_GRAVITY_ACCELERATION;
    staticFriction = 1.0;
    slipFriction = 1.0;
    cullingThresh = 0.005;
    errorCriterion = ConstraintForceSolver::defaultGaussSeidelErrorCriterion();
    maxNumIterations = ConstraintForceSolver::defaultGaussSeidelMaxNumIterations();
    contactCorrectionDepth = ConstraintForceSolver::defaultContactCorrectionDepth();
    contactCorrectionVelocityRatio = ConstraintForceSolver::defaultContactCorrectionVelocityRatio();
    
    epsilon = 0.0;
    is2Dmode = false;
}


AISTSimulatorItem::AISTSimulatorItem(const AISTSimulatorItem& org)
    : SimulatorItem(org),
      impl(new AISTSimulatorItemImpl(this, *org.impl))
{

}


AISTSimulatorItemImpl::AISTSimulatorItemImpl(AISTSimulatorItem* self, const AISTSimulatorItemImpl& org)
    : self(self),
      dynamicsMode(org.dynamicsMode),
      integrationMode(org.integrationMode)
{
    gravity = org.gravity;
    staticFriction = org.staticFriction;
    slipFriction = org.slipFriction;
    cullingThresh = org.cullingThresh;
    errorCriterion = org.errorCriterion;
    maxNumIterations = org.maxNumIterations;
    contactCorrectionDepth = org.contactCorrectionDepth;
    contactCorrectionVelocityRatio = org.contactCorrectionVelocityRatio;
    epsilon = org.epsilon;
    is2Dmode = org.is2Dmode;
}


AISTSimulatorItem::~AISTSimulatorItem()
{
    delete impl;
}


ItemPtr AISTSimulatorItem::doDuplicate() const
{
    return new AISTSimulatorItem(*this);
}


SimulationBodyPtr AISTSimulatorItem::createSimulationBody(const Body& orgBody)
{
    return make_shared<SimulationBody>(make_shared<DyBody>(orgBody));
}


ControllerItem* AISTSimulatorItem::createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem)
{
    return new HighGainControllerItem(bodyItem, bodyMotionItem);
}


bool AISTSimulatorItem::initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies)
{
    return impl->initializeSimulation(simBodies);
}


bool AISTSimulatorItemImpl::initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies)
{
    if(ENABLE_DEBUG_OUTPUT){
        static int ntest = 0;
        os.open((string("test-log-") + lexical_cast<string>(ntest++) + ".log").c_str());
        os << setprecision(30);
    }

    if(integrationMode.is(EULER_INTEGRATION)){
        world.setEulerMethod();
    } else if(integrationMode.is(RUNGE_KUTTA_INTEGRATION)){
        world.setRungeKuttaMethod();
    }
    world.setGravityAcceleration(gravity);
    world.enableSensors(true);
    world.setTimeStep(self->worldTimeStep());
    world.setCurrentTime(0.0);

    world.constraintForceSolver.setGaussSeidelErrorCriterion(errorCriterion.value());
    world.constraintForceSolver.setGaussSeidelMaxNumIterations(maxNumIterations);
    world.constraintForceSolver.setContactDepthCorrection(
        contactCorrectionDepth.value(), contactCorrectionVelocityRatio.value());

    world.clearBodies();
    bodyIndexMap.clear();
    for(size_t i=0; i < simBodies.size(); ++i){
        addBody(simBodies[i]);
    }

    world.constraintForceSolver.clearCollisionCheckLinkPairs();

    self->forAllColdetLinkPairs(bind(&AISTSimulatorItemImpl::addColdetLinkPair, this, _1, _2, _3, _4));
    
    if(is2Dmode){
        world.constraintForceSolver.set2Dmode(true);
    }

    world.initialize();

    return true;
}


void AISTSimulatorItemImpl::addBody(const SimulationBodyPtr& simBody)
{
    DyBodyPtr body = static_pointer_cast<DyBody>(simBody->body());

    DyLink* rootLink = body->rootLink();
    rootLink->v().setZero();
    rootLink->dv().setZero();
    rootLink->w().setZero();
    rootLink->dw().setZero();
    rootLink->vo().setZero();
    rootLink->dvo().setZero();

    bool isHighGainMode = dynamicsMode.is(HG_DYNAMICS);
    if(dynamic_cast<HighGainControllerItem*>(simBody->controller())){
        isHighGainMode = true;
    }

    // Using clone internal models may be better to avoid a bug caused by
    // accessing the same object from multi-threads.
    //for(int i=0; i < body->numLinks(); ++i){
    //    body->link(i)->coldetModel->cloneInternalModel();
    //}

    for(int i=0; i < body->numLinks(); ++i){
        Link* link = body->link(i);
        link->u() = 0.0;
        link->dq() = 0.0;
        link->ddq() = 0.0;
    }
    
    body->clearExternalForces();
    body->calcForwardKinematics(true, true);

    if(isHighGainMode){
        ForwardDynamicsCBMPtr cbm = make_shared_aligned<ForwardDynamicsCBM>(body);
        cbm->setHighGainModeForAllJoints();
        bodyIndexMap[body.get()] = world.addBody(body, cbm);
    } else {
        bodyIndexMap[body.get()] = world.addBody(body);
    }
}


void AISTSimulatorItemImpl::addColdetLinkPair(const BodyPtr& body1, Link* link1, const BodyPtr& body2, Link* link2)
{
    int body1Index = bodyIndexMap[body1.get()];
    int body2Index = bodyIndexMap[body2.get()];
    world.constraintForceSolver.addCollisionCheckLinkPair(
        body1Index, static_cast<DyLink*>(link1),
        body2Index, static_cast<DyLink*>(link2),
        staticFriction, slipFriction, cullingThresh.value(), epsilon);
}


bool AISTSimulatorItem::stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies)
{
    return impl->stepSimulation(activeSimBodies);
}


bool AISTSimulatorItemImpl::stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies)
{
    world.constraintForceSolver.clearExternalForces();

    if(dynamicsMode.is(KINEMATICS)){
        for(size_t i=0; i < activeSimBodies.size(); ++i){
            activeSimBodies[i]->body()->calcForwardKinematics();
        }
    } else {
        world.calcNextState();
    }

    return true;
}


void AISTSimulatorItem::finalizeSimulation()
{
    if(ENABLE_DEBUG_OUTPUT){
        impl->os.close();
    }
}


void AISTSimulatorItem::doPutProperties(PutPropertyFunction& putProperty)
{
    SimulatorItem::doPutProperties(putProperty);
    impl->doPutProperties(putProperty);
}


void AISTSimulatorItemImpl::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Dynamics mode"), dynamicsMode,
                bind((bool(Selection::*)(int))&Selection::select, &dynamicsMode, _1));
    putProperty(_("Integration mode"), integrationMode,
                bind((bool(Selection::*)(int))&Selection::select, &integrationMode, _1));
    putProperty(_("Gravity"), str(gravity), bind(toVector3, _1, ref(gravity)));
    putProperty.decimals(3).min(0.0);
    putProperty(_("Static friction"), staticFriction, changeProperty(staticFriction));
    putProperty(_("Slip friction"), slipFriction, changeProperty(slipFriction));
    putProperty(_("Culling thresh"), cullingThresh,
                (bind(&FloatingNumberString::setNonNegativeValue, ref(cullingThresh), _1)));
    putProperty(_("Error criterion"), errorCriterion,
                bind(&FloatingNumberString::setPositiveValue, ref(errorCriterion), _1));
    putProperty.min(1.0)(_("Max iterations"), maxNumIterations, changeProperty(maxNumIterations));
    putProperty(_("CC depth"), contactCorrectionDepth,
                bind(&FloatingNumberString::setNonNegativeValue, ref(contactCorrectionDepth), _1));
    putProperty(_("CC v-ratio"), contactCorrectionVelocityRatio,
                bind(&FloatingNumberString::setPositiveValue, ref(contactCorrectionVelocityRatio), _1));
    putProperty(_("2D mode"), is2Dmode, changeProperty(is2Dmode));
}


bool AISTSimulatorItem::store(Archive& archive)
{
    SimulatorItem::store(archive);
    return impl->store(archive);
}


bool AISTSimulatorItemImpl::store(Archive& archive)
{
    archive.write("dynamicsMode", dynamicsMode.selectedSymbol());
    archive.write("integrationMode", integrationMode.selectedSymbol());
    write(archive, "gravity", gravity);
    archive.write("staticFriction", staticFriction);
    archive.write("slipFriction", slipFriction);
    archive.write("cullingThresh", cullingThresh);
    archive.write("errorCriterion", errorCriterion);
    archive.write("maxNumIterations", maxNumIterations);
    archive.write("contactCorrectionDepth", contactCorrectionDepth);
    archive.write("contactCorrectionVelocityRatio", contactCorrectionVelocityRatio);
    archive.write("2Dmode", is2Dmode);
    return true;
}


bool AISTSimulatorItem::restore(const Archive& archive)
{
    SimulatorItem::restore(archive);
    return impl->restore(archive);
}


bool AISTSimulatorItemImpl::restore(const Archive& archive)
{
    string symbol;
    if(archive.read("dynamicsMode", symbol)){
        dynamicsMode.select(symbol);
    }
    if(archive.read("integrationMode", symbol)){
        integrationMode.select(symbol);
    }
    read(archive, "gravity", gravity);
    archive.read("staticFriction", staticFriction);
    archive.read("slipFriction", slipFriction);
    cullingThresh = archive.get("cullingThresh", cullingThresh.string());
    errorCriterion = archive.get("errorCriterion", errorCriterion.string());
    archive.read("maxNumIterations", maxNumIterations);
    contactCorrectionDepth = archive.get("contactCorrectionDepth", contactCorrectionDepth.string());
    contactCorrectionVelocityRatio = archive.get("contactCorrectionVelocityRatio", contactCorrectionVelocityRatio.string());
    archive.read("2Dmode", is2Dmode);
    return true;
}
