/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "OSGSceneWorld.h"
#include "OSGCollision.h"
#include "WorldItem.h"
#include <map>
#include <iostream>
#include <boost/signal.hpp>
#include <boost/bind.hpp>
#include <osg/Geode>
#include <cnoid/RootItem>
#include <cnoid/ItemTreeView>
#include <cnoid/MessageView>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    const bool TRACE_FUNCTIONS = false;
}


OSGSceneWorld::OSGSceneWorld(WorldItemPtr worldItem)
    : os(MessageView::mainInstance()->cout()),
      worldItem(worldItem)
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorld::OSGSceneWorld()" << endl;
    }
            
    osgCollision = new OSGCollision();
    osgCollision->setColdetPairs(worldItem->coldetLinkPairs());

    osg::Geode* geode = new osg::Geode();
    geode->addDrawable(osgCollision.get());

    addChild(geode);
}


OSGSceneWorld::~OSGSceneWorld()
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorld::~OSGSceneWorld()" << endl;
    }
}


void OSGSceneWorld::onAttachedToScene()
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorld::onAttachedToScene()" << endl;
    }
    
    connectionWithSigCollisionsUpdated = 
        worldItem->sigCollisionsUpdated().connect(
            bind(&OSGSceneWorld::onCollisionsUpdated, this));
}


void OSGSceneWorld::onDetachedFromScene()
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorld::onDetachedFromScene()" << endl;
    }
    
    connectionWithSigCollisionsUpdated.disconnect();
}


void OSGSceneWorld::onCollisionsUpdated()
{
    osgCollision->dirtyDisplayList();
    requestRedraw();
}


OSGSceneWorldManager::OSGSceneWorldManager()
    : os(MessageView::mainInstance()->cout())
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::OSGSceneWorldManager()" << endl;
    }
    
    itemTreeView = ItemTreeView::mainInstance();

    RootItem::mainInstance()->sigItemAdded().connect(
        bind(&OSGSceneWorldManager::onItemAdded, this, _1));
}


OSGSceneWorldManager::~OSGSceneWorldManager()
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::~OSGSceneWorldManager()" << endl;
    }
}


void OSGSceneWorldManager::onItemAdded(Item* item)
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::onItemAdded()" << endl;
    }
    
    if(WorldItem* worldItem = dynamic_cast<WorldItem*>(item)){

        worldItem->sigDetachedFromRoot().connect(
            bind(&OSGSceneWorldManager::onWorldItemDetached, this, worldItem));

        itemTreeView->sigCheckToggled(worldItem).connect(
            bind(&OSGSceneWorldManager::onWorldItemCheckToggled, this, worldItem, _1));
        
        if(itemTreeView->isItemChecked(worldItem)){
            showSceneWorld(worldItem, true);
        }
    }
}


/**
   @todo do disconnecting operation
*/
void OSGSceneWorldManager::onWorldItemDetached(WorldItem* worldItem)
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::onWorldItemDetached()" << endl;
    }
    
    showSceneWorld(worldItem, false);
}


void OSGSceneWorldManager::onWorldItemCheckToggled(WorldItem* worldItem, bool isChecked)
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::onWorldItemShown()" << endl;
    }

    showSceneWorld(worldItem, isChecked);
}


void OSGSceneWorldManager::showSceneWorld(WorldItem* worldItem, bool show)
{
    if(TRACE_FUNCTIONS){
        cout << "OSGSceneWorldManager::showOSGSceneWorld()" << endl;
    }

    SceneWorldMap::iterator p = sceneWorlds.find(worldItem);

    if(!show && p != sceneWorlds.end()){
        p->second->hide();
        sceneWorlds.erase(p);
        
    } else if(show && p == sceneWorlds.end()){
        OSGSceneWorldPtr sceneWorld = new OSGSceneWorld(worldItem);
        sceneWorlds[worldItem] = sceneWorld;
        sceneWorld->show();
    }
}
