/*!
  @file
  @author Shizuko Hattori
*/

#include "BulletSimulatorItem.h"
#include <cnoid/ItemManager>
#include <cnoid/Archive>
#include <cnoid/EigenUtil>
#include <cnoid/EigenArchive>
#include <cnoid/Body>
#include <cnoid/BodyItem>
#include <cnoid/Link>
#include <cnoid/FloatingNumberString>
#include <cnoid/BasicSensorSimulationHelper>
#include <boost/bind.hpp>
#include <boost/make_shared.hpp>
#include "gettext.h"
#include <btBulletDynamicsCommon.h>
#include <BulletCollision/Gimpact/btGImpactShape.h>
#include <BulletCollision/Gimpact/btGImpactCollisionAlgorithm.h>

using namespace std;
using namespace boost;
using namespace cnoid;

#define DEBUG_OUT 0

namespace {

	const btScalar DEFAULT_GRAVITY_ACCELERATION = 9.80665;
	const btScalar DEFAULT_ERP = 0.2;
	const int  DEFAULT_NUMITERATION = 10;
	const btScalar DEFAULT_RESTITUTION = 0.0;
	const btScalar DEFAULT_FRICTION = 0.7;
	const btScalar DEFAULT_ERP2 = 0.0;
	const btScalar DEFAULT_SPLITIMPULSEPENETRATIONTHRESHOLD = -0.0001;

	class BulletBody;

	class BulletLink : public Referenced
    {
    public:
		BulletLink* parent;
        Link* link;
		btTransform shift;
		btTransform invShift;
		btScalar*	vertices;
		int*	triangles;
		btTriangleIndexVertexArray* pMeshData;
		btCollisionShape* pColShape;
		btDefaultMotionState* motionState;
		btRigidBody* body;
		btTypedConstraint* joint;
		btDynamicsWorld* dynamicsWorld;
		double q_offset;
        BulletSimulatorItemImpl* simImpl;
		
		BulletLink(BulletSimulatorItemImpl* simImpl, BulletBody* bulletBody, BulletLink* parent,
                const Vector3& parentOrigin, Link* link, short group, bool isSelfCollisionDetectionEnabled);
        ~BulletLink();
		void createLinkBody(BulletSimulatorItemImpl* simImpl, BulletLink* parent, const Vector3& origin,
			short group, bool isSelfCollisionDetectionEnabled);
		void createGeometry();
		void getKinematicStateFromBullet();
		void setKinematicStateToBullet();
		void setTorqueToBullet();
	};
	typedef boost::intrusive_ptr<BulletLink> BulletLinkPtr;

    class BulletBody : public SimulationBody
    {
    public:
		vector<BulletLinkPtr> bulletLinks;
		vector<btTypedConstraint*> extraJoints;
		btDynamicsWorld* dynamicsWorld;
        vector<btJointFeedback*> forceSensorFeedbacks;
        BasicSensorSimulationHelper sensorHelper;

        BodyPtr body_;
        BulletBody(const Body& orgBody);
		~BulletBody();
		void createBody(BulletSimulatorItemImpl* simImpl, short group);
		void getKinematicStateFromBullet();
		void setKinematicStateToBullet();
		void setTorqueToBullet();
		void setExtraJoints();
        void updateForceSensors();
    };
}


namespace cnoid {
  
    class BulletSimulatorItemImpl
    {
    public:
        BulletSimulatorItem* self;
		
		btDefaultCollisionConfiguration* collisionConfiguration;
		btCollisionDispatcher* dispatcher;
		btBroadphaseInterface* broadphase;
		btConstraintSolver* solver;
		btDynamicsWorld* dynamicsWorld;

		Vector3 gravity;
		double timeStep;
		double erp;
		int numIterations;
		double restitution;
		double friction;
		double erp2;
		double splitImpulsePenetrationThreshold;

        BulletSimulatorItemImpl(BulletSimulatorItem* self);
        BulletSimulatorItemImpl(BulletSimulatorItem* self, const BulletSimulatorItemImpl& org);
        ~BulletSimulatorItemImpl();
        bool initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies);
        bool stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies);
        void doPutProperties(PutPropertyFunction& putProperty);
        void store(Archive& archive);
        void restore(const Archive& archive);
		void initialize();
		void clear();
		void addBody(BulletBody* bulletBody, short group);
		void setSolverParameter();
    };
}


static bool CustomMaterialCombinerCallback(btManifoldPoint& cp,	const btCollisionObjectWrapper* colObj0,int partId0,int index0,const btCollisionObjectWrapper* colObj1,int partId1,int index1)
{
    if(cp.m_distance1 < -0.001)
        return true;
    BulletLink* bulletLink0 = (BulletLink*)colObj0->getCollisionObject()->getUserPointer();
    Link* link0 = 0;
    if(bulletLink0)
        link0 = bulletLink0->link;
    BulletLink* bulletLink1 = (BulletLink*)colObj1->getCollisionObject()->getUserPointer();
    Link* link1 = 0;
    if(bulletLink1)
        link1 = bulletLink1->link;

    Link* crawlerlink;
    double sign = 1;
    double friction;
    if(link0 && (link0->jointType() == Link::CRAWLER_JOINT)){
        crawlerlink = link0;
        friction = bulletLink0->simImpl->friction;
    } else if(link1 && (link1->jointType() == Link::CRAWLER_JOINT)){
        crawlerlink = link1;
        sign = -1;
        friction = bulletLink1->simImpl->friction;
    }else
        return true;

    Vector3 axis = crawlerlink->R() * crawlerlink->a();
    btVector3 normal = cp.m_normalWorldOnB;
    Vector3 n(normal.x(), normal.y(), normal.z());  
    Vector3 dir = axis.cross(n);

    if(dir.norm()<1e-5)
        return true;
	
    dir *= sign;
	dir.normalize();
    Vector3 frictionDir1 = friction * dir;
    Vector3 frictionDir2 = 0.1 * axis;
	
	cp.m_lateralFrictionInitialized = true;
	cp.m_lateralFrictionDir1.setValue(frictionDir1.x(), frictionDir1.y(), frictionDir1.z());
	cp.m_lateralFrictionDir2.setValue(frictionDir2.x(), frictionDir2.y(), frictionDir2.z());
	cp.m_contactMotion1 = crawlerlink->u();

	return true;
}


///////////////////////////////////Link////////////////////////////////////////////////////
BulletLink::BulletLink(BulletSimulatorItemImpl* _simImpl, BulletBody* bulletBody, BulletLink* parent,
                       const Vector3& parentOrigin, Link* link, short group, bool isSelfCollisionDetectionEnabled)
{
    bulletBody->bulletLinks.push_back(this);
	dynamicsWorld = _simImpl->dynamicsWorld;
    this->simImpl = _simImpl;

    this->link = link;
	vertices = 0;
	triangles = 0;
	pMeshData = 0;
	pColShape = 0;
	motionState = 0;
	body = 0;
	joint=0;
	q_offset = 0;
    
    Vector3 o = parentOrigin + link->b();
    
	createLinkBody(simImpl, parent, o, group, isSelfCollisionDetectionEnabled);
	body->setActivationState(DISABLE_DEACTIVATION);

    for(Link* child = link->child(); child; child = child->sibling()){
        new BulletLink(simImpl, bulletBody, this, o, child, group, isSelfCollisionDetectionEnabled);
    }
}

void BulletLink::createGeometry()
{
    const ColdetModelPtr& cmodel = link->coldetModel();
    int numTriangles = cmodel->getNumTriangles();
    if(numTriangles <= 0){
        return;
    }

    int numVertices = cmodel->getNumVertices();
    vertices = new btScalar[numVertices*3];
	
    for(int i=0, j=0; i < numVertices; ++i){
        float x, y, z;
        cmodel->getVertex(i, x, y, z);
		
		btVector3 v((btScalar)x, (btScalar)y, (btScalar)z );
		btVector3 v0 = invShift * v;
		vertices[j++] = v0[0];
        vertices[j++] = v0[1];
        vertices[j++] = v0[2];
    }
        
    triangles = new int[numTriangles*3];
    for(int i=0; i < numTriangles; ++i){
        cmodel->getTriangle(i, triangles[i*3],  triangles[i*3+1],  triangles[i*3+2]);
    }

	pMeshData = new btTriangleIndexVertexArray( numTriangles, triangles, sizeof(int)*3, numVertices, vertices, sizeof(btScalar)*3);  

	btGImpactMeshShape * trimesh = new btGImpactMeshShape(pMeshData);
	trimesh->setLocalScaling(btVector3(1.f,1.f,1.f));
	trimesh->setMargin(0.0f);
	trimesh->updateBound();
	pColShape = trimesh;
}

void BulletLink::createLinkBody(BulletSimulatorItemImpl* simImpl, BulletLink* parent_, const Vector3& origin, 
	short group, bool isSelfCollisionDetectionEnabled)
{
	parent = parent_;
	const Vector3& c = link->c();
	btTransform transform;
	const Matrix3& R0 = link->Rs();
	btMatrix3x3 btR(R0(0,0), R0(0,1), R0(0,2),
					R0(1,0), R0(1,1), R0(1,2),
					R0(2,0), R0(2,1), R0(2,2));
	transform.setBasis(btR);
	transform.setOrigin(btVector3(origin(0), origin(1), origin(2)));

	btScalar mass(link->mass());
	if(link->isRoot() && link->isFixedJoint()){
		mass = 0.0;
    }
	const Matrix3& I = link->I();

	btVector3 localInertia(0,0,0);
	shift.setIdentity();
	shift.setOrigin(btVector3(c.x(), c.y(), c.z()));
	if (mass != 0.0){
		if(!I.isDiagonal()){
			btMatrix3x3 tensor(I(0,0), I(0,1), I(0,2),
							   I(1,0), I(1,1), I(1,2),
							   I(2,0), I(2,1), I(2,2));
			tensor.diagonalize(shift.getBasis(), btScalar(0.00001), 20);
			localInertia.setValue(tensor[0][0], tensor[1][1], tensor[2][2]);

		}else{
			localInertia.setValue(I(0,0), I(1,1), I(2,2));
		}
	}
	invShift = shift.inverse();

#if DEBUG_OUT
	std::cout <<  shift.getBasis()[0][0] << " " << shift.getBasis()[0][1]<< " " << shift.getBasis()[0][2] << " " << shift.getOrigin()[0] << std::endl;
	std::cout <<  shift.getBasis()[1][0] << " " << shift.getBasis()[1][1]<< " " << shift.getBasis()[1][2] << " " << shift.getOrigin()[1] <<std::endl;
	std::cout <<  shift.getBasis()[2][0] << " " << shift.getBasis()[2][1]<< " " << shift.getBasis()[2][2] << " " << shift.getOrigin()[2] <<std::endl;
#endif

	createGeometry();

	motionState = new btDefaultMotionState(transform * shift);
	btRigidBody::btRigidBodyConstructionInfo rbInfo(mass,motionState,pColShape,localInertia);
	body = new btRigidBody(rbInfo);
	//body->setDamping(0.1,0.1);
	body->setRestitution(simImpl->restitution);
	body->setFriction(simImpl->friction);
	body->setContactProcessingThreshold(BT_LARGE_FLOAT);
	body->forceActivationState(DISABLE_DEACTIVATION);
	
	short mask = 0xFFFF;
	if(!isSelfCollisionDetectionEnabled)
		mask ^= group;
	simImpl->dynamicsWorld->addRigidBody(body, group, mask); 

	const Vector3& a = link->a();
	const Vector3& b = link->b();
	const Matrix3& Rs = link->Rs();
	const Vector3& d = link->d();

	switch(link->jointType()){
        
    case Link::ROTATIONAL_JOINT: {
        const Matrix3& parentRs = parent->link->Rs();
        const Vector3 axisA0 = parentRs.transpose() * Rs * a;
		btVector3 axisA(axisA0(0), axisA0(1), axisA0(2));
		btVector3 axisB(a(0), a(1), a(2)); 
		btVector3 pivotA(b(0), b(1), b(2));
		btVector3 pivotB(0,0,0);
		joint = new btHingeConstraint(*body, *(parent->body), invShift*pivotB, parent->invShift*pivotA,
                                      invShift.getBasis()*axisB, parent->invShift.getBasis()*axisA );

	    if(link->q_upper() < numeric_limits<double>::max() && link->q_lower() > -numeric_limits<double>::max()){
		    ((btHingeConstraint*)joint)->setLimit(link->q_lower(), link->q_upper());
		} else {
            //Lowerlimit ＞ Upperlimit → axis is free
			((btHingeConstraint*)joint)->setLimit(1.0, -1.0); 
        }
		simImpl->dynamicsWorld->addConstraint(joint, true);
		q_offset = ((btHingeConstraint*)joint)->getHingeAngle();

#if DEBUG_OUT
		std::cout << "offset= " << q_offset << std::endl;
#endif
        break;
    }
        
    case Link::SLIDE_JOINT: {
		Vector3 u(0,0,1);
		Vector3 ty = d.cross(u);
		btMatrix3x3 btR;
		Matrix3 R0;
		if(ty.norm() == 0){
			btR.setIdentity();
			R0.setIdentity();
		} else {
			ty.normalized();
			Vector3 tx = ty.cross(d).normalized();
			R0.col(0) = tx;
			R0.col(1) = ty;
			R0.col(2) = d;
			btR.setValue(R0(0,0), R0(0,1), R0(0,2),
						 R0(1,0), R0(1,1), R0(1,2),
						 R0(2,0), R0(2,1), R0(2,2));
		}
		btTransform frameA,frameB;
		frameB.setBasis(btR);
		frameB.setOrigin(btVector3(0,0,0));

		Matrix3& parentRs = parent->link->Rs();
		const Matrix3 R1 = parentRs.transpose() * Rs * R0;
		btMatrix3x3 btR1(R1(0,0), R1(0,1), R1(0,2),
			   			 R1(1,0), R1(1,1), R1(1,2),
						 R1(2,0), R1(2,1), R1(2,2));
		frameA.setBasis(btR1);
		frameA.setOrigin(btVector3(b(0), b(1), b(2)));
		joint = new btGeneric6DofConstraint(*(parent->body), *body, parent->invShift*frameA, invShift*frameB, false);

		if(link->q_upper() < numeric_limits<double>::max() &&
           link->q_lower() > -numeric_limits<double>::max() &&
           link->q_lower() < link->q_upper()){
			((btGeneric6DofConstraint*)joint)->setLinearLowerLimit(btVector3(0.0, 0.0, link->q_lower()));
			((btGeneric6DofConstraint*)joint)->setLinearUpperLimit(btVector3(0.0, 0.0, link->q_upper()));
			((btGeneric6DofConstraint*)joint)->setAngularLowerLimit(btVector3(0.0, 0.0, 0.0));
			((btGeneric6DofConstraint*)joint)->setAngularUpperLimit(btVector3(0.0, 0.0, 0.0));
		} else {
            //Lowerlimit ＞ Upperlimit → axis is free
			((btGeneric6DofConstraint*)joint)->setLinearLowerLimit(btVector3(0.0, 0.0, 1.0)); 
			((btGeneric6DofConstraint*)joint)->setLinearUpperLimit(btVector3(0.0, 0.0, -1.0));
			((btGeneric6DofConstraint*)joint)->setAngularLowerLimit(btVector3(0.0, 0.0, 0.0));
			((btGeneric6DofConstraint*)joint)->setAngularUpperLimit(btVector3(0.0, 0.0, 0.0));
		}
		((btGeneric6DofConstraint*)joint)->calculateTransforms();
		simImpl->dynamicsWorld->addConstraint(joint, true);
		q_offset = ((btGeneric6DofConstraint*)joint)->getRelativePivotPosition(2);

        break;
    }

    case Link::FREE_JOINT:
        break;
        
    case Link::FIXED_JOINT:
    default:
		if(!link->isRoot()){
			btTransform frameA,frameB;
			frameA.setIdentity();
			frameA.setOrigin(btVector3(b(0), b(1), b(2)));
			frameB.setIdentity();
			joint = new btGeneric6DofConstraint(*(parent->body), *body, parent->invShift*frameA, invShift*frameB, false);
			((btGeneric6DofConstraint*)joint)->calculateTransforms();
			simImpl->dynamicsWorld->addConstraint(joint, true);
			((btGeneric6DofConstraint*)joint)->setLinearLowerLimit(btVector3(0.0, 0.0, 0.0));
			((btGeneric6DofConstraint*)joint)->setLinearUpperLimit(btVector3(0.0, 0.0, 0.0));
			((btGeneric6DofConstraint*)joint)->setAngularLowerLimit(btVector3(0.0, 0.0, 0.0));
			((btGeneric6DofConstraint*)joint)->setAngularUpperLimit(btVector3(0.0, 0.0, 0.0));
            if(link->jointType() == Link::CRAWLER_JOINT){
                body->setCollisionFlags(body->getCollisionFlags()  | btCollisionObject::CF_CUSTOM_MATERIAL_CALLBACK);
                body->setUserPointer(this);
                if(!gContactAddedCallback)
                    gContactAddedCallback = CustomMaterialCombinerCallback;
            }
		}
        break;
    }
}


BulletLink::~BulletLink()
{
	if(vertices)
		delete[] vertices;
	if(triangles)
		delete[] triangles;
	if(pMeshData)
		delete pMeshData;
	if(pColShape)
		delete pColShape;
	if(motionState)
		delete motionState;
	if(body){
		dynamicsWorld->removeRigidBody(body);
		delete body;
	}
}

void BulletLink::setKinematicStateToBullet()
{
	const Matrix3& R = link->attitude();
	btTransform transform;
	btMatrix3x3 btR(R(0,0), R(0,1), R(0,2),
					R(1,0), R(1,1), R(1,2),
					R(2,0), R(2,1), R(2,2));
	transform.setBasis(btR);
	const Vector3& p = link->p();
	transform.setOrigin(btVector3(p(0), p(1), p(2)));
	body->setWorldTransform(transform*shift);

	const Vector3 lc = R * link->c();
	const Vector3 v = link->v() + link->w().cross(lc);
	body->setLinearVelocity(btVector3(v(0), v(1), v(2)));
	const Vector3& w = link->w();
	body->setAngularVelocity(btVector3(w(0), w(1), w(2)));
	body->activate(true);
}

void BulletLink::getKinematicStateFromBullet()
{
	if(link->isRoot() && link->isFixedJoint()){
		return;
    }
	btTransform trans;
	motionState->getWorldTransform(trans);
	btTransform strans =  trans * invShift;
	Matrix3 R;
	btMatrix3x3& btR = strans.getBasis();
	R << btR[0][0], btR[0][1], btR[0][2],
 		 btR[1][0], btR[1][1], btR[1][2],
		 btR[2][0], btR[2][1], btR[2][2];
	link->setAttitude(R);
	btVector3& p = strans.getOrigin();
    link->p() << p[0], p[1], p[2];

#if DEBUG_OUT
	std::cout << link->q() << std::endl;
	std::cout << R(0,0) << " " << R(0,1) << " " << R(0,2) << std::endl;
	std::cout << R(1,0) << " " << R(1,1) << " " << R(1,2) << std::endl;
	std::cout << R(2,0) << " " << R(2,1) << " " << R(2,2) << std::endl;
	std::cout << link->p().x() << "," << link->p().y() << "," << link->p().z() << std::endl;
	std::cout << std::endl;
#endif

	if(link->isRotationalJoint()){
		link->q() = ((btHingeConstraint*)joint)->getHingeAngle() - q_offset;
	} else if(link->isSlideJoint()){
		link->q() = ((btGeneric6DofConstraint*)joint)->getRelativePivotPosition(2) - q_offset;      
	}
}

void BulletLink::setTorqueToBullet()
{
	if(link->isRotationalJoint()){
		const Vector3 u = link->u() * link->a();
		const Vector3 uu = link->attitude() * u;
		btVector3 torque(uu(0), uu(1), uu(2));
		body->applyTorque(torque);
		parent->body->applyTorque(-torque);

    } else if(link->isSlideJoint()){
		const Vector3 u = link->u() * link->d();
		const Vector3 uu = link->attitude() * u;
		btVector3 torque(uu(0), uu(1), uu(2));
		btVector3 pos(0,0,0);
		body->applyForce(torque, invShift*pos);
		const Vector3& b = link->b();
		btVector3 pos1(b(0), b(1), b(2));
		parent->body->applyForce(-torque, parent->invShift*pos1);
	}
}

////////////////////////////////Body////////////////////////////////////////
BulletBody::BulletBody(const Body& orgBody)
    : SimulationBody(make_shared<Body>(orgBody))
{

}


BulletBody::~BulletBody()
{
	for(size_t i=0; i < bulletLinks.size(); ++i){
		if(bulletLinks[i]->joint){
			dynamicsWorld->removeConstraint(bulletLinks[i]->joint);
			delete bulletLinks[i]->joint;
		}
    }
	for(size_t i=0; i < extraJoints.size(); ++i){
		if(extraJoints[i]){
			dynamicsWorld->removeConstraint(extraJoints[i]);
			delete extraJoints[i];
		}
	}
    for(size_t i=0; i < forceSensorFeedbacks.size(); ++i){
		if(forceSensorFeedbacks[i]){
			delete forceSensorFeedbacks[i];
		}
	}
}

void BulletBody::createBody(BulletSimulatorItemImpl* simImpl, short group)
{
    body_ = body();
	dynamicsWorld = simImpl->dynamicsWorld;

    BulletLink* rootLink = new BulletLink(simImpl, this, 0, Vector3::Zero(), body_->rootLink(), 
		group, bodyItem()->isSelfCollisionDetectionEnabled());

	setKinematicStateToBullet();
	setExtraJoints();
	setTorqueToBullet();

    sensorHelper.initialize(body_, simImpl->timeStep, simImpl->gravity);

    // set joint feedbacks for force sensors
    const DeviceList<ForceSensor>& forceSensors = sensorHelper.forceSensors();
    forceSensorFeedbacks.resize(forceSensors.size());
    for(size_t i=0; i < forceSensors.size(); ++i){
        forceSensorFeedbacks[i] = new btJointFeedback;
        bulletLinks[forceSensors[i]->link()->index()]->joint->setJointFeedback(forceSensorFeedbacks[i]);
    }
}

void BulletBody::setExtraJoints()
{
	extraJoints.clear();
	int n = body_->extraJoints.size();

    for(int j=0; j < n; ++j){
		Body::ExtraJoint& extraJoint = body_->extraJoints[j];
		
		BulletLinkPtr bulletLinkPair[2];
        for(int i=0; i < 2; ++i){
            BulletLinkPtr bulletLink;
            Link* link = extraJoint.link[i];
            if(link->index() < bulletLinks.size()){
                bulletLink = bulletLinks[link->index()];
                if(bulletLink->link == link){
                    bulletLinkPair[i] = bulletLink;
                }
            }
            if(!bulletLink){
                break;
            }
        }

		if(bulletLinkPair[1]){
		    Link* link0 = bulletLinkPair[0]->link;
			Link* link1 = bulletLinkPair[1]->link;
            Vector3& p0 = extraJoint.point[0];  // link0 local position
            Vector3& a = extraJoint.axis;	    // link0 local axis
			Vector3 wp = link0->attitude() * p0 + link0->p();  // world position
			Vector3 p1 = link1->attitude().transpose() * (wp - link1->p());  // link1 local position;

			if(extraJoint.type == Body::EJ_PISTON){
				Vector3 u(0,0,1);
				Vector3 ty = a.cross(u);
				btMatrix3x3 btR;
				Matrix3 R0;
				if(ty.norm() == 0){
					btR.setIdentity();
					R0.setIdentity();
				}else{
					ty.normalized();
					Vector3 tx = ty.cross(a).normalized();
					R0.col(0) = tx;
					R0.col(1) = ty;
					R0.col(2) = a;
					btR.setValue(R0(0,0), R0(0,1), R0(0,2),
								 R0(1,0), R0(1,1), R0(1,2),
								 R0(2,0), R0(2,1), R0(2,2));
				}
				btTransform frameA,frameB;
				frameB.setBasis(btR);
				frameB.setOrigin(btVector3(p0(0), p0(1), p0(2)));

				Matrix3 R1 = link1->attitude().transpose() * link0->attitude() * R0;
				btMatrix3x3 btR1(R1(0,0), R1(0,1), R1(0,2),
			   					 R1(1,0), R1(1,1), R1(1,2),
								 R1(2,0), R1(2,1), R1(2,2));
				frameA.setBasis(btR1);
				frameA.setOrigin(btVector3(p1(0), p1(1), p1(2)));

				btGeneric6DofConstraint* joint = new btGeneric6DofConstraint(*(bulletLinkPair[1]->body), *(bulletLinkPair[0]->body),
					bulletLinkPair[1]->invShift*frameA, bulletLinkPair[0]->invShift*frameB, false);
				joint->setLinearLowerLimit(btVector3(0,0,1));  //Lowerlimit ＞ Upperlimit → axis is free
				joint->setLinearUpperLimit(btVector3(0,0,-1));
				joint->setAngularLowerLimit(btVector3(0,0,1));  
				joint->setAngularUpperLimit(btVector3(0,0,-1));	
				joint->calculateTransforms();
				dynamicsWorld->addConstraint(joint, true);
				extraJoints.push_back(joint);
			}else if(extraJoint.type == Body::EJ_BALL){
				btVector3 pivotInA(p0(0), p0(1), p0(2));
				btVector3 pivotInB(p1(0), p1(1), p1(2));
				btPoint2PointConstraint* joint = new btPoint2PointConstraint(*(bulletLinkPair[0]->body), *(bulletLinkPair[1]->body),
					bulletLinkPair[0]->invShift*pivotInA, bulletLinkPair[1]->invShift*pivotInB);
				dynamicsWorld->addConstraint(joint, true);
				extraJoints.push_back(joint);
			}
		}
	}
}

void BulletBody::setKinematicStateToBullet()
{
	for(size_t i=0; i < bulletLinks.size(); ++i){
		bulletLinks[i]->setKinematicStateToBullet();
    }
}

void BulletBody::getKinematicStateFromBullet()
{
	for(size_t i=0; i < bulletLinks.size(); ++i){
		bulletLinks[i]->getKinematicStateFromBullet();
    }
}

void BulletBody::setTorqueToBullet()
{
	for(size_t i=1; i < bulletLinks.size(); ++i){
        bulletLinks[i]->setTorqueToBullet();
    }
}

void BulletBody::updateForceSensors()
{
    const DeviceList<ForceSensor>& forceSensors = sensorHelper.forceSensors();
    for(int i=0; i < forceSensors.size(); ++i){
        ForceSensor* sensor = forceSensors.get(i);
        const Link* link = sensor->link();
        btTypedConstraint* joint = bulletLinks[link->index()]->joint;
        btJointFeedback* fb = joint->getJointFeedback();
        Vector3 f, tau;
        if(link->isRotationalJoint()){
            f   << fb->m_appliedForceBodyB.x(), fb->m_appliedForceBodyB.y(), fb->m_appliedForceBodyB.z();
            tau << fb->m_appliedTorqueBodyB.x(),  fb->m_appliedTorqueBodyB.y(),  fb->m_appliedTorqueBodyB.z();
        }else{
            f   << fb->m_appliedForceBodyA.x(), fb->m_appliedForceBodyA.y(), fb->m_appliedForceBodyA.z();
            tau << fb->m_appliedTorqueBodyA.x(),  fb->m_appliedTorqueBodyA.y(),  fb->m_appliedTorqueBodyA.z();
        }
       
        const Matrix3 R = link->R() * sensor->R_local();
        const Vector3 p = link->p() + link->R() * sensor->p_local();
        sensor->f()   = R.transpose() * f;
        sensor->tau() = R.transpose() * (tau - p.cross(f));
        sensor->notifyStateChange();
    }
}

////////////////////////////////////SimItem imple////////////////////////////////////////
void BulletSimulatorItem::initialize(ExtensionManager* ext)
{
    ext->itemManager().registerClass<BulletSimulatorItem>(N_("BulletSimulatorItem"));
    ext->itemManager().addCreationPanel<BulletSimulatorItem>();
}


BulletSimulatorItem::BulletSimulatorItem()
{
    impl = new BulletSimulatorItemImpl(this);
}


BulletSimulatorItemImpl::BulletSimulatorItemImpl(BulletSimulatorItem* self)
    : self(self)

{
	initialize();
    gravity << 0.0, 0.0, -DEFAULT_GRAVITY_ACCELERATION;
	erp = DEFAULT_ERP;
	numIterations = DEFAULT_NUMITERATION;
	restitution = DEFAULT_RESTITUTION;
	friction = DEFAULT_FRICTION;
	erp2 = DEFAULT_ERP2;
	splitImpulsePenetrationThreshold = DEFAULT_SPLITIMPULSEPENETRATIONTHRESHOLD;
}


BulletSimulatorItem::BulletSimulatorItem(const BulletSimulatorItem& org)
    : SimulatorItem(org)
{
    impl = new BulletSimulatorItemImpl(this, *org.impl);
}


BulletSimulatorItemImpl::BulletSimulatorItemImpl(BulletSimulatorItem* self, const BulletSimulatorItemImpl& org)
    : self(self)
{
	initialize();
	gravity = org.gravity;
	erp = org.erp;
	numIterations = org.numIterations;
	restitution = org.restitution;
	friction = org.friction;
	erp2 = org.erp2;
	splitImpulsePenetrationThreshold = org.splitImpulsePenetrationThreshold;
}

void BulletSimulatorItemImpl::initialize()
{
	collisionConfiguration =0;
	dispatcher = 0;
	broadphase = 0;
	solver =0;
	dynamicsWorld = 0;
    self->SimulatorItem::setAllLinkPositionOutputMode(true);
}

BulletSimulatorItem::~BulletSimulatorItem()
{
    delete impl;
}


BulletSimulatorItemImpl::~BulletSimulatorItemImpl()
{
	clear();
}


void BulletSimulatorItem::setAllLinkPositionOutputMode(bool on)
{
    // The mode is not changed.
    // This simulator only supports the all link position output
    // because joint positions may be slightly changed
}


ItemPtr BulletSimulatorItem::doDuplicate() const
{
    return new BulletSimulatorItem(*this);
}


SimulationBodyPtr BulletSimulatorItem::createSimulationBody(const Body& orgBody)
{
    return make_shared<BulletBody>(orgBody);
}


bool BulletSimulatorItem::initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies)
{
    return impl->initializeSimulation(simBodies);
}

bool BulletSimulatorItemImpl::initializeSimulation(const std::vector<SimulationBodyPtr>& simBodies)
{
	clear();

	collisionConfiguration = new btDefaultCollisionConfiguration();
	dispatcher = new btCollisionDispatcher(collisionConfiguration);
	broadphase = new btDbvtBroadphase();
	solver = new btSequentialImpulseConstraintSolver();
	dynamicsWorld = new btDiscreteDynamicsWorld(dispatcher,broadphase,solver,collisionConfiguration);

    btVector3 g(gravity.x(), gravity.y(), gravity.z());
	dynamicsWorld->setGravity(g);
	timeStep = self->worldTimeStep();
	setSolverParameter();

	short group = 1;
	 for(size_t i=0; i < simBodies.size(); ++i){
         addBody(static_cast<BulletBody*>(simBodies[i].get()), group<<i);
    }

	btGImpactCollisionAlgorithm::registerAlgorithm(dispatcher);


    return true;
}

void BulletSimulatorItemImpl::clear()
{
	if(dynamicsWorld)
		delete dynamicsWorld;
	if(solver)
		delete solver;
	if(dispatcher)
		delete dispatcher;
	if(collisionConfiguration)
		delete collisionConfiguration;	
	if(broadphase)
		delete broadphase;
}

void BulletSimulatorItemImpl::addBody(BulletBody* bulletBody, short group)
{
	Body& body = *bulletBody->body();

    Link* rootLink = body.rootLink();
    rootLink->v().setZero();
    rootLink->dv().setZero();
    rootLink->w().setZero();
    rootLink->dw().setZero();

    for(int i=0; i < body.numJoints(); ++i){
        Link* joint = body.joint(i);
        joint->u() = 0.0;
        joint->dq() = 0.0;
        joint->ddq() = 0.0;
    }
    
    body.clearExternalForces();
    body.calcForwardKinematics(true, true);

    bulletBody->createBody(this, group);
}

void BulletSimulatorItem::initializeSimulationThread()
{

}


bool BulletSimulatorItem::stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies)
{
    return impl->stepSimulation(activeSimBodies);
}


bool BulletSimulatorItemImpl::stepSimulation(const std::vector<SimulationBodyPtr>& activeSimBodies)
{
	for(size_t i=0; i < activeSimBodies.size(); ++i){
        BulletBody* bulletBody = static_cast<BulletBody*>(activeSimBodies[i].get());
		bulletBody->body_->setVirtualJointForces();
		bulletBody->setTorqueToBullet();
	}

	dynamicsWorld->stepSimulation(timeStep,2,timeStep/2.);

#if DEBUG_OUT
	int numManifolds = dispatcher->getNumManifolds();
    for (int i=0;i<numManifolds;i++)
    {
		btPersistentManifold* contactManifold = dispatcher->getManifoldByIndexInternal(i);
		int numContacts = contactManifold->getNumContacts();
        for (int j=0;j<numContacts;j++)
        {
			btManifoldPoint& pt = contactManifold->getContactPoint(j);
            btVector3 ptA = pt.getPositionWorldOnA();
            btVector3 ptB = pt.getPositionWorldOnB();
			btVector3 normal = pt.m_normalWorldOnB; 
			std::cout << i << " " << j <<" pos= " <<ptA[0]<<" "<<ptA[1]<<" "<<ptA[2]<<"norm= " <<
			normal[0] << " " << normal[1] << " " << normal[2] << " Impulse= " << pt.getAppliedImpulse() <<std::endl;
            std::cout << pt.m_lateralFrictionDir1.x() << " " << pt.m_lateralFrictionDir1.y() << " " << pt.m_lateralFrictionDir1.z() << " " << pt.m_appliedImpulseLateral1 << std::endl;
            std::cout << pt.m_lateralFrictionDir2.x() << " " << pt.m_lateralFrictionDir2.y() << " " << pt.m_lateralFrictionDir2.z() << " " << pt.m_appliedImpulseLateral2 << std::endl;
            std::cout << pt.m_contactMotion1 << std::endl;
		}
	}
#endif

	for(size_t i=0; i < activeSimBodies.size(); ++i){
        BulletBody* bulletBody = static_cast<BulletBody*>(activeSimBodies[i].get());
        if(!bulletBody->sensorHelper.forceSensors().empty()){
            bulletBody->updateForceSensors();
        }
        bulletBody->getKinematicStateFromBullet();
        if(bulletBody->sensorHelper.hasGyroOrAccelSensors()){
            bulletBody->sensorHelper.updateGyroAndAccelSensors();
        }
    }
	return true;
}


void BulletSimulatorItem::doPutProperties(PutPropertyFunction& putProperty)
{
    SimulatorItem::doPutProperties(putProperty);
    impl->doPutProperties(putProperty);
}


void BulletSimulatorItemImpl::doPutProperties(PutPropertyFunction& putProperty)
{
	putProperty.min(0.0).max(1.0);
	putProperty(_("Error Reduction Parameter"), erp, changeProperty(erp));
	putProperty(_("Restitution"), restitution, changeProperty(restitution));
	putProperty(_("Friction"), friction, changeProperty(friction));
	putProperty(_("ERP2"), erp2, changeProperty(erp2));
	putProperty.min(1);
	putProperty(_("Num of Iterations"), numIterations, changeProperty(numIterations));
	putProperty.decimals(4).min(-1.0).max(1.0);
	putProperty(_("SplitImpulsePenetrationThreshold"), splitImpulsePenetrationThreshold, changeProperty(splitImpulsePenetrationThreshold));
}


bool BulletSimulatorItem::store(Archive& archive)
{
    SimulatorItem::store(archive);
    impl->store(archive);
    return true;
}


void BulletSimulatorItemImpl::store(Archive& archive)
{
	archive.write("ErrorReductionParameter", erp);
	archive.write("NumIterations", numIterations);
	archive.write("Restitution", restitution);
	archive.write("Friction", friction);
	archive.write("ERP2", erp2);
	archive.write("SplitImpulsePenetrationThreshold", splitImpulsePenetrationThreshold);
}


bool BulletSimulatorItem::restore(const Archive& archive)
{
    SimulatorItem::restore(archive);
    impl->restore(archive);
    return true;
}


void BulletSimulatorItemImpl::restore(const Archive& archive)
{
	archive.read("ErrorReductionParameter", erp);
	archive.read("NumIterations", numIterations);
	archive.read("Restitution", restitution);
	archive.read("Friction", friction);
	archive.read("ERP2", erp2);
	archive.read("SplitImpulsePenetrationThreshold", splitImpulsePenetrationThreshold);
}

void BulletSimulatorItemImpl::setSolverParameter()
{
    if(dynamicsWorld != NULL){
        btContactSolverInfo& slvInfo = dynamicsWorld->getSolverInfo();
        slvInfo.m_erp = erp;
        slvInfo.m_numIterations = numIterations;
        slvInfo.m_erp2 = erp2;
        slvInfo.m_splitImpulsePenetrationThreshold = splitImpulsePenetrationThreshold;
        
        slvInfo.m_solverMode |= SOLVER_DISABLE_VELOCITY_DEPENDENT_FRICTION_DIRECTION+SOLVER_USE_2_FRICTION_DIRECTIONS;
        slvInfo.m_solverMode |= SOLVER_ENABLE_FRICTION_DIRECTION_CACHING;
    }
}
