/**
   @file
   @author Shin'ichiro Nakaoka
*/

#include "PointCloudUtil.h"
#include "EasyScanner.h"
#include "Exception.h"
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {

    enum Element { E_X, E_Y, E_Z, E_NORMAL_X,E_NORMAL_Y, E_NORMAL_Z };

    SgPointSetPtr readPoints(EasyScanner& scanner, const std::vector<Element>& elements, int numPoints)
    {
        SgPointSetPtr pointSet;
        
        SgVertexArrayPtr vertices = new SgVertexArray();
        vertices->reserve(numPoints);

        const int numElements = elements.size();

        SgNormalArrayPtr normals;
        bool hasNormals = false;
        for(int i=0; i < numElements; ++i){
            if(elements[i] >= E_NORMAL_X){
                hasNormals = true;
                normals = new SgNormalArray();
                normals->reserve(numPoints);
            }
        }

        SgVector3 vertex = SgVector3::Zero();
        SgVector3 normal = SgVector3::Zero();

        while(!scanner.isEOF()){
            scanner.skipBlankLines();
            for(int i=0; i < numElements; ++i){
                double value = scanner.readDoubleEx(_("Illeagel point values"));
                switch(elements[i]){
                case E_X: vertex.x() = value; break;
                case E_Y: vertex.y() = value; break;
                case E_Z: vertex.z() = value; break;
                case E_NORMAL_X: normal.x() = value; break;
                case E_NORMAL_Y: normal.y() = value; break;
                case E_NORMAL_Z: normal.z() = value; break;
                }
            }
            vertices->push_back(vertex);
            if(hasNormals){
                normals->push_back(normal);
            }
            scanner.readLFEOF();
        }

        if(!vertices->empty()){
            pointSet = new SgPointSet();
            pointSet->setVertices(vertices);
            if(normals){
                pointSet->setNormals(normals);
            }
        }

        return pointSet;
    }
}


SgPointSetPtr cnoid::loadPCD(const std::string& filename)
{
    SgPointSetPtr pointSet;
    
    try {
        EasyScanner scanner(filename);
        scanner.setCommentChar('#');

        int numPoints = 0;

        std::vector<Element> elements;

        while(true){
            scanner.skipBlankLines();
            scanner.readWordEx(_("Illegal header key"));

            if(scanner.stringValue == "FIELDS"){
                while(scanner.readWord()){
                    if(scanner.stringValue == "x"){
                        elements.push_back(E_X);
                    } else if(scanner.stringValue == "y"){
                        elements.push_back(E_Y);
                    } else if(scanner.stringValue == "z"){
                        elements.push_back(E_Z);
                    } else if(scanner.stringValue == "normal_x"){
                        elements.push_back(E_NORMAL_X);
                    } else if(scanner.stringValue == "normal_y"){
                        elements.push_back(E_NORMAL_Y);
                    } else if(scanner.stringValue == "normal_z"){
                        elements.push_back(E_NORMAL_Z);
                    }
                }
            } else if(scanner.stringValue == "POINTS"){
                numPoints = scanner.readIntEx(_("The 'POINTS' field is not correctly specified."));
            } else if(scanner.stringValue == "DATA"){
                scanner.readWordEx(_("The 'DATA' field is not correctly specified."));
                if(scanner.stringValue == "ascii"){
                    scanner.readLFex();
                    if(elements.empty()){
                        scanner.throwException(_("The specification of field elements is not found."));
                    }
                    pointSet = readPoints(scanner, elements, numPoints);
                    break;
                } else {
                    scanner.throwException(_("The 'ascii' format is only supported for the point DATA."));
                }
            } else {
                scanner.skipToLineEnd();
            }
            scanner.readLFEOFex(_("The field value is not correctly specified."));
        }
    } catch(EasyScanner::Exception& ex){
        throw file_read_error() << error_info_message(ex.getFullMessage());
    }

    return pointSet;
}
