/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "SceneGraph.h"

using namespace std;
using namespace cnoid;

namespace {
    const double PI = 3.14159265358979323846;
}


SgUpdate::~SgUpdate()
{

}


void SgBoundingBox::clear()
{
    empty_ = true;
    min_.setConstant(std::numeric_limits<SgVector3::Scalar>::max());
    max_.setConstant(-std::numeric_limits<SgVector3::Scalar>::max());
}


SgBoundingBox::SgBoundingBox()
{
    clear();
}

        
SgBoundingBox::SgBoundingBox(const SgVector3& min, const SgVector3& max)
{
    set(min, max);
}


void SgBoundingBox::set(const SgVector3& min, const SgVector3& max)
{
    min_ = min;
    max_ = max;
    empty_ = (min.x() >= max.x()) && (min.y() >= max.y()) && (min.z() >= max.z());
}


const SgVector3 SgBoundingBox::center() const
{
    return (min_ + max_) / 2.0;
}


const SgFloat SgBoundingBox::boundingSphereRadius() const
{
    return (max_ - center()).norm();
}


void SgBoundingBox::expandBy(const SgFloat x,  const SgFloat y, const SgFloat z )
{
    if(x < min_.x()){
        min_.x() = x;
    }
    if(x > max_.x()){
        max_.x() = x;
    }
    if(y < min_.y()){
        min_.y() = y;
    }
    if(y > max_.y()){
        max_.y() = y;
    }
    if(z < min_.z()){
        min_.z() = z;
    }
    if(z > max_.z()){
        max_.z() = z;
    }
    if(empty_){
        empty_ = (min_.x() >= max_.x()) && (min_.y() >= max_.y()) && (min_.z() >= max_.z());
    }
}


void SgBoundingBox::expandBy(const SgBoundingBox& bbox)
{
    if(!bbox.empty()){
        if(bbox.min().x() < min_.x()){
            min_.x() = bbox.min().x();
        }
        if(bbox.max().x() > max_.x()){
            max_.x() = bbox.max().x();
        }
        if(bbox.min().y() < min_.y()){
            min_.y() = bbox.min().y();
        }
        if(bbox.max().y() > max_.y()){
            max_.y() = bbox.max().y();
        }
        if(bbox.min().z() < min_.z()){
            min_.z() = bbox.min().z();
        }
        if(bbox.max().z() > max_.z()){
            max_.z() = bbox.max().z();
        }
        if(empty_){
            empty_ = (min_.x() >= max_.x()) && (min_.y() >= max_.y()) && (min_.z() >= max_.z());
        }
    }
}


void SgBoundingBox::transform(const SgAffine3& T)
{
    if(!empty()){
        const SgVector3 p1 = min_;
        const SgVector3 p2 = max_;
        clear();
        expandBy(T * SgVector3(p1.x(), p1.y(), p1.z()));
        expandBy(T * SgVector3(p1.x(), p2.y(), p1.z()));
        expandBy(T * SgVector3(p2.x(), p1.y(), p1.z()));
        expandBy(T * SgVector3(p2.x(), p2.y(), p1.z()));
        expandBy(T * SgVector3(p1.x(), p1.y(), p2.z()));
        expandBy(T * SgVector3(p1.x(), p2.y(), p2.z()));
        expandBy(T * SgVector3(p2.x(), p1.y(), p2.z()));
        expandBy(T * SgVector3(p2.x(), p2.y(), p2.z()));
    }
}


SgObject::SgObject()
{

}


SgObject::SgObject(const SgObject& org)
    : name_(org.name_)
{

}


SgObject* SgObject::clone(const CopyFlags& copyFlags) const
{
    return new SgObject(*this);
}


int SgObject::numElements() const
{
    return 0;
}


SgObject* SgObject::element(int index)
{
    return 0;
}


void SgObject::notifyUpdate(SgUpdate& update)
{
    update.push(this);
    sigUpdated_(update);
    for(NodeContainer::const_iterator p = upperNodes.begin(); p != upperNodes.end(); ++p){
        (*p)->notifyUpdate(update);
    }
    update.pop();
}


void SgObject::addUpperNode(SgObject* node)
{
    upperNodes.insert(node);
    if(upperNodes.size() == 1){
        sigGraphConnection_(true);
    }
}


void SgObject::removeUpperNode(SgObject* node)
{
    upperNodes.erase(node);
    if(upperNodes.empty()){
        sigGraphConnection_(false);
    }
}
            

SgObject* SgNode::clone(const CopyFlags& copyFlags) const
{
    return new SgNode(*this);
}


void SgNode::accept(SgVisitor& visitor)
{
    visitor.visitNode(this);
}


const SgBoundingBox& SgNode::boundingBox() const
{
    static const SgBoundingBox bbox; // empty one
    return bbox;
}


SgAffine3 cnoid::calcTotalTransform(const SgNodePath& path)
{
    SgAffine3 T = SgAffine3::Identity();
    for(SgNodePath::const_iterator p = path.begin(); p != path.end(); ++p){
        SgTransform* transform = dynamic_cast<SgTransform*>(*p);
        if(transform){
            T = T * transform->T();
        }
    }
    return T;
}


SgGroup::SgGroup()
{
    isBboxCacheValid = false;
}


SgGroup::SgGroup(const SgGroup& org, const CopyFlags& copyFlags)
    : SgNode(org)
{
    children.reserve(org.numChildren());
    if(copyFlags & DEEP_COPY_NODES){
        for(const_iterator p = org.begin(); p != org.end(); ++p){
            children.push_back(static_cast<SgNode*>((*p)->clone(copyFlags)));
        }
    } else {
        // shallow copy
        for(const_iterator p = org.begin(); p != org.end(); ++p){
            children.push_back(*p);
        }
    }
    isBboxCacheValid = true;
    bboxCache = org.bboxCache;
}


SgGroup::~SgGroup()
{
    for(iterator p = begin(); p != end(); ++p){
        (*p)->removeUpperNode(this);
    }
}


SgObject* SgGroup::clone(const CopyFlags& copyFlags) const
{
    return new SgGroup(*this, copyFlags);
}


int SgGroup::numElements() const
{
    return children.size();
}


SgObject* SgGroup::element(int index)
{
    return children[index].get();
}


void SgGroup::accept(SgVisitor& visitor)
{
    visitor.visitGroup(this);
}


void SgGroup::notifyUpdate(SgUpdate& update)
{
    //if(update.action() & SgUpdate::BBOX_UPDATED){
    invalidateBoundingBox();
    SgNode::notifyUpdate(update);
    //}
}


const SgBoundingBox& SgGroup::boundingBox() const
{
    if(isBboxCacheValid){
        return bboxCache;
    }
    bboxCache.clear();
    for(const_iterator p = begin(); p != end(); ++p){
        bboxCache.expandBy((*p)->boundingBox());
    }
    isBboxCacheValid = true;
    return bboxCache;
}


bool SgGroup::contains(SgNodePtr node) const
{
    for(const_iterator p = begin(); p != end(); ++p){
        if((*p) == node){
            return true;
        }
    }
    return false;
}


void SgGroup::clearChildren(bool doNotify)
{
    if(!children.empty()){
        children.clear();
        if(doNotify){
            SgUpdate update(SgUpdate::REMOVED);
            notifyUpdate(update);
        }
    }
}


void SgGroup::addChild(SgNodePtr node, bool doNotify)
{
    if(node){
        children.push_back(node);
        node->addUpperNode(this);
    }
    if(doNotify){
        SgUpdate update(SgUpdate::ADDED);
        notifyUpdate(update);
    }
}


bool SgGroup::removeChild(SgNodePtr node, bool doNotify)
{
    bool removed = false;
    iterator p = begin();
    while(p != end()){
        if((*p) == node){
            SgNodePtr child = *p;
            p = children.erase(p);
            child->removeUpperNode(this);
            removed = true;
        } else {
            ++p;
        }
    }
    if(doNotify && removed){
        SgUpdate update(SgUpdate::REMOVED);
        notifyUpdate(update);
    }
    return removed;
}
    

void SgGroup::removeChild(int index, bool doNotify)
{
    iterator p = children.begin() + index;
    (*p)->removeUpperNode(this);
    children.erase(p);
    if(doNotify){
        SgUpdate update(SgUpdate::REMOVED);
        notifyUpdate(update);
    }
}


SgInvariantGroup::SgInvariantGroup(const SgInvariantGroup& org, const CopyFlags& copyFlags)
    : SgGroup(org, copyFlags)
{

}


SgObject* SgInvariantGroup::clone(const CopyFlags& copyFlags) const
{
    return new SgInvariantGroup(*this, copyFlags);
}


void SgInvariantGroup::accept(SgVisitor& visitor)
{
    visitor.visitInvariantGroup(this);
}


SgTransform::SgTransform()
    : T_(SgAffine3::Identity())
{

}


SgTransform::SgTransform(const SgAffine3& T)
  : T_(T)
{

}        


SgObject* SgTransform::clone(const CopyFlags& copyFlags) const
{
    return new SgTransform(*this,copyFlags);
}


void SgTransform::accept(SgVisitor& visitor)
{
    visitor.visitTransform(this);
}


const SgBoundingBox& SgTransform::boundingBox() const
{
    if(isBboxCacheValid){
        return bboxCache;
    }
    bboxCache.clear();
    for(const_iterator p = begin(); p != end(); ++p){
        bboxCache.expandBy((*p)->boundingBox());
    }
    untransformedBboxCache = bboxCache;
    bboxCache.transform(T_);
    isBboxCacheValid = true;
    return bboxCache;
}


const SgBoundingBox& SgTransform::untransformedBoundingBox() const
{
    if(!isBboxCacheValid){
        boundingBox();
    }
    return untransformedBboxCache;
}


SgMaterial::SgMaterial()
{
    ambientIntensity_ = 0.02;
    diffuseColor_ << 0.8, 0.8, 0.8;
    emissiveColor_.setZero();
    specularColor_.setZero();
    shininess_ = 0.2;
    transparency_ = 0.0;
}


SgMaterial::SgMaterial(const SgMaterial& org)
    : SgObject(org)
{
    ambientIntensity_ = org.ambientIntensity_;
    diffuseColor_ = org.diffuseColor_;
    emissiveColor_ = org.emissiveColor_;
    specularColor_ = org.specularColor_;
    shininess_ = org.shininess_;
    transparency_ = org.transparency_;
}


SgObject* SgMaterial::clone(const CopyFlags& copyFlags) const
{
    return new SgMaterial(*this);
}


SgTexture::SgTexture()
{
    width_  = 0;
    height_ = 0;
    numComponents_ = 0;
    repeatS_ = true; 
    repeatT_ = true; 
}


SgTexture::SgTexture(const SgTexture& org)
    : SgObject(org),
      pixels_(org.pixels_)
{
    width_ = org.width_;
    height_ = org.height_;
    numComponents_ = org.numComponents_;
    repeatS_ = org.repeatS_;
    repeatT_ = org.repeatT_;
}


SgObject* SgTexture::clone(const CopyFlags& copyFlags) const
{
    return new SgTexture(*this);
}


SgMesh::SgMesh()
{
    colorPerVertex_ = false;
    isSolid_ = false;
}


SgMesh::SgMesh(const SgMesh& org, const CopyFlags& copyFlags)
    : SgObject(org),
      triangles_(org.triangles_),
      normalIndices_(org.normalIndices_),
      colorIndices_(org.colorIndices_),
      primitive_(org.primitive_)
{
    if(copyFlags & (DEEP_COPY_ARRAYS | DEEP_COPY_OBJECTS)){
        vertices_ = new SgVertexArray(*org.vertices_);
        normals_ = new SgNormalArray(*org.normals_);
        colors_ = new SgVertexArray(*org.colors_);
    } else {
        vertices_ = org.vertices_;
        normals_ = org.normals_;
        colors_ = org.colors_;
    }
        
    colorPerVertex_ = org.colorPerVertex_;
    isSolid_ = org.isSolid_;
    bbox = org.bbox;
}

    
SgObject* SgMesh::clone(const CopyFlags& copyFlags) const
{
    return new SgMesh(*this, copyFlags);
}


int SgMesh::numElements() const
{
    int n = 0;
    if(vertices_) ++n;
    if(normals_) ++n;
    if(colors_) ++n;
    return n;
}


SgObject* SgMesh::element(int index)
{
    SgObject* elements[3] = { 0, 0, 0 };
    int i = 0;
    if(vertices_) elements[i++] = vertices_.get();
    if(normals_) elements[i++] = normals_.get();
    if(colors_) elements[i++] = colors_.get();
    return elements[index];
}


void SgMesh::setVertices(SgVertexArrayPtr vertices)
{
    if(vertices_){
        vertices_->removeUpperNode(this);
    }
    vertices_ = vertices;
    vertices->addUpperNode(this);
}


void SgMesh::setNormals(SgNormalArrayPtr normals)
{
    if(normals_){
        normals_->removeUpperNode(this);
    }
    normals_ = normals;
    normals->addUpperNode(this);
}


void SgMesh::setColors(SgColorArrayPtr colors)
{
    if(colors_){
        colors_->removeUpperNode(this);
    }
    colors_ = colors;
    colors->addUpperNode(this);
}


void SgMesh::updateBoundingBox()
{
    bbox.clear();
    if(vertices_){
        for(SgVertexArray::const_iterator p = vertices_->begin(); p != vertices_->end(); ++p){
            bbox.expandBy(*p);
        }
    }
}

void SgMesh::setBox(SgVector3 size)
{
    SgFloat x = size.x()*0.5;
    SgFloat y = size.y()*0.5;
    SgFloat z = size.z()*0.5;
    SgVertexArrayPtr vertices = new SgVertexArray();
    vertices->reserve(8);
    setVertices(vertices);

    vertices->push_back(SgVector3(x,y,z));
    vertices->push_back(SgVector3(-x,y,z));
    vertices->push_back(SgVector3(-x,-y,z));
    vertices->push_back(SgVector3(x,-y,z));
    vertices->push_back(SgVector3(x,y,-z));
    vertices->push_back(SgVector3(-x,y,-z));
    vertices->push_back(SgVector3(-x,-y,-z));
    vertices->push_back(SgVector3(x,-y,-z));

    triangles_.clear();
    triangles_.reserve(12);
    addTriangle(0,1,2);
    addTriangle(2,3,0);
    addTriangle(0,5,1);
    addTriangle(0,4,5);
    addTriangle(1,5,6);
    addTriangle(1,6,2);
    addTriangle(2,6,7);
    addTriangle(2,7,3);
    addTriangle(3,7,4);
    addTriangle(3,4,0);
    addTriangle(4,6,5);
    addTriangle(4,7,6);

    generateNormals(PI);

    primitive_ = Box(size);
}


void SgMesh::setSphere(SgFloat radius, int divisionNumber)
{
    const int vdn = divisionNumber / 2;  // latitudinal division number
    const int hdn = divisionNumber;      // longitudinal division number

    SgVertexArrayPtr vertices = new SgVertexArray();
    vertices->reserve((vdn - 1) * hdn + 2);
    setVertices(vertices);

    for(int i=1; i < vdn; i++){ // latitudinal direction
        const double tv = i * PI / vdn;
        for(int j=0; j < hdn; j++){ // longitudinal direction
            const double th = j * 2.0 * PI / hdn;
            vertices->push_back(SgVector3(radius * sin(tv) * cos(th), radius * cos(tv), radius * sin(tv) * sin(th)));
        }
    }
    
    const int topIndex  = vertices->size();
    vertices->push_back(SgVector3(0.0,  radius, 0.0));
    const int bottomIndex = vertices->size();
    vertices->push_back(SgVector3(0.0, -radius, 0.0));

    triangles_.clear();
    triangles_.reserve(vdn * hdn * 2 * 3);

    // top faces
    for(int i=0; i < hdn; ++i){
        addTriangle(topIndex, (i+1) % hdn, i);
    }

    // side faces
    for(int i=0; i < vdn - 2; ++i){
        const int upper = i * hdn;
        const int lower = (i + 1) * hdn;
        for(int j=0; j < hdn; ++j) {
            // upward convex triangle
            addTriangle(j + upper, ((j + 1) % hdn) + lower, j + lower);
            // downward convex triangle
            addTriangle(j + upper, ((j + 1) % hdn) + upper, ((j + 1) % hdn) + lower);
        }
    }
    
    // bottom faces
    const int offset = (vdn - 2) * hdn;
    for(int i=0; i < hdn; ++i){
        addTriangle(bottomIndex, (i % hdn) + offset, ((i+1) % hdn) + offset);
    }

    //! \todo set normals directly without using the following function
    generateNormals(PI);

    primitive_ = Sphere(radius);
}


void SgMesh::setCone(SgFloat radius, SgFloat height, bool bottom, bool side, int divisionNumber)
{
    SgVertexArrayPtr vertices = new SgVertexArray();
    vertices->reserve(divisionNumber + 1);
    setVertices(vertices);

    for(int i=0;  i < divisionNumber; ++i){
        const double angle = i * 2.0 * PI / divisionNumber;
        vertices->push_back(SgVector3(radius * cos(angle), -height/2.0, radius * sin(angle)));
    }

    const int topIndex = vertices->size();
    vertices->push_back(SgVector3(0.0f, height / 2.0, 0.0f));
    const int bottomCenterIndex = vertices->size();
    vertices->push_back(SgVector3(0.0f, -height / 2.0, 0.0f));

    triangles_.clear();
    triangles_.reserve((divisionNumber * 2) * 4);

    for(int i=0; i < divisionNumber; ++i){
        // side faces
        if(side){
            addTriangle(topIndex, (i + 1) % divisionNumber, i);
        }
        // bottom faces
        if(bottom){
            addTriangle(bottomCenterIndex, i, (i + 1) % divisionNumber);
        }
    }

    generateNormals(PI);

    primitive_ = Cone(radius, height);

}


void SgMesh::setCylinder(SgFloat radius, SgFloat height, bool bottom, bool side, bool top, int divisionNumber)
{
    SgVertexArrayPtr vertices = new SgVertexArray();
    vertices->resize(divisionNumber * 2);
    setVertices(vertices);

    const double y = height / 2.0;
    for(int i=0 ; i < divisionNumber ; i++ ){
        const double angle = i * 2.0 * PI / divisionNumber;
        SgVector3& vtop = vertices->at(i);
        SgVector3& vbottom = vertices->at(i + divisionNumber);
        vtop[0] = vbottom[0] = radius * cos(angle);
        vtop[2] = vbottom[2] = radius * sin(angle);
        vtop[1]    =  y;
        vbottom[1] = -y;
    }

    const int topCenterIndex = vertices->size();
    vertices->push_back(SgVector3(0.0f,  y, 0.0f));
    const int bottomCenterIndex = vertices->size();
    vertices->push_back(SgVector3(0.0f, -y, 0.0f));

    triangles_.clear();
    triangles_.reserve((divisionNumber * 4) * 4);

    for(int i=0; i < divisionNumber; ++i){
        // top face
        if(top){
            addTriangle(topCenterIndex, (i+1) % divisionNumber, i);
        }
        // side face (upward convex triangle)
        if(side){        
            addTriangle(i, ((i+1) % divisionNumber) + divisionNumber, i + divisionNumber);
            // side face (downward convex triangle)
            addTriangle(i, (i+1) % divisionNumber, ((i + 1) % divisionNumber) + divisionNumber);
        }
        // bottom face
        if(bottom){
            addTriangle(bottomCenterIndex, i + divisionNumber, ((i+1) % divisionNumber) + divisionNumber);
        }
    }

    generateNormals(PI);

    primitive_ = Cylinder(radius, height);
}

namespace {
    
    void calculateFaceNormals(SgMesh* mesh, SgNormalArray& faceNormals, vector<vector<int> >& vertexIndexToFaceIndicesMap)
    {
        const SgVertexArray& vertices = *mesh->vertices();
        const int numVertices = vertices.size();
        const SgIndexArray& triangles = mesh->triangles();
        const int numFaces = triangles.size() / 3;
        faceNormals.reserve(numFaces);
        vertexIndexToFaceIndicesMap.resize(numVertices);

        for(int faceIndex=0; faceIndex < numFaces; ++faceIndex){

            const SgVector3& v0 = vertices[triangles[faceIndex * 3]];
            const SgVector3& v1 = vertices[triangles[faceIndex * 3 + 1]];
            const SgVector3& v2 = vertices[triangles[faceIndex * 3 + 2]];
            const SgVector3 normal((v1 - v0).cross(v2 - v0).normalized());
            faceNormals.push_back(normal);

            for(int i=0; i < 3; ++i){
                const int vertexIndex = triangles[faceIndex * 3 + i];
                vector<int>& facesOfVertex = vertexIndexToFaceIndicesMap[vertexIndex];

                /**
                   \todo Angle between adjacent edges should be taken into account
                   to generate natural normals
                */
                bool isSameNormalFaceFound = false;
                for(size_t j=0; j < facesOfVertex.size(); ++j){
                    const SgVector3& otherNormal = faceNormals[facesOfVertex[j]];
                    // the same face is not appended
                    if((otherNormal - normal).squaredNorm() <= numeric_limits<SgFloat>::epsilon()){
                        isSameNormalFaceFound = true;
                        break;
                    }
                }
                if(!isSameNormalFaceFound){
                    facesOfVertex.push_back(faceIndex);
                }
            }
        }
    }


    void setVertexNormals(SgMesh* mesh, SgNormalArray& faceNormals,
                          vector<vector<int> >& vertexIndexToFaceIndicesMap, float creaseAngle)
    {
        const SgVertexArray& vertices = *mesh->vertices();
        const int numVertices = vertices.size();
        const SgIndexArray& triangles = mesh->triangles();
        const int numFaces = triangles.size() / 3;

        SgNormalArray& normals = *mesh->normals();
        SgIndexArray& normalIndices = mesh->normalIndices();
        normalIndices.clear();
        normalIndices.reserve(triangles.size());

        vector<vector<int> > vertexIndexToNormalIndicesMap(numVertices);

        // const float cosCreaseAngle = cosf(creaseAngle);

        for(int faceIndex=0; faceIndex < numFaces; ++faceIndex){

            for(int i=0; i < 3; ++i){

                const int vertexIndex = triangles[faceIndex * 3 + i];
                const vector<int>& facesOfVertex = vertexIndexToFaceIndicesMap[vertexIndex];
                const SgVector3& currentFaceNormal = faceNormals[faceIndex];
                SgVector3 normal = currentFaceNormal;
                bool normalIsFaceNormal = true;
                
                // avarage normals of the faces whose crease angle is below the 'creaseAngle' variable
                for(size_t j=0; j < facesOfVertex.size(); ++j){
                    const int adjacentFaceIndex = facesOfVertex[j];
                    const SgVector3& adjacentFaceNormal = faceNormals[adjacentFaceIndex];
                    const float angle = acosf(currentFaceNormal.dot(adjacentFaceNormal)
                                              / (currentFaceNormal.norm() * adjacentFaceNormal.norm()));
                    if(angle > 1.0e-6f && angle < creaseAngle){
                        normal += adjacentFaceNormal;
                        normalIsFaceNormal = false;
                    }
                }
                if(!normalIsFaceNormal){
                    normal.normalize();
                }
                
                int normalIndex = -1;
                
                for(int j=0; j < 3; ++j){
                    int vertexIndex2 = triangles[faceIndex * 3 + j];
                    const vector<int>& normalIndicesOfVertex = vertexIndexToNormalIndicesMap[vertexIndex2];
                    for(size_t k=0; k < normalIndicesOfVertex.size(); ++k){
                        int index = normalIndicesOfVertex[k];
                        if((normals[index] - normal).squaredNorm() <= numeric_limits<SgFloat>::epsilon()){
                            normalIndex = index;
                            goto normalIndexFound;
                        }
                    }
                }
                if(normalIndex < 0){
                    normalIndex = normals.size();
                    normals.push_back(normal);
                    vertexIndexToNormalIndicesMap[vertexIndex].push_back(normalIndex);
                }
                
              normalIndexFound:
                normalIndices.push_back(normalIndex);
            }
        }
    }
}


void SgMesh::generateNormals(double creaseAngle)
{
    if(!vertices_ || triangles_.empty()){
        return;
    }
    SgNormalArrayPtr normals = new SgNormalArray();
    setNormals(normals);

    SgNormalArray faceNormals;
    vector<vector<int> > vertexIndexToFaceIndicesMap;
    calculateFaceNormals(this, faceNormals, vertexIndexToFaceIndicesMap);
    setVertexNormals(this, faceNormals, vertexIndexToFaceIndicesMap, creaseAngle);
}


SgShape::SgShape()
{

}


SgShape::SgShape(const SgShape& org, const CopyFlags& copyFlags)
    : SgNode(org)
{
    if(copyFlags & DEEP_COPY_NODES){
        setMesh(static_cast<SgMesh*>(org.mesh()->clone()));
        setMaterial(static_cast<SgMaterial*>(org.material()->clone()));
        setTexture(static_cast<SgTexture*>(org.texture()->clone()));
    } else {
        setMesh(const_cast<SgMesh*>(org.mesh()));
        setMaterial(const_cast<SgMaterial*>(org.material()));
        setTexture(const_cast<SgTexture*>(org.texture()));
    }
}


SgObject* SgShape::clone(const CopyFlags& copyFlags) const
{
    return new SgShape(*this, copyFlags);
}


int SgShape::numElements() const
{
    int n = 0;
    if(mesh_) ++n;
    if(material_) ++n;
    if(texture_) ++n;
    return n;
}


SgObject* SgShape::element(int index)
{
    SgObject* elements[3] = { 0, 0, 0 };
    int i = 0;
    if(mesh_) elements[i++] = mesh_.get();
    if(material_) elements[i++] = material_.get();
    if(texture_) elements[i++] = texture_.get();
    return elements[index];
}


void SgShape::accept(SgVisitor& visitor)
{
    visitor.visitShape(this);
}


const SgBoundingBox& SgShape::boundingBox() const
{
    if(mesh()){
        return mesh()->boundingBox();
    }
    return SgNode::boundingBox();
}


void SgShape::setMesh(SgMeshPtr mesh)
{
    if(mesh_){
        mesh_->removeUpperNode(this);
    }
    mesh_ = mesh;
    mesh->addUpperNode(this);
}


void SgShape::setMaterial(SgMaterialPtr material)
{
    if(material_){
        material_->removeUpperNode(this);
    }
    material_ = material;
    material->addUpperNode(this);
}


void SgShape::setTexture(SgTexturePtr texture)
{
    if(texture_){
        texture_->removeUpperNode(this);
    }
    texture_ = texture;
    texture->addUpperNode(this);
}


SgPointSetBase::SgPointSetBase()
{


}
        

SgPointSetBase::SgPointSetBase(const SgPointSetBase& org, const CopyFlags& copyFlags)
    : SgNode(org)
{
    if(copyFlags & DEEP_COPY_NODES){
        vertices_ = new SgVertexArray(*org.vertices_);
        colors_ = new SgVertexArray(*org.colors_);
        setMaterial(static_cast<SgMaterial*>(org.material()->clone()));
    } else {
        vertices_ = org.vertices_;
        colors_ = org.colors_;
        setMaterial(const_cast<SgMaterial*>(org.material()));
    }
    bbox = org.bbox;
}


int SgPointSetBase::numElements() const
{
    int n = 0;
    if(vertices_) ++n;
    if(colors_) ++n;
    return n;
}
    

SgObject* SgPointSetBase::element(int index)
{
    SgObject* elements[2] = { 0, 0 };
    int i = 0;
    if(vertices_) elements[i++] = vertices_.get();
    if(colors_) elements[i++] = colors_.get();
    return elements[index];
}
    

const SgBoundingBox& SgPointSetBase::boundingBox() const
{
    return bbox;
}


void SgPointSetBase::updateBoundingBox()
{
    bbox.clear();
    if(vertices_){
        for(SgVertexArray::const_iterator p = vertices_->begin(); p != vertices_->end(); ++p){
            bbox.expandBy(*p);
        }
    }
}


void SgPointSetBase::setVertices(SgVertexArrayPtr vertices)
{
    if(vertices_){
        vertices_->removeUpperNode(this);
    }
    vertices_ = vertices;
    vertices->addUpperNode(this);
}


void SgPointSetBase::setNormals(SgNormalArrayPtr normals)
{
    if(normals_){
        normals_->removeUpperNode(this);
    }
    normals_ = normals;
    normals->addUpperNode(this);
}


void SgPointSetBase::setMaterial(SgMaterialPtr material)
{
    if(material_){
        material_->removeUpperNode(this);
    }
    material_ = material;
    material->addUpperNode(this);
}


void SgPointSetBase::setColors(SgColorArrayPtr colors)
{
    if(colors_){
        colors_->removeUpperNode(this);
    }
    colors_ = colors;
    colors->addUpperNode(this);
}


SgPointSet::SgPointSet()
{
    pointSize_ = 0.0;
}


SgPointSet::SgPointSet(const SgPointSet& org, const CopyFlags& copyFlags)
    : SgPointSetBase(org, copyFlags)
{
    pointSize_ = org.pointSize_;
}


SgObject* SgPointSet::clone(const CopyFlags& copyFlags) const
{
    return new SgPointSet(*this, copyFlags);
}


void SgPointSet::accept(SgVisitor& visitor)
{
    visitor.visitPointSet(this);
}

   
SgLineSet::SgLineSet()
{
    colorPerVertex_ = false;
    lineWidth_ = 0.0;
}


SgLineSet::SgLineSet(const SgLineSet& org, const CopyFlags& copyFlags)
    : SgPointSetBase(org, copyFlags)
{
    colorIndices_ = org.colorIndices_;
    colorPerVertex_ = org.colorPerVertex_;
    lineWidth_ = org.lineWidth_;
}

    
SgObject* SgLineSet::clone(const CopyFlags& copyFlags) const
{
    return new SgLineSet(*this, copyFlags);
}
    

void SgLineSet::accept(SgVisitor& visitor)
{
    visitor.visitLineSet(this);
}


SgPreprocessed::SgPreprocessed()
{

}


SgPreprocessed::SgPreprocessed(const SgPreprocessed& org)
    : SgNode(org)
{

}


SgObject* SgPreprocessed::clone(const CopyFlags& copyFlags) const
{
    return new SgPreprocessed(*this);
}


void SgPreprocessed::accept(SgVisitor& visitor)
{
    visitor.visitPreprocessed(this);
}


SgLight::SgLight()
{
    on_ = true;
    color_.setOnes();
    intensity_ = 1.0;
    ambientIntensity_ = 0.0;
}


SgLight::SgLight(const SgLight& org)
    : SgPreprocessed(org)
{
    on_ = org.on_;
    color_ = org.color_;
    intensity_ = org.intensity_;
    ambientIntensity_ = org.ambientIntensity_;
}


SgObject* SgLight::clone(const CopyFlags& copyFlags) const
{
    return new SgLight(*this);
}


void SgLight::accept(SgVisitor& visitor)
{
    visitor.visitLight(this);
}


SgDirectionalLight::SgDirectionalLight()
{
    direction_ << 0.0, 0.0, -1.0;
}


SgDirectionalLight::SgDirectionalLight(const SgDirectionalLight& org)
    : SgLight(org)
{
    direction_ = org.direction_;
}


SgObject* SgDirectionalLight::clone(const CopyFlags& copyFlags) const
{
    return new SgDirectionalLight(*this);
}


void SgDirectionalLight::accept(SgVisitor& visitor)
{
    visitor.visitLight(this);
}


SgPointLight::SgPointLight()
{
    location_.setZero();
    attenuation_ << 1.0, 0.0, 0.0;
}


SgPointLight::SgPointLight(const SgPointLight& org)
    : SgLight(org)
{
    location_ = org.location_;
    attenuation_ = org.attenuation_;
}


SgObject* SgPointLight::clone(const CopyFlags& copyFlags) const
{
    return new SgPointLight(*this);
}


void SgPointLight::accept(SgVisitor& visitor)
{
    visitor.visitLight(this);
}


SgSpotLight::SgSpotLight()
{
    direction_ << 0.0, 0.0, -1.0;
    beamWidth_ = 1.570796;
    cutOffAngle_ = 0.785398;
}


SgSpotLight::SgSpotLight(const SgSpotLight& org)
    : SgPointLight(org)
{
    direction_ = org.direction_;
    beamWidth_ = org.beamWidth_;
    cutOffAngle_ = org.cutOffAngle_;
}


SgObject* SgSpotLight::clone(const CopyFlags& copyFlags) const
{
    return new SgSpotLight(*this);
}


void SgSpotLight::accept(SgVisitor& visitor)
{
    visitor.visitLight(this);
}


SgCamera::SgCamera()
{

}


SgCamera::SgCamera(const SgCamera& org)
    : SgPreprocessed(org)
{

}


void SgCamera::accept(SgVisitor& visitor)
{
    visitor.visitCamera(this);
}


SgIsometry3 SgCamera::positionLookingFor(const SgVector3& eye, const SgVector3& direction, const SgVector3& up)
{
    SgVector3 d = direction.normalized();
    SgVector3 c = d.cross(up).normalized();
    SgVector3 u = c.cross(d);
    SgIsometry3 C;
    C.linear() << c, u, -d;
    C.translation() = eye;
    return C;
}


SgIsometry3 SgCamera::positionLookingAt(const SgVector3& eye, const SgVector3& center, const SgVector3& up)
{
    return positionLookingFor(eye, (center - eye), up);
}


SgVector3 SgCamera::right(const SgAffine3& T)
{
   return T.linear().col(0);
}


SgVector3 SgCamera::direction(const SgAffine3& T)
{
    return -T.linear().col(2);
}


SgVector3 SgCamera::up(const SgAffine3& T)
{
    return T.linear().col(1);
}





SgPerspectiveCamera::SgPerspectiveCamera()
{
    fieldOfView_ = 0.785398;
}


SgPerspectiveCamera::SgPerspectiveCamera(const SgPerspectiveCamera& org)
    : SgCamera(org)
{
    fieldOfView_ = org.fieldOfView_;
}


SgObject* SgPerspectiveCamera::clone(const CopyFlags& copyFlags) const
{
    return new SgPerspectiveCamera(*this);
}


void SgPerspectiveCamera::accept(SgVisitor& visitor)
{
    visitor.visitCamera(this);
}


SgFloat SgPerspectiveCamera::fovy(SgFloat aspectRatio) const
{
    if(aspectRatio >= 1.0){
        return fieldOfView_;
    } else {
        return 2.0 * atanf(tanf(fieldOfView_ / 2.0) / aspectRatio);
    }
}


SgOrthographicCamera::SgOrthographicCamera()
{
    height_ = 2.0;
}


SgOrthographicCamera::SgOrthographicCamera(const SgOrthographicCamera& org)
    : SgCamera(org)
{
    height_ = org.height_;
}


SgObject* SgOrthographicCamera::clone(const CopyFlags& copyFlags) const
{
    return new SgOrthographicCamera(*this);
}


void SgOrthographicCamera::accept(SgVisitor& visitor)
{
    visitor.visitCamera(this);
}


void SgVisitor::visitNode(SgNode* node)
{

}


void SgVisitor::visitGroup(SgGroup* group)
{
    visitNode(group);
    
    for(SgGroup::iterator p = group->begin(); p != group->end(); ++p){
        (*p)->accept(*this);
    }
}


void SgVisitor::visitTransform(SgTransform* transform)
{
    visitGroup(transform);
}


void SgVisitor::visitInvariantGroup(SgInvariantGroup* group)
{
    visitGroup(group);
}


void SgVisitor::visitShape(SgShape* shape)
{
    visitNode(shape);
}


void SgVisitor::visitPointSet(SgPointSet* pointSet)
{
    visitNode(pointSet);
}


void SgVisitor::visitLineSet(SgLineSet* lineSet)
{
    visitNode(lineSet);
}


void SgVisitor::visitPreprocessed(SgPreprocessed* preprocessed)
{
    visitNode(preprocessed);

}


void SgVisitor::visitLight(SgLight* light)
{
    visitPreprocessed(light);

}


void SgVisitor::visitCamera(SgCamera* camera)
{
    visitPreprocessed(camera);
}


