/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_UTIL_SCENE_GRAPH_H_INCLUDED
#define CNOID_UTIL_SCENE_GRAPH_H_INCLUDED

#include "Referenced.h"
#include "EigenTypes.h"
#include "SignalProxy.h"
#include <boost/variant.hpp>
#include <string>
#include <vector>
#include <set>
#include <bitset>
#include "exportdecl.h"

namespace cnoid {

#define CNOID_SCENE_GRAPH_USE_FLOAT
    
#ifdef  CNOID_SCENE_GRAPH_USE_FLOAT
    typedef float SgFloat;
    typedef Vector3f SgVector3;
    typedef Affine3f SgAffine3;
    typedef Isometry3f SgIsometry3;
    typedef Matrix4f SgMatrix4;
    typedef Translation3f SgTranslation3;
    typedef AngleAxisf SgAngleAxis;
    typedef Vector2f SgVector2;
    typedef Vector4f SgVector4;
#else
    typedef double SgFloat;
    typedef Vector3d SgVector3;
    typedef Affine3d SgAffine3;
    typedef Isometry3d SgIsometry3;
    typedef Matrix4d SgMatrix4;
    typedef Translation3d SgTranslation3;
    typedef AngleAxisd SgAngleAxis;
    typedef Vector2d SgVector2;
    typedef Vector4d SgVector4;
#endif

    class SgObject;
    typedef boost::intrusive_ptr<SgObject> SgObjectPtr;

    class SgVisitor;

    class CNOID_EXPORT SgUpdate
    {
      public:
        enum Action { 
            ADDED = 1 << 0,
            REMOVED = 1 << 1,
            BBOX_UPDATED = 1 << 2,
            MODIFIED = 1 << 3
        };

        typedef std::vector<SgObject*> Path;
        
        SgUpdate(Action action) : action_(action) { path_.reserve(16); }
        virtual ~SgUpdate();
        Action action() const { return action_; }
        const Path& path() const { return path_; }
        void push(SgObject* node) { path_.push_back(node); }
        void pop() { path_.pop_back(); }

    private:
        Path path_;
        Action action_;
    };


    class CNOID_EXPORT SgBoundingBox
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SgBoundingBox();
        SgBoundingBox(const SgVector3& min, const SgVector3& max);
        
        void set(const SgVector3& min, const SgVector3& max);
        void clear();

        bool empty() const { return empty_; }
        const SgVector3& min() const { return min_; }
        const SgVector3& max() const { return max_; }
        const SgVector3 center() const;
        const SgFloat boundingSphereRadius() const;
        
        void expandBy(const SgBoundingBox& bbox);
        void expandBy(const SgFloat x,  const SgFloat y, const SgFloat z );
        void expandBy(const SgVector3& v){ expandBy(v.x(), v.y(), v.z()); }

        void transform(const SgAffine3& T);

    private:
        bool empty_;
        SgVector3 min_;
        SgVector3 max_;
    };


    class CNOID_EXPORT SgObject : public Referenced
    {
      public:
        enum { SHALLOW_COPY = 0,
               DEEP_COPY_OBJECTS = 1 << 0,
               DEEP_COPY_NODES = 1 << 1,
               DEEP_COPY_MATERIALS = 1 << 2,
               DEEP_COPY_TEXTURES = 1 << 3,
               DEEP_COPY_MESHES = 1 << 4,
               DEEP_COPY_ARRAYS = 1 << 5,
               DEEP_COPY_ALL = 0x7FFFFFFF
        };
        typedef unsigned int CopyFlags;

        SgObject();
        SgObject(const SgObject& org);

        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;

        const std::string& name() { return name_; }
        void setName(const std::string& name) { name_ = name; }

        virtual int numElements() const;
        virtual SgObject* element(int index);

        SignalProxy<boost::signal<void(const SgUpdate& update)> > sigUpdated() {
            return sigUpdated_;
        }
        
        virtual void notifyUpdate(SgUpdate& update);
            
        void notifyUpdate(SgUpdate::Action action = SgUpdate::MODIFIED) {
            SgUpdate update(action);
            notifyUpdate(update);
        }

        void addUpperNode(SgObject* node);
        void removeUpperNode(SgObject* node);

        /**
           This signal is emitted when the object is first attached to an upper node
           or the object is detached from all the upper node.
        */
        SignalProxy<boost::signal<void(bool on)> > sigGraphConnection() {
            return sigGraphConnection_;
        }

      private:
        std::string name_;
        typedef std::set<SgObject*> NodeContainer;
        NodeContainer upperNodes;
        boost::signal<void(const SgUpdate& update)> sigUpdated_;
        boost::signal<void(bool on)> sigGraphConnection_;
    };


    class CNOID_EXPORT SgNode : public SgObject
    {
      protected:
        SgNode() { }
        SgNode(const SgNode& org) : SgObject(org) { }
        
      public:
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
        virtual const SgBoundingBox& boundingBox() const;
    };
    typedef boost::intrusive_ptr<SgNode> SgNodePtr;

    typedef std::vector<SgNode*> SgNodePath;

    CNOID_EXPORT SgAffine3 calcTotalTransform(const SgNodePath& path);


    class CNOID_EXPORT SgGroup : public SgNode
    {
        typedef std::vector<SgNodePtr> Container;
        
      public:
        typedef Container::iterator iterator;
        typedef Container::const_iterator const_iterator;
        typedef Container::reverse_iterator reverse_iterator;
        typedef Container::const_reverse_iterator const_reverse_iterator;

        SgGroup();
        SgGroup(const SgGroup& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        ~SgGroup();
        
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual int numElements() const;
        virtual SgObject* element(int index);
        virtual void accept(SgVisitor& visitor);
        virtual void notifyUpdate(SgUpdate& update);

        void notifyUpdate(SgUpdate::Action action = SgUpdate::MODIFIED) {
            SgUpdate update(action);
            notifyUpdate(update);
        }
        
        virtual const SgBoundingBox& boundingBox() const;
        
        void invalidateBoundingBox() { isBboxCacheValid = false; }

        iterator begin() { return children.begin(); }
        const_iterator begin() const { return children.begin(); }
        iterator end() { return children.end(); }
        const_iterator end() const { return children.end(); }

        reverse_iterator rbegin() { return children.rbegin(); }
        const_reverse_iterator rbegin() const { return children.rbegin(); }
        reverse_iterator rend() { return children.rend(); }
        const_reverse_iterator rend() const { return children.rend(); }

        bool contains(SgNodePtr node) const;

        bool empty() const { return children.empty(); }
        int numChildren() const { return children.size(); }
        SgNode* child(int index) { return children[index].get(); }

        virtual void clearChildren(bool doNotify = false);
        virtual void addChild(SgNodePtr node, bool doNotify = false);
        virtual bool removeChild(SgNodePtr node, bool doNotify = false);
        virtual void removeChild(int index, bool doNotify = false);

      private:
        Container children;

      protected:
        mutable SgBoundingBox bboxCache;
        mutable bool isBboxCacheValid;
    };
    typedef boost::intrusive_ptr<SgGroup> SgGroupPtr;


    class CNOID_EXPORT SgInvariantGroup : public SgGroup
    {
      public:
        SgInvariantGroup() { }
        SgInvariantGroup(const SgInvariantGroup& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
    };
    typedef boost::intrusive_ptr<SgInvariantGroup> SgInvariantGroupPtr;
    
    
    class CNOID_EXPORT SgTransform : public SgGroup
    {
      public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SgTransform();
        SgTransform(const SgAffine3& T);
        SgTransform(const SgTransform& org, const CopyFlags& copyFlags = DEEP_COPY_NODES)
            : SgGroup(org, copyFlags), T_(org.T_) { }
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
        virtual const SgBoundingBox& boundingBox() const;

        const SgBoundingBox& untransformedBoundingBox() const;

        SgAffine3& T() { return T_; }
        const SgAffine3& T() const { return T_; }

        SgAffine3::TranslationPart translation() { return T_.translation(); }
        SgAffine3::ConstTranslationPart translation() const { return T_.translation(); }

        SgAffine3::LinearPart rotation() { return T_.linear(); }
        SgAffine3::ConstLinearPart rotation() const { return T_.linear(); }

        template<class Scalar, int Mode, int Options>
        void setPosition(const Eigen::Transform<Scalar, 3, Mode, Options>& T) {
            T_ = T.template cast<SgFloat>();
        }
        template<class Scalar, int Mode, int Options>
        void setTransform(const Eigen::Transform<Scalar, 3, Mode, Options>& T) {
            T_ = T.template cast<SgFloat>();
        }
        template<typename Derived>
        void setRotation(const Eigen::MatrixBase<Derived>& R) {
            T_.linear() = R.template cast<SgFloat>();
        }
        template<typename Derived>
        void setTranslation(const Eigen::MatrixBase<Derived>& p) {
            T_.translation() = p.template cast<SgFloat>();
        }

      private:
        SgAffine3 T_; // shoudl be Isometry3?
        mutable SgBoundingBox untransformedBboxCache;
    };
    typedef boost::intrusive_ptr<SgTransform> SgTransformPtr;


    class CNOID_EXPORT SgMaterial : public SgObject
    {
      public:
        SgMaterial();
        SgMaterial(const SgMaterial& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;

        SgFloat ambientIntensity() const { return ambientIntensity_; }
        void setAmbientIntensity(SgFloat intensity) { ambientIntensity_ = intensity; }
        const SgVector3& diffuseColor() const { return diffuseColor_; }
        template<typename Derived> void setDiffuseColor(const Eigen::MatrixBase<Derived>& c) {
            diffuseColor_ = c.template cast<SgFloat>(); }
        const SgVector3& emissiveColor() const { return emissiveColor_; }
        template<typename Derived> void setEmissiveColor(const Eigen::MatrixBase<Derived>& c) {
            emissiveColor_ = c.template cast<SgFloat>(); }
        SgFloat shininess() const { return shininess_; }
        void setShininess(SgFloat s) { shininess_ = s; }
        const SgVector3& specularColor() const { return specularColor_; }
        template<typename Derived> void setSpecularColor(const Eigen::MatrixBase<Derived>& c) {
            specularColor_ = c.template cast<SgFloat>(); }
        SgFloat transparency() const { return transparency_; }
        void setTransparency(SgFloat t) { transparency_ = t; }
        
      private:
        SgFloat ambientIntensity_;
        SgVector3 diffuseColor_;
        SgVector3 emissiveColor_;
        SgFloat shininess_;
        SgVector3 specularColor_;
        SgFloat transparency_;
    };
    typedef boost::intrusive_ptr<SgMaterial> SgMaterialPtr;


    class CNOID_EXPORT SgTexture : public SgObject
    {
      public:
        SgTexture();
        SgTexture(const SgTexture& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;

        const std::vector<unsigned char>& pixels() const { return pixels_; }
        std::vector<unsigned char>& pixels() { return pixels_; }
        int width() const { return width_; }
        int height() const { return height_; }
        void setSize(int width, int height) { width_ = width; height_ = height; }
        int numComponents() const { return numComponents_; }
        void setNumComponents(int n) { numComponents_ = n; }
        bool repeatS() const { return repeatS_; }
        bool repeatT() const { return repeatT_; }
        void setRepeat(bool s, bool t) { repeatS_ = s; repeatT_ = t; }

      private:
        std::vector<unsigned char> pixels_;
        int width_;
        int height_;
        int numComponents_;
        bool repeatS_;
        bool repeatT_;
    };
    typedef boost::intrusive_ptr<SgTexture> SgTexturePtr;


    template<class T> class SgVectorArray : public SgObject
    {
        typedef std::vector<T> Container;
        
      public:
        typedef typename Container::iterator iterator;
        typedef typename Container::const_iterator const_iterator;
        typedef typename Container::size_type size_type;
        typedef typename Container::value_type value_type;
        typedef typename Container::reference reference;
        typedef typename Container::const_reference const_reference;
        typedef typename Container::pointer pointer;
        typedef typename Container::const_pointer const_pointer;
        typedef typename T::Scalar Scalar;

        SgVectorArray() { }
        SgVectorArray(const std::vector<T>& org) : values(org) { }

        template<class Element>
        SgVectorArray(const std::vector<Element>& org) {
            values.reserve(org.size());
            for(typename std::vector<Element>::const_iterator p = org.begin(); p != org.end(); ++p){
                values.push_back(p->template cast<typename T::Scalar>());
            }
        }
        
        SgVectorArray(const SgVectorArray& org)
            : SgObject(org), values(org.values) { }
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const { return new SgVectorArray(*this); }
        
        SgVectorArray<T>& operator=(const SgVectorArray<T>& rhs) {
            values = rhs.value;
        }
        iterator begin() { return values.begin(); }
        const_iterator begin() const { return values.begin(); }
        iterator end() { return values.end(); }
        const_iterator end() const { return values.end(); }
        size_type size() const { return values.size(); }
        void resize(size_type s) { values.resize(s); }
        void resize(size_type s, const T& v) { values.resize(s, v); }
        bool empty() const { return values.empty(); }
        void reserve(size_type s) { values.reserve(s); }
        T& operator[](size_type i) { return values[i]; }
        const T& operator[](size_type i) const { return values[i]; }
        T& at(size_type i) { return values[i]; }
        const T& at(size_type i) const { return values[i]; }
        T& front() { return values.front(); }
        const T& front() const { return values.front(); }
        T& back() { return values.back(); }
        const T& back() const { return values.back(); }
        Scalar* data() { return values.front().data(); }
        const Scalar* data() const { return values.front().data(); }
        void push_back(const T& v) { values.push_back(v); }
        void pop_back() { values.pop_back(); }
        iterator erase(iterator p) { return values.erase(p); }
        iterator erase(iterator first, iterator last) { return values.erase(first, last); }
        void clear() { values.clear(); }

      private:
        Container values;
    };

    typedef SgVectorArray<SgVector3> SgVertexArray;
    typedef boost::intrusive_ptr<SgVertexArray> SgVertexArrayPtr;

    typedef SgVectorArray<SgVector3> SgNormalArray;
    typedef boost::intrusive_ptr<SgNormalArray> SgNormalArrayPtr;

    typedef SgVectorArray<SgVector3> SgColorArray;
    typedef boost::intrusive_ptr<SgColorArray> SgColorArrayPtr;

    typedef std::vector<unsigned int> SgIndexArray;


    class CNOID_EXPORT SgMesh : public SgObject
    {
      public:
        SgMesh();
        SgMesh(const SgMesh& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual int numElements() const;
        virtual SgObject* element(int index);

        void setVertices(SgVertexArrayPtr vertices);
        void setNormals(SgNormalArrayPtr normals);
        void setColors(SgColorArrayPtr colors);

        SgVertexArray* vertices() { return vertices_.get(); }
        const SgVertexArray* vertices() const { return vertices_.get(); }
        SgNormalArray* normals() { return normals_.get(); }
        const SgNormalArray* normals() const { return normals_.get(); }
        SgColorArray* colors() { return colors_.get(); }
        const SgColorArray* colors() const { return colors_.get(); }

        const SgBoundingBox& boundingBox() const { return bbox; }
        void updateBoundingBox();

        /**
           Triangle indices (triangles variable) should be CCW.
        */
        const SgIndexArray& triangles() const { return triangles_; }
        SgIndexArray& triangles() { return triangles_; }

        void addTriangle(int v0, int v1, int v2){
            triangles_.push_back(v0);
            triangles_.push_back(v1);
            triangles_.push_back(v2);
        }
        
        /**
           Normals are assinged for vertices in triangles.
           If this is empty, the accumulated vertex indices in triangles are used for specifying normals.
        */
        const SgIndexArray& normalIndices() const { return normalIndices_; }
        SgIndexArray& normalIndices() { return normalIndices_; }

        void generateNormals(double creaseAngle);
        
        const SgIndexArray& colorIndices() const { return colorIndices_; }
        SgIndexArray& colorIndices() { return colorIndices_; }

        bool colorPerVertex() const { return colorPerVertex_; }
        void setColorPerVertex(bool on) { colorPerVertex_ = on; }
        bool isSolid() const { return isSolid_; }
        void setSolid(bool on) { isSolid_ = on; }

        enum PrimitiveType { MESH = 0, BOX, SPHERE, CYLINDER, CONE };

        class Mesh { }; // defined for no primitive information

        class Box {
        public:
            Box() { }
            Box(SgVector3 size) : size(size) { }
            SgVector3 size;
        };
        class Sphere {
        public:
            Sphere() { }
            Sphere(SgFloat radius) : radius(radius) { }
            SgFloat radius;
        };
       class Cylinder {
       public:
            Cylinder() { }
            Cylinder(SgFloat radius, SgFloat height) :
                radius(radius), height(height), bottom(true), side(true), top(true) { }
            SgFloat radius;
            SgFloat height;
            bool bottom;
            bool side;
            bool top;
        };
        class Cone {
        public:
            Cone() { }
            Cone(SgFloat radius, SgFloat height) :
                radius(radius), height(height), bottom(true), side(true) { }
            SgFloat radius;
            SgFloat height;
            bool bottom;
            bool side;
        };

        typedef boost::variant<Mesh, Box, Sphere, Cylinder, Cone> Primitive;

        const int primitiveType() const { return primitive_.which(); }
        template<class TPrimitive> const TPrimitive& primitive() const { return boost::get<TPrimitive>(primitive_); }
        template<class TPrimitive> void setPrimitive(const TPrimitive& prim) { primitive_ = prim; }

        void setSphere(SgFloat radius, int divisionNumber = 20);
        void setBox(SgVector3 size);
        void setCone(SgFloat radius, SgFloat height, bool bottom = true, bool side = true, int divisionNumber = 20);
        void setCylinder(SgFloat radius, SgFloat height, bool bottom = true, bool side = true, bool top = true, int divisionNumber = 20);

      private:
        SgBoundingBox bbox;
        SgVertexArrayPtr vertices_;
        SgNormalArrayPtr normals_;
        SgColorArrayPtr colors_;
        SgIndexArray triangles_;
        SgIndexArray normalIndices_;
        SgIndexArray colorIndices_;
        bool colorPerVertex_;
        bool isSolid_;
        Primitive primitive_;
    };
    typedef boost::intrusive_ptr<SgMesh> SgMeshPtr;


    class CNOID_EXPORT SgShape : public SgNode
    {
      public:
        SgShape();
        SgShape(const SgShape& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual int numElements() const;
        virtual SgObject* element(int index);
        virtual void accept(SgVisitor& visitor);
        virtual const SgBoundingBox& boundingBox() const;
        
        void setMesh(SgMeshPtr mesh);
        void setMaterial(SgMaterialPtr material);
        void setTexture(SgTexturePtr texture);

        SgMesh* mesh() { return mesh_.get(); }
        const SgMesh* mesh() const { return mesh_.get(); }
        SgMaterial* material() { return material_.get(); }
        const SgMaterial* material() const { return material_.get(); }
        SgTexture* texture() { return texture_.get(); }
        const SgTexture* texture() const { return texture_.get(); }

      private:
        SgMeshPtr mesh_;
        SgMaterialPtr material_;
        SgTexturePtr texture_;
    };
    typedef boost::intrusive_ptr<SgShape> SgShapePtr;


    class CNOID_EXPORT SgPointSetBase : public SgNode
    {
      public:
        SgPointSetBase();
        SgPointSetBase(const SgPointSetBase& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual int numElements() const;
        virtual SgObject* element(int index);
        virtual const SgBoundingBox& boundingBox() const;
        void updateBoundingBox();
        
        void setVertices(SgVertexArrayPtr vertices);
        void setMaterial(SgMaterialPtr material);
        void setNormals(SgNormalArrayPtr normals);
        void setColors(SgColorArrayPtr colors);
        
        SgVertexArray* vertices() { return vertices_.get(); }
        const SgVertexArray* vertices() const { return vertices_.get(); }
        SgMaterial* material() { return material_.get(); }
        const SgMaterial* material() const { return material_.get(); }
        SgNormalArray* normals() { return normals_.get(); }
        const SgNormalArray* normals() const { return normals_.get(); }
        SgColorArray* colors() { return colors_.get(); }
        const SgColorArray* colors() const { return colors_.get(); }

      private:
        SgBoundingBox bbox;
        SgVertexArrayPtr vertices_;
        SgMaterialPtr material_;
        SgNormalArrayPtr normals_;
        SgColorArrayPtr colors_;
    };
    typedef boost::intrusive_ptr<SgPointSetBase> SgPointSetBasePtr;


    class CNOID_EXPORT SgPointSet : public SgPointSetBase
    {
      public:
        SgPointSet();
        SgPointSet(const SgPointSet& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
        
        void setPointSize(double size) { pointSize_ = size; }

        /**
           The default value of this is zero and the current system value is used then.
        */
        double pointSize() const { return pointSize_; }

      private:
        double pointSize_;
    };
    typedef boost::intrusive_ptr<SgPointSet> SgPointSetPtr;


    class CNOID_EXPORT SgLineSet : public SgPointSetBase
    {
      public:
        SgLineSet();
        SgLineSet(const SgLineSet& org, const CopyFlags& copyFlags = DEEP_COPY_NODES);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        const SgIndexArray& colorIndices() const { return colorIndices_; }
        SgIndexArray& colorIndices() { return colorIndices_; }

        bool colorPerVertex() const { return colorPerVertex_; }
        void setColorPerVertex(bool on) { colorPerVertex_ = on; }

        void setLineWidth(double width) { lineWidth_ = width; }

        /**
           The default value of this is zero and the current system value is used then.
        */
        double lineWidth() const { return lineWidth_; }

      private:
        SgIndexArray colorIndices_;
        bool colorPerVertex_;
        double lineWidth_;
    };
    typedef boost::intrusive_ptr<SgLineSet> SgLineSetPtr;
    

    class CNOID_EXPORT SgPreprocessed : public SgNode
    {
      protected:
        SgPreprocessed();
        SgPreprocessed(const SgPreprocessed& org);

      public:
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
    };


    class CNOID_EXPORT SgLight : public SgPreprocessed
    {
      protected:
        SgLight();
        SgLight(const SgLight& org);
      public:
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        bool on() const { return on_; }
        void on(bool on) { on_ = on; }
        const SgVector3& color() const { return color_; }
        template<typename Derived> void setColor(const Eigen::MatrixBase<Derived>& c) {
            color_ = c.template cast<SgFloat>(); }
        SgFloat intensity() const { return intensity_; }
        void setIntensity(SgFloat intensity) { intensity_ = intensity; }
        SgFloat ambientIntensity() const { return ambientIntensity_; }
        void setAmbientIntensity(SgFloat intensity) { ambientIntensity_ = intensity; }
        
      private:
        bool on_;
        SgVector3 color_;
        SgFloat intensity_;
        SgFloat ambientIntensity_;
    };
    typedef boost::intrusive_ptr<SgLight> SgLightPtr;


    class CNOID_EXPORT SgDirectionalLight : public SgLight
    {
      public:
        SgDirectionalLight();
        SgDirectionalLight(const SgDirectionalLight& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        const SgVector3& direction() const { return direction_; }
        template<typename Derived> void setDirection(const Eigen::MatrixBase<Derived>& d) {
            direction_ = d.template cast<SgFloat>(); }

      private:
        SgVector3 direction_;
    };
    typedef boost::intrusive_ptr<SgDirectionalLight> SgDirectionalLightPtr;


    class CNOID_EXPORT SgPointLight : public SgLight
    {
      public:
        SgPointLight();
        SgPointLight(const SgPointLight& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        const SgVector3& location() const { return location_; }
        template<typename Derived> void setLocation(const Eigen::MatrixBase<Derived>& l) {
            location_ = l.template cast<SgFloat>(); }

        const SgVector3& attenuation() const { return attenuation_; }
        template<typename Derived> void setAttenuation(const Eigen::MatrixBase<Derived>& a) {
            attenuation_ = a.template cast<SgFloat>(); }
        
      private:
        SgVector3 location_;
        SgVector3 attenuation_;
    };
    typedef boost::intrusive_ptr<SgPointLight> SgPointLightPtr;


    class CNOID_EXPORT SgSpotLight : public SgPointLight
    {
      public:
        SgSpotLight();
        SgSpotLight(const SgSpotLight& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        const SgVector3& direction() const { return direction_; }
        template<typename Derived> void setDirection(const Eigen::MatrixBase<Derived>& d) {
            direction_ = d.template cast<SgFloat>(); }

        SgFloat beamWidth() const { return beamWidth_; }
        void setBeamWidth(SgFloat w) { beamWidth_ = w; }
        SgFloat cutOffAngle() const { return cutOffAngle_; }
        void setCutOffAngle(SgFloat a) { cutOffAngle_ = a; }
        
      private:
        SgVector3 direction_;
        SgFloat beamWidth_;
        SgFloat cutOffAngle_;
    };
    typedef boost::intrusive_ptr<SgSpotLight> SgSpotLightPtr;


    class CNOID_EXPORT SgCamera : public SgPreprocessed
    {
      protected:
        SgCamera();
        SgCamera(const SgCamera& org);
      public:
        virtual void accept(SgVisitor& visitor);

        static SgIsometry3 positionLookingFor(const SgVector3& eye, const SgVector3& direction, const SgVector3& up);
        static SgIsometry3 positionLookingAt(const SgVector3& eye, const SgVector3& center, const SgVector3& up);
        static SgVector3 right(const SgAffine3& T);
        static SgVector3 direction(const SgAffine3& T);
        static SgVector3 up(const SgAffine3& T);
    };
    typedef boost::intrusive_ptr<SgCamera> SgCameraPtr;


    class CNOID_EXPORT SgPerspectiveCamera : public SgCamera
    {
      public:
        SgPerspectiveCamera();
        SgPerspectiveCamera(const SgPerspectiveCamera& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        SgFloat fieldOfView() const { return fieldOfView_; }
        void setFieldOfView(SgFloat fov) { fieldOfView_ = fov; }
        
        SgFloat fovy(SgFloat aspectRatio) const;

      private:
        SgFloat fieldOfView_;
    };
    typedef boost::intrusive_ptr<SgPerspectiveCamera> SgPerspectiveCameraPtr;


    class CNOID_EXPORT SgOrthographicCamera : public SgCamera
    {
      public:
        SgOrthographicCamera();
        SgOrthographicCamera(const SgOrthographicCamera& org);
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);

        SgFloat height() const { return height_; }
        void setHeight(SgFloat h) { height_ = h; }

      private:
        SgFloat height_;
    };
    typedef boost::intrusive_ptr<SgOrthographicCamera> SgOrthographicCameraPtr;


    class CNOID_EXPORT SgFog : public SgPreprocessed
    {
      protected:
        SgFog();
        SgFog(const SgFog& org);
      public:
        virtual SgObject* clone(const CopyFlags& copyFlags = DEEP_COPY_NODES) const;
        virtual void accept(SgVisitor& visitor);
    };
    typedef boost::intrusive_ptr<SgFog> SgFogPtr;
    

    class CNOID_EXPORT SgVisitor
    {
    public:
        virtual void visitNode(SgNode* node);
        virtual void visitGroup(SgGroup* group);
        virtual void visitInvariantGroup(SgInvariantGroup* group);
        virtual void visitTransform(SgTransform* transform);
        virtual void visitShape(SgShape* shape);
        virtual void visitPointSet(SgPointSet* pointSet);        
        virtual void visitLineSet(SgLineSet* lineSet);        
        virtual void visitPreprocessed(SgPreprocessed* preprocessed);
        virtual void visitLight(SgLight* light);
        virtual void visitCamera(SgCamera* camera);
    };

}

#endif
