
/**
   @author Shin'ichiro Nakaoka
*/

#include "VRMLToSGConverter.h"
#include <map>
#include <string>
#include <typeinfo>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace cnoid {

    class VRMLToSGConverterImpl
    {
    public:
        VRMLToSGConverterImpl(std::ostream& os__);
        SgNode* convertNode(VRMLNode* vnode);
        SgGroup* convertGroupNode(AbstractVRMLGroup* vgroup);
        SgTransform* createTransformNode(VRMLTransform* vt);
        SgShape* convertShapeNode(VRMLShape* vshape);
        SgMaterial* createMaterial(VRMLMaterial* vm);
        SgMesh* createMeshFromIndexedFaceSet(VRMLIndexedFaceSet* vface);
        SgLight* convertLightNode(VRMLLight* vnode);
        SgPointLight* createPointLight(VRMLPointLight* vlight);
        SgSpotLight* createSpotLight(VRMLSpotLight* vlight);
        SgDirectionalLight* createDirectionalLight(VRMLDirectionalLight* vlight);

        ostream* os_;
        ostream& os() { return *os_; }

        vector<int> removedFaceIndices;
        vector<int> removedFaceVertexIndices;
        vector<int> dummyRemovedFaceIndices;
        
        VRMLMaterialPtr defaultMaterial;

        typedef map<VRMLNode*, SgMaterialPtr> VRMLNodeToSgMaterialMap;
        VRMLNodeToSgMaterialMap vrmlNodeToSgMaterialMap;
        
        typedef map<VRMLNode*, SgNode*> VRMLNodeToSgNodeMap;
        VRMLNodeToSgNodeMap vrmlNodeToSgNodeMap;

        typedef map<VRMLGeometry*, SgMesh*> VRMLGeometryToSgMeshMap;
        VRMLGeometryToSgMeshMap vrmlGeometryToSgMeshMap;
    };
}


VRMLToSGConverter::VRMLToSGConverter(std::ostream& os)
{
    impl = new VRMLToSGConverterImpl(os);
}


VRMLToSGConverterImpl::VRMLToSGConverterImpl(std::ostream& os__)
{
    os_ = &os__;
    dummyRemovedFaceIndices.push_back(std::numeric_limits<SFInt32>::max());
    defaultMaterial = new VRMLMaterial();
}


VRMLToSGConverter::~VRMLToSGConverter()
{
    delete impl;
}


SgNodePtr VRMLToSGConverter::convert(VRMLNodePtr vrmlNode)
{
    if(vrmlNode){
        return impl->convertNode(vrmlNode.get());
    }
    return 0;
}


SgNode* VRMLToSGConverterImpl::convertNode(VRMLNode* vnode)
{
    SgNode* node = 0;

    VRMLNodeToSgNodeMap::iterator p = vrmlNodeToSgNodeMap.find(vnode);
    if(p != vrmlNodeToSgNodeMap.end()){
        node = p->second;

    } else {
        if(VRMLProtoInstance* protoInstance = dynamic_cast<VRMLProtoInstance*>(vnode)){
            vnode = protoInstance->actualNode.get();
        }

        if(vnode){
            if(AbstractVRMLGroup* group = dynamic_cast<AbstractVRMLGroup*>(vnode)){
                node = convertGroupNode(group);
            } else if(VRMLShape* shape = dynamic_cast<VRMLShape*>(vnode)){
                node = convertShapeNode(shape);
            } else if(VRMLLight* light = dynamic_cast<VRMLLight*>(vnode)){
                node = convertLightNode(light);
            }
            if(node){
                node->setName(vnode->defName);
                vrmlNodeToSgNodeMap.insert(make_pair(vnode, node));
            }
        }
    }
        
    return node;
}


SgGroup* VRMLToSGConverterImpl::convertGroupNode(AbstractVRMLGroup* vgroup)
{
    SgGroup* group;

    if(VRMLTransform* transform = dynamic_cast<VRMLTransform*>(vgroup)){
        group = createTransformNode(transform);
    } else {
        group = new SgGroup;
    }
    
    int num = vgroup->countChildren();
    for(int i=0; i < num; i++){
        SgNode* child = convertNode(vgroup->getChild(i));
        if(child){
            group->addChild(child);
        }
    }
    
    if(group->numChildren() == 0){
        delete group;
        group = 0;
    }
    
    return group;
}


SgTransform* VRMLToSGConverterImpl::createTransformNode(VRMLTransform* vt)
{
    Translation3d C(vt->center);
    const AngleAxisd& SR = vt->scaleOrientation;
    Affine3d S;
    S.linear() = vt->scale.asDiagonal();
    S.translation().setZero();
    const AngleAxisd& R = vt->rotation;
    Translation3d T(vt->translation);

    SgTransform* transform = new SgTransform;
    transform->setTransform(T * C * R * SR * S * SR.inverse() * C.inverse());
    return transform;
}


SgShape* VRMLToSGConverterImpl::convertShapeNode(VRMLShape* vshape)
{
    SgShape* shape = 0;

    VRMLGeometry* vrmlGeometry = dynamic_node_cast<VRMLGeometry>(vshape->geometry).get();
    if(vrmlGeometry){
        SgMesh* mesh = 0;
        VRMLGeometryToSgMeshMap::iterator p = vrmlGeometryToSgMeshMap.find(vrmlGeometry);
        if(p != vrmlGeometryToSgMeshMap.end()){
            mesh = p->second;
        } else {
            if(VRMLIndexedFaceSet* faceSet = dynamic_cast<VRMLIndexedFaceSet*>(vrmlGeometry)){
                mesh = createMeshFromIndexedFaceSet(faceSet);
                if(mesh){
                    vrmlGeometryToSgMeshMap.insert(make_pair(vrmlGeometry, mesh));
                }
            }
        }
        if(mesh){
            shape = new SgShape;
            shape->setMesh(mesh);

            VRMLMaterial* vm;
            if(vshape->appearance && vshape->appearance->material){
                vm = vshape->appearance->material.get();
            } else {
                vm = defaultMaterial.get();
            }
            VRMLNodeToSgMaterialMap::iterator p = vrmlNodeToSgMaterialMap.find(vm);
            if(p != vrmlNodeToSgMaterialMap.end()){
                shape->setMaterial(p->second);
            } else {
                shape->setMaterial(createMaterial(vm));
                vrmlNodeToSgMaterialMap.insert(make_pair(vm, shape->material()));
            }
        }
    }

    return shape;
}


SgMaterial* VRMLToSGConverterImpl::createMaterial(VRMLMaterial* vm)
{
    SgMaterial* material = new SgMaterial;
    material->setDiffuseColor(vm->diffuseColor);
    material->setAmbientIntensity(vm->ambientIntensity);
    material->setEmissiveColor(vm->emissiveColor);
    material->setSpecularColor(vm->specularColor);
    material->setShininess(vm->shininess);
    material->setTransparency(vm->transparency);
    return material;
}


SgMesh* VRMLToSGConverterImpl::createMeshFromIndexedFaceSet(VRMLIndexedFaceSet* vface)
{
    SgMesh* mesh = new SgMesh;

    mesh->setSolid(vface->solid);
    
    if(vface->coord){
        mesh->setVertices(new SgVertexArray(vface->coord->point));
    }

    removedFaceIndices.clear();
    removedFaceVertexIndices.clear();

    const MFInt32& coordIndex = vface->coordIndex;
    const int coordIndexSize = coordIndex.size();
    mesh->triangles().reserve((coordIndexSize + 1) * 3 / 4);

    int faceIndex = 0;
    int faceVertexIndex = 0;
    int faceVertexPosBegin = 0;
    int numFaceVertices = 0;
    const bool ccw = vface->ccw;
        
    for(int i=0; i < coordIndexSize; ++i){
        int index = coordIndex[i];
        if(index >= 0){
            ++numFaceVertices;
        } else {
            if(numFaceVertices == 3) { // Triangle ?
                if(ccw){
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin]);
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin + 1]);
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin + 2]);
                } else { // flip
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin]);
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin + 2]);
                    mesh->triangles().push_back(coordIndex[faceVertexPosBegin + 1]);
                }
                faceVertexIndex += 3;
            } else {
                removedFaceIndices.push_back(faceIndex);
                for(int j=0; j < numFaceVertices; ++j){
                    removedFaceVertexIndices.push_back(faceVertexIndex++);
                }
            }
            faceVertexPosBegin = i + 1; // next position
            numFaceVertices = 0;
            ++faceIndex;
        }
    }

    /*
    if(!removedFaceIndices.empty()){
        os() << str(fmt(_("There are %2% non-triangle polygon(s) in a IndexedFaceSet node%1%."))
                    % (vface->defName.empty() ? "" : (string(" ") + vface->defName))
                    % removedFaceIndices.size())
             << endl;
    }
    */

    // Sentinel value
    removedFaceIndices.push_back(std::numeric_limits<SFInt32>::max());
    removedFaceVertexIndices.push_back(std::numeric_limits<SFInt32>::max());

    if(vface->normal){
        
        mesh->setNormals(new SgNormalArray(vface->normal->vector));
        
        if(!vface->normalIndex.empty()){
            const bool perVertex = vface->normalPerVertex;
            const int size = vface->normalIndex.size();
            size_t skipIndex = 0;
            const MFInt32& indicesToSkip = perVertex ? removedFaceVertexIndices : removedFaceIndices;
            mesh->normalIndices().reserve(perVertex ? size : size * 3);

            for(int i=0; i < size; ++i){
                if(i == indicesToSkip[skipIndex]){
                    if(skipIndex < indicesToSkip.size()){
                        ++skipIndex;
                    }
                } else {
                    int normalIndex = vface->normalIndex[i];
                    if(normalIndex >= 0){
                        if(perVertex){
                            mesh->normalIndices().push_back(normalIndex);
                        } else {
                            mesh->normalIndices().push_back(normalIndex);
                            mesh->normalIndices().push_back(normalIndex);
                            mesh->normalIndices().push_back(normalIndex);
                        }
                    }
                }
            }
        }
    }
    
    if(vface->color){
        const MFColor& c = vface->color->color;
        if(!c.empty()){
            mesh->setColors(new SgColorArray(c));
            mesh->setColorPerVertex(vface->colorPerVertex);
            
            if(!vface->colorIndex.empty()){
                const bool perVertex = vface->colorPerVertex;
                const int size = vface->colorIndex.size();
                size_t skipIndex = 0;
                MFInt32& indicesToSkip = perVertex ? removedFaceVertexIndices : removedFaceIndices;
                mesh->colorIndices().reserve(perVertex ? size : size * 3);
                
                for(int i=0; i < size; ++i){
                    if(i == indicesToSkip[skipIndex]){
                        if(skipIndex < indicesToSkip.size()){
                            ++skipIndex;
                        }
                    } else {
                        int colorIndex = vface->colorIndex[i];
                        if(colorIndex >= 0){
                            mesh->colorIndices().push_back(colorIndex);
                        }
                    }
                }
            }
        }
    }

    mesh->updateBoundingBox();
    
    return mesh;
}


SgLight* VRMLToSGConverterImpl::convertLightNode(VRMLLight* vlight)
{
    SgLight* light = 0;

    if(VRMLPointLight* vPointLight = dynamic_cast<VRMLSpotLight*>(vlight)){
        light = createPointLight(vPointLight);
    } else if(VRMLDirectionalLight* vDirectionalLight = dynamic_cast<VRMLDirectionalLight*>(vlight)){
        light = createDirectionalLight(vDirectionalLight);
    }
    if(light){
        light->on(vlight->on);
        light->setColor(vlight->color);
        light->setIntensity(vlight->intensity);
        light->setAmbientIntensity(vlight->ambientIntensity);
    }
    return light;
}


SgPointLight* VRMLToSGConverterImpl::createPointLight(VRMLPointLight* vlight)
{
    SgPointLight* light;
    
    if(VRMLSpotLight* vSpotLight = dynamic_cast<VRMLSpotLight*>(vlight)){
        light = createSpotLight(vSpotLight);
    } else {
        light = new SgPointLight();
    }
    light->setLocation(vlight->location);
    light->setAttenuation(vlight->attenuation);

    return light;
}


SgSpotLight* VRMLToSGConverterImpl::createSpotLight(VRMLSpotLight* vlight)
{
    SgSpotLight* light = new SgSpotLight();
    
    light->setDirection(vlight->direction);
    light->setBeamWidth(vlight->beamWidth);
    light->setCutOffAngle(vlight->cutOffAngle);

    return light;
}


SgDirectionalLight* VRMLToSGConverterImpl::createDirectionalLight(VRMLDirectionalLight* vlight)
{
    SgDirectionalLight* light = new SgDirectionalLight();
    light->setDirection(vlight->direction);
    return light;
}

